#include "acctd-config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <syslog.h>
#include <sysexits.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <limits.h>
#include <time.h>
#include <fcntl.h>

#ifdef HAVE_DB_185_H
#include <db_185.h>
#else
#include <db.h>
#endif

#include "own_wtmp.h"

#ifdef __STDC__
#include <stdarg.h>	/* ANSI C, variable length args */
#else
#include <varargs.h>	/* has 'vararg' definitions */
#endif

#ifndef STDIO_VSNPRINTF
extern int vsnprintf(char *, size_t, const char *, va_list);
#endif

char *pname;
char *utmp_path = TAC_PLUS_UTMPFILE;
char *dir_path = NULL;
int utmp_file = -1;
time_t scan_time = SCANTIME;
time_t reinit = REINIT;
char *killer = ACTPROG;
char *dbfile = TAC_PLUS_DBFILE;
#ifdef ACCTD_NOTIFY
char *notifier = NOTIFYPROG;
int  notifytime = NOTIFYTIME;
#endif

static int debug = 0;
int nodetach = 0;

DB *db = NULL;
DBT key[1], val[1];

#define pidfilename RUNPATH "/acctd.pid"

#ifdef __STDC__
static void report(int priority, char *fmt, ...)
#else /* VARARGS2 */
static void report(priority, fmt, va_alist)
int priority;
char *fmt;
va_dcl                        /* no terminating semi-colon */
#endif
{
  char msg[1024];             /* temporary string */
  va_list ap;

#ifdef __STDC__
  va_start(ap, fmt);
#else
  va_start(ap);
#endif
  vsnprintf(msg, sizeof(msg), fmt, ap);
  va_end(ap);

  syslog(priority, "%s", msg);
  if (debug) printf("%s\n", msg);
}

#if defined(CRON) && defined(STRICT_TIME_CHECK) || defined(PARROT)
#include "tac_plus.h"
#endif

extern int find_ug(DB *, TQ_st *, char *, int (*)(TQ_st *), DBT *, char **);
extern int find_ug_tl_cond(TQ_st *);
extern int find_ug_dl_cond(TQ_st *);
#ifdef PARROT
static double parrot_off=PARROT_OFF; /* avoid rounding errors */
extern int find_ug_pl_cond(TQ_st *);
extern FILE *find_parrot_file(char *);
#endif

unsigned long bindaddress;

#if defined(CRON) && defined(STRICT_TIME_CHECK)
extern int cfg_read_config(char *);
extern int set_time_status(char *, struct authen_data *, struct passwd *, int);
/* int initialised = 0; */
char *cfgfile=TAC_PLUS_CFGFILE;

int strict_time_check(char *name, char *nas, char *port) {
  struct authen_data a1;
  struct identity a2;
  struct passwd pw;

  bzero(&a2, sizeof(a2));
  a2.username = name;
  a2.NAS_name = nas;
  a2.NAS_port = port;

  bzero(&a1, sizeof(a1));
  a1.NAS_id = &a2;

  bzero(&pw, sizeof(pw));
  pw.pw_name = name;
  pw.pw_uid = NOPUID;
  pw.pw_gid = NOPGID;
  pw.pw_dir = NOPHOME;
  pw.pw_shell = NOPSHELL;

  return set_time_status(name, &a1, &pw, TAC_PLUS_RECURSE);
}
#endif

void usage()
{
  fprintf(stderr, "\
Usage: %s [-I] [-d] [-w file] [-r [dir]] [-t time] [-a [time]] [-k path]\n\
 -I          - don't detach from controlling terminal\n\
 -d          - run in debug mode (do not daemonize)\n\
 -w <file>   - use utmp file <file> instead of %s\n\
 -u <file>   - use user database file <file> instead of %s\n\
 -r [<dir>]  - place restrict files in <dir> (%s as default) directory\n\
 -t <time>   - scan wtmp each <time> seconds instead of %d\n\
 -a <time>   - kill users with great than <time> seconds on-line time\n\
               (if <time>==0, then never) instead of %d\n\
 -k <path>   - path to killer program instead of %s\n"
#ifdef ACCTD_NOTIFY
" -n <path>   - path to notifier program instead of %s\n\
 -o <time>   - send notification when left <time> minutes instead of %d\n"
#endif
, pname, TAC_PLUS_UTMPFILE, TAC_PLUS_DBFILE, TAC_PLUS_DENYDIR, SCANTIME, REINIT, ACTPROG
#ifdef ACCTD_NOTIFY
, NOTIFYPROG, NOTIFYTIME
#endif
);
  exit(EX_USAGE);
}

void hangup(int s) {
  report(LOG_INFO, "got %d signal -- exiting", s);
  if (db)
    (*db->close)(db);
  closelog();
  unlink(pidfilename);
  exit(EX_OK);
}

void killuser(UTMP_T *w, int act) {
  int pid;

  if ((pid = fork())) {
    int c;
    waitpid(pid, &c, 0);
    return;
  }
#ifdef HAVE_DAEMON
  daemon(1, 0);
#else
  setsid();
#endif
  if ((pid = fork())) {
    sleep(1);
    exit(EX_OK);
  }
#ifdef HAVE_DAEMON
  daemon(1, 0);
#else
  setsid();
#endif
  if (dir_path) {
    time_t t;
    char lck[256];
    FILE *f;

    time(&t);
    if (dir_path && act) {
      if(sizeof(lck) <= strlen(dir_path) + strlen(w->ut_user) + 1)
	report(LOG_ERR, "can't create deny file for %s: internal buffer overflow", w->ut_user);
      else {
	sprintf(lck, "%s/%s", dir_path, w->ut_user);
	if ((f = fopen(lck, "w"))) {
	  fprintf(f, "\n*** Sorry, your time limit is exceeded at %s", ctime(&t));
	  fclose(f);
	} else
	  report(LOG_ERR, "can't create deny file %s: %s", lck, strerror(errno));
      }
    }
  }
  execl(killer, "-killuser", w->ut_user, w->ut_line, w->ut_host, NULL);
  report(LOG_ERR, "can't exec %s %s %s %s: %s", killer, w->ut_user, w->ut_line, w->ut_host, strerror(errno));
  exit(EX_OSERR);
}

#ifdef CRON
#include "../lib/libcron/cron.h"
#endif

void db_update(time_t t, UTMP_T *w) {
#ifdef DBF_1
  TQ_st tq;
#else
  time_t tq;
#endif
  int res, kill = 0;
  char *gname;

#ifdef DBF_1
  res = find_ug(db, &tq, w->ut_user, find_ug_tl_cond, key, &gname);
  switch (res) {
  case 0: /* found valid user record */
    tq.tot -= t;
    break;
  case 1: /* found valid group record */
    /* report(LOG_INFO, "db_update tl: found for user %s in group %s", w->ut_user, key->data); */
    break;
  case 3: /* not found in get_hash_user (should not happen?), user record present */
    tq.tot -= t;
  case 2: /* not found in get_hash_user (should not happen?), no user record */
    /* report(LOG_ERR, "db_update tl: not found user info for %s", w->ut_user); */
    break;
  case 5: /* no groups found, user record present */
    tq.tot -= t;
  case 4: /* no groups found, no user record */
    /* report(LOG_INFO, "db_update tl: not found groups containing user %s", w->ut_user); */
    break;
  case 7: /* not found in groups, user record present */
    tq.tot -= t;
  case 6: /* not found in groups, no user record */
    /* report(LOG_INFO, "db_update tl: for %s in groups exceeded", w->ut_user); */
    break;
  }
  if (res < 2 || res % 2) { /* update u/g record */
    if (tq.tl != TIME_LIMIT_OFF) tq.tll += t;
    val->data = (void *) &tq;
    val->size = sizeof(tq);
    (*db->put)(db, key, val, NULL);
  } /* else
    report(LOG_INFO, "not updating tl for %s", w->ut_user); */
  if (!kill && (kill = (res < 2 && !find_ug_tl_cond(&tq) || res > 2 && res != 4))) {
    report(LOG_INFO, "user %s on line %s nas %s total time limit reached", w->ut_user, w->ut_line, w->ut_host);
    killuser(w, 1);
  }

  res = find_ug(db, &tq, w->ut_user, find_ug_dl_cond, key, &gname);
  switch (res) {
  case 0: /* found valid user record */
    break;
  case 1: /* found valid group record */
    /* report(LOG_INFO, "db_update dl: found for user %s in group %s", w->ut_user, key->data); */
    break;
  case 2: /* not found in get_hash_user (should not happen?), no user record */
  case 3: /* not found in get_hash_user (should not happen?), user record present */
    /* report(LOG_ERR, "db_update dl: not found user info for %s", w->ut_user); */
    break;
  case 4: /* no groups found, no user record */
  case 5: /* no groups found, user record present */
    /* report(LOG_INFO, "db_update dl: not found groups containing user %s", w->ut_user); */
    break;
  case 6: /* not found in groups, no user record */
  case 7: /* not found in groups, user record present */
    /* report(LOG_INFO, "db_update dl: for %s in groups exceeded", w->ut_user); */
    break;
  }
  if (res < 2 || res % 2) { /* update u/g record */
    if (tq.dl != TIME_LIMIT_OFF) {
      tq.dll += t;
      val->data = (void *) &tq;
      val->size = sizeof(tq);
      (*db->put)(db, key, val, NULL);
    }
  } /* else
    report(LOG_INFO, "not updating dl for %s", w->ut_user); */
  if (!kill && (kill = (res < 2 && !find_ug_dl_cond(&tq) || res > 2 && res != 4))) {
    report(LOG_INFO, "user %s on line %s nas %s day time limit reached", w->ut_user, w->ut_line, w->ut_host);
    killuser(w, 1);
  }

#ifdef PARROT
  res = find_ug(db, &tq, w->ut_user, find_ug_pl_cond, key, &gname);
  switch (res) {
  case 0: /* found valid user record */
    break;
  case 1: /* found valid group record */
    /* report(LOG_INFO, "db_update pl: found for user %s in group %s", w->ut_user, key->data); */
    break;
  case 2: /* not found in get_hash_user (should not happen?), no user record */
  case 3: /* not found in get_hash_user (should not happen?), user record present */
    /* report(LOG_ERR, "db_update pl: not found user info for %s", w->ut_user); */
    break;
  case 4: /* no groups found, no user record */
  case 5: /* no groups found, user record present */
    /* report(LOG_INFO, "db_update pl: not found groups containing user %s", w->ut_user); */
    break;
  case 6: /* not found in groups, no user record */
  case 7: /* not found in groups, user record present */
    /* report(LOG_INFO, "db_update pl: for %s in groups exceeded", w->ut_user); */
    break;
  }
  if (res < 2 || res % 2) { /* update u/g record */
    if (tq.parrot != parrot_off) {
      double veloc = -1.0;
      FILE *f = find_parrot_file(key->data);

      if (f) {
	struct passwd pw;
	char *envp[] = { NULL };
	time_t t = time(NULL);
	struct tm *tm = localtime(&t);
	entry *e;
	int minute, hour, dom, month, dow;

	bzero(&pw, sizeof(pw));
	pw.pw_name = key->data;
	pw.pw_uid = NOPUID;
	pw.pw_gid = NOPGID;
	pw.pw_dir = NOPHOME;
	pw.pw_shell = NOPSHELL;

	minute = tm->tm_min -FIRST_MINUTE;
	hour = tm->tm_hour -FIRST_HOUR;
	dom = tm->tm_mday -FIRST_DOM;
	month = tm->tm_mon +1 /* 0..11 -> 1..12 */ -FIRST_MONTH;
	dow = tm->tm_wday -FIRST_DOW;

	while ((e = load_entry(f, NULL, &pw, envp))) {
	  if(bit_test(e->minute, minute)
	     && bit_test(e->hour, hour)
	     && bit_test(e->month, month)
	     && ( ((e->flags & DOM_STAR) || (e->flags & DOW_STAR))
		  ? (bit_test(e->dow,dow) && bit_test(e->dom,dom))
		  : (bit_test(e->dow,dow) || bit_test(e->dom,dom))
		  )
	     ) {
	    veloc = atof(e->cmd); /* FIXME use strtod with diagnostic msg */
	    free_entry(e);
	    break;
	  }
	  free_entry(e);
	}
	fclose(f);

	if (veloc < 0.0)
	  report(LOG_INFO, "user %s: no parrot rule for current moment", w->ut_user);
      } else
	report(LOG_INFO, "parrot file not found for %s", w->ut_user);
      if (veloc < 0.0)
	veloc = 0.0;

      tq.parrot += t * veloc / PARROT_INTERVAL;
      val->data = (void *) &tq;
      val->size = sizeof(tq);
      (*db->put)(db, key, val, NULL);
#ifdef ACCTD_NOTIFY
      if (res < 2) {
	if (tq.parrot + tq.parrot_credit <= notifytime * veloc
	    && tq.parrot + tq.parrot_credit > (notifytime-1) * veloc) {
	  int pid;
	  report(LOG_INFO, "notifying user %s on line %s nas %s about expiration of parrot limit", w->ut_user, w->ut_line, w->ut_host);
	  if ((pid = fork())) {
	    int c;
	    waitpid(pid, &c, 0);
	    return;
	  }
#ifdef HAVE_DAEMON
	  daemon(1, 0);
#else
	  setsid();
#endif
	  if ((pid = fork())) {
	    sleep(1);
	    exit(EX_OK);
	  }
#ifdef HAVE_DAEMON
	  daemon(1, 0);
#else
	  setsid();
#endif
	  /* send notify to user */
	  execl(notifier, "-notifier", w->ut_user, w->ut_line, w->ut_host, inet_ntoa(w->ut_addr), NULL);
	  report(LOG_ERR, "can't exec %s %s %s %s %s: %s",
		 notifier, w->ut_user, w->ut_line, w->ut_host, inet_ntoa(w->ut_addr), strerror(errno));
	  exit(EX_OSERR);
	}
      }
#endif
    }
  } /* else
    report(LOG_INFO, "not updating pl for %s", w->ut_user); */
  if (!kill && (kill = (res < 2 && !find_ug_pl_cond(&tq) || res > 2 && res != 4))) {
    report(LOG_INFO, "user %s on line %s nas %s parrot limit reached", w->ut_user, w->ut_line, w->ut_host);
    killuser(w, 1);
  }
#endif /* PARROT */

#else /* DBF_1 */
  key->data = w->ut_user;
  key->size = strlen(w->ut_user) + 1;
  if ((*db->get) (db, key, val, NULL)) {
    report(LOG_INFO, "no such user: %s", w.ut_user);
    return;
  }

  tq = *(time_t *)val->data;
  tq += t;
  if (tq < 0) {
    killuser(w, 1);
    report(LOG_INFO, "user %s on line %s nas %s time limit reached", w->ut_user, w->ut_line, w->ut_host);
    tq = 0;
  }

  val->data = (void *) &tq;
  val->size = sizeof(tq);
  (*db->put) (db, key, val, NULL);
#endif /* DBF_1 */
}

int main(int ac, char *av[]) {
  int ch;
  UTMP_T w;
#ifndef HAVE_DAEMON
  int nf;
#endif
  FILE *pidfile;
  pid_t pid, pid_old; /* Our pid */
  time_t told, tnow;

  pname = strrchr(av[0], '/');
  pname = pname ? ++pname : av[0];

  opterr = 0;
  while ((ch = getopt(ac, av, "Iu:dw:r::t:a:k:"
#ifdef ACCTD_NOTIFY
		      "n:o:"
#endif
		      )) != EOF) {
    if(ch == ':')
      ch = optopt;
    switch (ch) {
    case 'I':
      nodetach = 1;
      break;
    case 'w':
      if(!optarg)
	usage();
      utmp_path = optarg;
      break;
    case 'u':
      if(!optarg)
	usage();
      dbfile = optarg;
      break;
    case 'r':
      dir_path = optarg ? optarg : TAC_PLUS_DENYDIR;
      break;
    case 't':
      if(!optarg)
	usage();
      scan_time = atoi(optarg);
      break;
    case 'a':
      if(!optarg)
	usage();
      reinit = atoi(optarg);
      break;
    case 'k':
      if(!optarg)
	usage();
      killer = optarg;
      break;
#ifdef ACCTD_NOTIFY
    case 'n':
      if(!optarg)
	usage();
      notifier = optarg;
      break;
    case 'o':
      if(!optarg)
	usage();
      notifytime = atoi(optarg);
      break;
#endif
    case 'd':
      debug = 1;
      break;
    default:
      usage();
      break;
    }
  }
  ac -= optind;
  av += optind;

  openlog(pname, LOG_PID | LOG_NDELAY, LOG_FACIL);

  utmp_file = open(utmp_path, O_RDONLY);
  if (utmp_file < 0) {
    report(LOG_ERR, "unable to open wtmp file %s: %s", utmp_path, strerror(errno));
    exit(EX_OSFILE);
  }
  signal(SIGHUP, hangup);
  signal(SIGINT, hangup);
  signal(SIGQUIT, hangup);
  signal(SIGTERM, hangup);

  report(LOG_INFO, "started (version %s)", ACCTD_VERSION);

  if (!debug && !nodetach) {
#ifdef HAVE_DAEMON
    daemon(1, 0);
#else
    nf = open("/dev/null", O_RDWR, 0600);
    dup2(nf, 0);
    dup2(nf, 1);
    dup2(nf, 2);
    close(nf);
    setsid();
    if (fork())
      exit(0);
#endif
  }

#if defined(CRON) && defined(STRICT_TIME_CHECK) || defined(PARROT)
  parser_init();

  /* if (initialised) */
    cfg_clean_config();

  report(LOG_INFO, "Reading config");

  if (!cfgfile) {
    report(LOG_ERR, "no config file specified");
    return 1;
  }

  if (cfg_read_config(cfgfile)) {
    report(LOG_ERR, "Parsing %s", cfgfile);
    return 1;
  }
#endif

/* write pid to file -- patch from Ivan V. Sidorenko */
  pid = getpid();
  if ((pidfile = fopen(pidfilename, "r")) != NULL) { /* file exists */
    fscanf(pidfile,"%d",(int *)&pid_old);
    if (kill(pid_old,0) != -1) {
      printf("File .PID exists, process running!\n");
      exit(0);
    } else {
      printf("File .PID exists, process not running!\n");
      fclose(pidfile);
      unlink(pidfilename);
    }
  }
  if ((pidfile = fopen(pidfilename, "w")) != NULL) {
    fprintf(pidfile, "%d\n", (int)pid);
    fclose(pidfile);
  } else {
    printf("Can't open file %s: %s\n", pidfilename, strerror(errno));
    exit(0);
  }

  told = tnow = time(NULL);
  while (1) {
    sleep(scan_time);
    /* FIXME time interval between two sleep()s may be > than scan_time
     * because of update time; alarm() is more robust than
     * while(1) { sleep(); update(); }
     */
    if (!tac_lockfd(utmp_path, utmp_file, 0)) {
      report(LOG_WARNING, "cannot lock utmp file %s: %s", utmp_path, strerror(errno));
      exit(EX_OSFILE);
    }
    db = dbopen(dbfile, O_RDWR, 0600, DB_HASH, NULL);
    if (!db) {
      report(LOG_ERR, "unable to open user database %s: %s", dbfile, strerror(errno));
      exit(EX_OSFILE);
    }
    if (!tac_lockfd(dbfile, (*db->fd)(db), 1)) {
      report(LOG_ERR, "unable to lock user database %s: %s", dbfile, strerror(errno));
      exit(EX_OSFILE);
    }
    lseek(utmp_file, 0L, SEEK_SET);
    tnow = time(NULL);
    while (read(utmp_file, &w, sizeof(w)) == sizeof(w)) {
      if (!*w.ut_user)
	continue;
      w.ut_user[UT_NAMESIZE-1] = '\0'; w.ut_line[UT_LINESIZE-1] = '\0'; w.ut_host[UT_HOSTSIZE-1] = '\0';
      db_update(told - tnow, &w);
      if (reinit && (tnow - w.ut_time) > reinit) {
	report(LOG_INFO, "user %s on line %s nas %s logged in too long", w.ut_user, w.ut_line, w.ut_host);
	killuser(&w, 0);
      }
#if defined(CRON) && defined(STRICT_TIME_CHECK)
      if (strict_time_check(w.ut_user, w.ut_host, w.ut_line)) {
	report(LOG_INFO, "user %s on line %s nas %s time interval is over", w.ut_user, w.ut_line, w.ut_host);
	killuser(&w, 0);
      }
#endif
    }
    tac_unlockfd(utmp_path, utmp_file, 0);
    told = tnow;
    (*db->close)(db);
    db = NULL;
  }
}
