#include "acctd-config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <syslog.h>
#include <string.h>
#include <errno.h>
#include "getopt.h"

#include <sys/types.h>
#include <limits.h>

#ifdef HAVE_DB_185_H
#include <db_185.h>
#else
#include <db.h>
#endif

#include <sysexits.h>
#include <fcntl.h>
#include <time.h>
#include <ctype.h>

#ifdef DBF_1
#define SET_TL		1
#define SET_TLL		2
#define SET_DL		4
#define SET_DLL		8
#define SET_TU		16
#define SET_PL		32
#define SET_PC		64
int	var_flags = 0;

static char	lmt_name[] = LMTNAME;
char	output_string_format[128] = FORMAT;
#endif

typedef enum {
    do_nothing,
    do_create,
    do_delete,
    do_list,
    do_modify,
    do_set,
    do_time
} action;
action	what_to_do = do_nothing;
char	*user  = ""; /* user to operate */
int	fmt_out[12];

DB	*db = NULL;
DBT	key[1], val[1];
int	listonly = 0;

#ifdef PARROT
static double	parrot_off = PARROT_OFF; /* avoid rounding errors */
#endif

int opt_err(char *e) {
    printf("Incorrect parameters (double %s). Use -h to help.\n", e);
    exit(10);
}

void set_key(action num, char *u) {
    if (!what_to_do) {
	what_to_do = num;
	user = u;
    } else
	opt_err ("action");
    return;
}

#ifdef DBF_1
void format_parser(void) {
    int i = 1;
    char *j = output_string_format;

    j = strtok(j, " ,");
    while(j) {
#ifdef PARROT
	if (!strcmp(j,"pl"))  fmt_out[i++] = 1;
	if (!strcmp(j,"pc"))  fmt_out[i++] = 2;
#endif
	if (!strcmp(j,"tl"))  fmt_out[i++] = 3;
	if (!strcmp(j,"tll")) fmt_out[i++] = 4;
	if (!strcmp(j,"dl"))  fmt_out[i++] = 5;
	if (!strcmp(j,"dll")) fmt_out[i++] = 6;
	if (!strcmp(j,"tu"))  fmt_out[i++] = 7;
	j = strtok(NULL," ,");
    }
    return;
}
#endif

void usage(void) {
#ifndef DBF_1
    fprintf (stderr,"Usage: macct -c|d|l|m user|- [--tl|time-lim time]\n");
#else
    fprintf (stderr,
	    "Usage: macct -c|d|l|m|s|t user|-|c [--tl|time-lim time] [--tll|time-lim-left time]"
	    " [--dl|day-lim time] [--dll|day-lim-left time] [--tu|total-used time|c]"
#ifdef PARROT
	    " [--parrot-limit|pl number] [--parrot-credit|pc number]"
#endif
	    " [--format|-f format]\n");
#endif
    exit(EX_USAGE);
}

void openbase(void) {
    struct flock flock;

    if (db)
	return;

    db = dbopen(TAC_PLUS_DBFILE, listonly ? O_RDONLY : O_RDWR | O_CREAT, 0600, DB_HASH, NULL);
    if (!db) {
	perror("unable to open user database " TAC_PLUS_DBFILE);
	exit(EX_OSFILE);
    }
    flock.l_type   = (listonly) ? F_RDLCK : F_WRLCK;
    flock.l_whence = SEEK_SET;
    flock.l_start  = 0L;
    flock.l_len    = 0L;
    if (fcntl((*db->fd)(db), F_SETLKW, &flock) < 0) {
      perror("unable to lock user database " TAC_PLUS_DBFILE);
      exit(EX_OSFILE);
    }
}

int adduser(char *u,
#ifdef DBF_1
        	    long tl, long tll, long dl, long dll, long tu,
#ifdef PARROT
		    double parrot, double parrot_credit,
#endif
	            int tuflag)
#else
              time_t tq)
#endif
{
#ifdef DBF_1
    TQ_st tq;
#endif

    openbase();

    key->data = u;
    key->size = strlen(u) + 1;
    (*db->get) (db, key, val, 0);
    if (val->data) {
	printf("User `%s' already exists.\n", u);
	exit(EX_NOUSER);
    }

#ifdef DBF_1
    tq.tl  = (var_flags & SET_TL ) ? tl  : 0;
    tq.tll = (var_flags & SET_TLL) ? tll : 0;
    tq.dl  = (var_flags & SET_DL ) ? dl  : 0;
    tq.dll = (var_flags & SET_DLL) ? dll : 0;
    tq.tot = (tuflag && (tq.tl != TIME_LIMIT_NO) && (tq.tl != TIME_LIMIT_OFF)) ?
				(tq.tl - tq.tll) : ((var_flags & SET_TU) ? tu : 0);

#ifdef PARROT
    tq.parrot        = (var_flags & SET_PL) ? parrot        : parrot_off;
    tq.parrot_credit = (var_flags & SET_PC) ? parrot_credit : 0;
#endif
#endif

    key->data = u;
    key->size = strlen(u) + 1;
    val->data = (void *) &tq;
    val->size = sizeof(tq);
    return (*db->put) (db, key, val, 0);
}

int deluser(char *u) {
    openbase();
    key->data = u;
    key->size = strlen(u) + 1;
    return (*db->del) (db, key, 0);
}

#ifdef DBF_1
static int ph = 1;
int listprint(char *u, TQ_st *tq) {
    int i = 0;

    if (!strcmp(u, lmt_name))
	return 0;

    if (ph > 0) {
	ph = -1;
	printf(" %-*s", USER_LEN_MAX,"  user");
	while(fmt_out[++i]){
	    switch (fmt_out[i]){
#ifdef PARROT
		case 1:
		    printf("  prt-lim");
		break;
		case 2:
		    printf(" prt-crdt");
		break;
#endif
		case 3:
		    printf("  t-limit");
		break;
		case 4:
		    printf("   t-left");
		break;
		case 5:
		    printf("  d-limit");
		break;
		case 6:
		    printf("   d-left");
		break;
		case 7:
		    printf("   t-used");
		break;
	    }
	}
	printf("\n");
    }

    printf(" %-*s", USER_LEN_MAX, u);
    i = 0;
    while(fmt_out[++i]){
	switch (fmt_out[i]){
#ifdef PARROT
	case 1:
	    if (tq->parrot == parrot_off)
		printf("      off");
	    else
		printf(" %8.4f", tq->parrot);
	break;
	case 2:
	    printf(" %8.4f", tq->parrot_credit);
	break;
#endif
	case 3:
	    switch (tq->tl) {
		case TIME_LIMIT_NO:  printf("  nolimit"); break;
		case TIME_LIMIT_OFF: printf("      off"); break;
		default: printf(" %8li", tq->tl / 60); break;
	    }
	break;
	case 4:
	    printf(" %8li", tq->tll / 60);
	break;
	case 5:
	    switch (tq->dl) {
		case TIME_LIMIT_NO:  printf("  nolimit"); break;
		case TIME_LIMIT_OFF: printf("      off"); break;
		default: printf(" %8li", tq->dl / 60); break;
	    }
	break;
	case 6:
	    printf(" %8li", tq->dll / 60);
	break;
	case 7:
	    printf(" %8li", tq->tot / 60);
	break;
	}
    }
    printf ("\n");
    return 0;
}

int lmtput(void)
{
    char buf[80];

    if(!ph)
	return 0;

    printf("Last modify 'day limits' table: ");
    key->data = lmt_name;
    key->size = strlen(lmt_name) + 1;
    if ((*db->get) (db, key, val, 0))
	printf("undefined\n");
    else {
	strftime(buf, sizeof(buf), "%Y/%m/%d %H:%M:%S", localtime((time_t *) val->data));
	printf("%s\n", buf);
    }
    return (EX_NOUSER);
}
#endif

int listuser(char *u)
{
#ifdef DBF_1
    TQ_st *tq;
#endif

    listonly = 1;
    openbase();

    if (strcmp(u, "-") && strcmp(u, "*")) {
	key->data = u;
	key->size = strlen(u) + 1;
	if (!(*db->get) (db, key, val, 0)) {
#ifdef DBF_1
	    tq = (TQ_st *) val->data;
	    listprint(u, tq);
#else
            printf("*** User `%s', time left %ld minutes\n",
      	           u, *(time_t *)val->data / 60);
#endif
	} else {
	    printf("*** Time limit not set for user `%s'\n", u);
	    exit(EX_NOUSER);
	}
    }
    else {
	while(!(*db->seq) (db, key, val, R_NEXT)) {
#ifdef DBF_1
	    tq = (TQ_st *) val->data;
	    listprint(key->data, tq);
#else
            printf("%-10s %6ld\n", (char *)key->data, *(time_t *)val->data / 60);
#endif
	}
    }
#ifdef DBF_1
    lmtput();
#endif
    return 0;
}

#ifdef DBF_1
int lmtset(char *u)
{
    time_t t;
    struct tm tm;

    openbase();

    if (!strcmp(u,"c") || !strcmp(u, "current"))
	time(&t);
    else {
	sscanf(u, "%d/%d/%d %d:%d:%d", &tm.tm_year, &tm.tm_mon, &tm.tm_mday, &tm.tm_hour, &tm.tm_min, &tm.tm_sec);
	tm.tm_year -= 1900;
	tm.tm_mon--;
	tm.tm_isdst = -1;
	t = mktime(&tm);
    }

    key->data = lmt_name;
    key->size = strlen(lmt_name) + 1;
    val->data = (void *) &t;
    val->size = sizeof(t);
    return (*db->put) (db, key, val, 0);
}
#endif

int moduser(char *u,
#ifdef DBF_1
		    long tl, long tll, long dl, long dll, long tu,
#ifdef PARROT
		    double parrot, double parrot_credit,
#endif
		int tuflag)
#else
		time_t tq)
#endif
{
#ifdef DBF_1
    TQ_st tq;
#endif

    openbase();

    key->data = u;
    key->size = strlen(u) + 1;
    if ((*db->get) (db, key, val, 0)) {
	printf("User `%s' not found\n", u);
	exit(EX_NOUSER);
    }
#ifdef DBF_1
    memcpy(&tq, val->data, sizeof(tq));
    if (what_to_do == do_set) {
	if (var_flags & SET_TL)  tq.tl  = tl;
        if (var_flags & SET_TLL) tq.tll = tll;
	if (var_flags & SET_DL)  tq.dl  = dl;
	if (var_flags & SET_DLL) tq.dll = dll;
	tq.tot = (tuflag && (tq.tl != TIME_LIMIT_NO) && (tq.tl != TIME_LIMIT_OFF))
		? (tq.tl - tq.tll) : ((var_flags & SET_TU) ? tu : tq.tot);
#ifdef PARROT
	if (var_flags & SET_PL) tq.parrot        = parrot;
	if (var_flags & SET_PC) tq.parrot_credit = parrot_credit;
#endif
    }
#endif

    if (what_to_do == do_modify){
#ifdef DBF_1
	if (var_flags & SET_TL) {
	    if ((tq.tl != TIME_LIMIT_NO) && (tq.tl != TIME_LIMIT_OFF))
		tq.tl  += tl;
	    else
		printf ("macct: tl - can't modify `off` or `nolimit` value. Use `-s` option.\n");
	}
	tq.tll += tll;
	if (var_flags & SET_DL) {
	    if ((tq.dl != TIME_LIMIT_NO) && (tq.dl != TIME_LIMIT_OFF))
		tq.dl  += dl;
	    else
		printf ("macct: dl - can't modify `off` or `nolimit` value. Use `-s` option.\n");
	}
	tq.dll += dll;
	tq.tot = (tuflag && (tq.tl != TIME_LIMIT_NO) && (tq.tl != TIME_LIMIT_OFF))
		? (tq.tl - tq.tll) : ((var_flags & SET_TU) ? tu : tq.tot);
#ifdef PARROT
	if (var_flags & SET_PL){
	    if (tq.parrot != parrot_off)
		tq.parrot += parrot;
	    else
		printf ("macct: pl - can't modify `off` value. Use `-s` option.\n");
	}
	tq.parrot_credit += parrot_credit;
#endif

#else
	tq += *(time_t *)val->data;
#endif
    }

    val->data = (void *) &tq;
    val->size = sizeof(tq);
    return (*db->put) (db, key, val, 0);
}

#ifdef DBF_1
int renew_day_limit(void)
{
    DB *dbl;
    TQ_st tq;
    DBT key_l[1], val_l[1];
    char u[17];

    dbl = dbopen(TAC_PLUS_DBFILE, O_RDONLY, 0600, DB_HASH, NULL);
    if (!dbl) {
	perror("unable to open database file (renew) " TAC_PLUS_DBFILE);
	exit(EX_OSFILE);
    }

    while(!(*dbl->seq) (dbl, key_l, val_l, R_NEXT)) {
	if (strcmp(key_l->data, lmt_name)) {
	    strncpy(u, key_l->data, key_l->size);
	    memcpy(&tq, val_l->data, sizeof(tq));
	    if ((tq.dl != TIME_LIMIT_NO) && (tq.dl != TIME_LIMIT_OFF))
		tq.dll = tq.dl;
	    key->data = u;
	    key->size = strlen(u) + 1;
	    val->data = (void *) &tq;
	    val->size = sizeof(tq);
	    (*db->put) (db, key, val, 0);
	}
    }
    (*dbl->close) (dbl);
    return 0;
}

#endif

int main(int argc, char **argv) {

    int c;

#ifdef DBF_1
    long tl    = 0;
    long tll   = 0;
    long dl    = 0;
    long dll   = 0;
    long tu    = 0;
    int tuflag = 0;
#ifdef PARROT
    double pl = 0.0;
    double pc = 0.0;
#endif

# else
    int tl = 0;
#endif

    if (argc == 1) {
#ifdef DBF_1
	format_parser();
#endif
	listuser("-");
	exit(EX_OK);
    }

    while (1){
	int option_index = 0;
	char *my_opt;
	static struct option long_options[] =
	{
    	    {"delete",        1, 0, 0},
    	    {"modify",        1, 0, 0},
    	    {"create",        1, 0, 0},
    	    {"list",          1, 0, 0},
    	    {"help",          0, 0, 0},
#ifdef DBF_1
    	    {"set",           1, 0, 0},
    	    {"time",          1, 0, 0},
#ifdef PARROT
    	    {"parrot-lim",    1, 0, 0},
    	    {"pl",            1, 0, 0},
    	    {"parrot-credit", 1, 0, 0},
    	    {"pc",            1, 0, 0},
#endif
    	    {"time-lim-left", 1, 0, 0},
    	    {"tll",           1, 0, 0},
    	    {"day-lim",       1, 0, 0},
    	    {"dl",            1, 0, 0},
    	    {"day-lim-left",  1, 0, 0},
    	    {"dll",           1, 0, 0},
    	    {"total-used",    1, 0, 0},
    	    {"tu",            1, 0, 0},
    	    {"format",        1, 0, 0},
#endif
    	    {"time-lim",      1, 0, 0},
    	    {"tl",            1, 0, 0}
	};


	c = getopt_long (argc, argv, "c:d:l:m:h"
#ifdef DBF_1
	"s:f:t:"
#endif
	, long_options, &option_index);
	if (c == -1) break;

	switch (c){
        case 0:
	    my_opt = (char*) long_options[option_index].name;

	    if (!strcasecmp(my_opt, "help")) {
		usage();
		break;
	    }
	    if (!strcasecmp(my_opt, "create")) {
		set_key(do_create, optarg);
		break;
	    }
	    if (!strcasecmp(my_opt, "delete")) {
		set_key(do_delete, optarg);
		break;
	    }
	    if (!strcasecmp(my_opt, "list")) {
		set_key(do_list, optarg);
		break;
	    }
	    if (!strcasecmp(my_opt, "modify")) {
		set_key(do_modify, optarg);
		break;
	    }
	    if ((!strcasecmp(my_opt, "time-lim")) || (!strcasecmp(my_opt, "tl"))) {
		if (!tl) {
#ifdef DBF_1
		    var_flags |= SET_TL;
		    if (!strcasecmp(optarg, "off"))
	    		tl = TIME_LIMIT_OFF;
		    else if (!strcasecmp(optarg, "nolimit"))
	    		tl = TIME_LIMIT_NO;
		    else
	        	tl = atol(optarg)*60;
#else
		    tl = atoi(optarg)*60;
#endif
		    break;
		} else
		    opt_err("time limit");
	    }

#ifdef DBF_1
	    if (!strcasecmp(my_opt, "set")) {
		set_key(do_set, optarg);
		break;
	    }
	    if (!strcasecmp(my_opt, "format")) {
		strncpy(output_string_format,optarg,sizeof(output_string_format));
		break;
	    }
	    if (!strcasecmp(my_opt, "time")) {
		set_key(do_time, optarg);
		break;
	    }
	    if ((!strcasecmp(my_opt, "total-used")) || (!strcasecmp(my_opt, "tu"))) {
		if (!tu && !tuflag) {
		    if (!strcasecmp(optarg, "c"))
			tuflag = 1;
		    else
			var_flags |= SET_TU;
			tu = atol(optarg)*60;
		    break;
		} else
		    opt_err("total used");
	    }
	    if ((!strcasecmp(my_opt, "time-lim-left")) || (!strcasecmp(my_opt, "tll"))) {
		if (!tll) {
		    var_flags |= SET_TLL;
		    tll = atol(optarg)*60;
		    break;
		} else
		    opt_err("time limit left");
	    }
	    if ((!strcasecmp(my_opt, "day-lim")) || (!strcasecmp(my_opt, "dl"))) {
		if (!dl) {
		    var_flags |= SET_DL;
		    if (!strcasecmp(optarg, "off"))
	    		dl = TIME_LIMIT_OFF;
		    else if (!strcasecmp(optarg, "nolimit"))
	        	dl = TIME_LIMIT_NO;
		    else
	    		dl = atol(optarg)*60;
		    break;
		} else
		    opt_err("day limit");
	    }
	    if ((!strcasecmp(my_opt, "day-lim-left")) || (!strcasecmp(my_opt, "dll"))) {
		if (!dll) {
		    var_flags |= SET_DLL;
		    dll = atol(optarg)*60;
		    break;
		} else
		    opt_err("day limit left");
	    }

#ifdef PARROT
	    if ((!strcasecmp(my_opt, "parrot-lim")) || (!strcasecmp(my_opt, "pl"))) {
		if (!pl) {
		    var_flags |= SET_PL;
            	    if (!strcasecmp(optarg, "off"))
			pl = parrot_off;
                    else
			pl = atof(optarg);
		    break;
		} else
		    opt_err("parrot limit");
	    }
	    if ((!strcasecmp(my_opt, "parrot-credit")) || (!strcasecmp(my_opt, "pc"))) {
		if (!pc) {
		    var_flags |= SET_PC;
		    pc = atof(optarg);
		    break;
		} else
		    opt_err("parrot credit");
	    }
#endif
#endif
        case 'c':
	    set_key(do_create, optarg);
    	break;

        case 'd':
	    set_key(do_delete, optarg);
    	break;

        case 'l':
	    set_key(do_list, optarg);
    	break;

        case 'm':
	    set_key(do_modify, optarg);
    	break;

#ifdef DBF_1
        case 's':
	    set_key(do_set, optarg);
    	break;

        case 't':
	    set_key(do_time, optarg);
	break;

        case 'f':
	    strncpy(output_string_format, optarg, sizeof(output_string_format));
    	break;
#endif
        case 'h':
        default:
	    usage();
	break;
        }
    }

    if (!what_to_do || ((what_to_do != do_list) && (what_to_do != do_time) &&
	(what_to_do != do_delete) && !tl
#ifdef DBF_1
	&& !var_flags && !tuflag
#endif
    )) {
	printf ("Nothing to do. Use -h to help. Bye.\n\n");
	exit (EX_OK);
    }

#ifdef DBF_1
    format_parser();
#endif

    switch (what_to_do) {
    case do_delete:
	deluser(user);
	break;

    case do_list:
#ifdef DBF_1
	if (!strncmp(user, lmt_name, USER_LEN_MAX)) {
	    printf("User name `%s' is reserved.\n", user);
	    exit(EX_NOUSER);
	}
#endif
	listuser(user);
	break;

#ifdef DBF_1
    case do_time:
	lmtset(user);
	renew_day_limit();
	break;
#endif

    case do_create:
#ifdef DBF_1
	adduser(user, tl, tll, dl, dll, tu,
#ifdef PARROT
		pl, pc,
#endif
		tuflag);
#else
	adduser(user, tl);
#endif
	break;

    case do_modify:
#ifdef DBF_1
    case do_set:
	moduser(user, tl, tll, dl, dll, tu,
#ifdef PARROT
		pl, pc,
#endif
		tuflag);
#else
	moduser(user, tl);
#endif
	break;

    default:
	usage();
	break;
    }

    if(db)
	(*db->close) (db);
    exit(EX_OK);
}
