AC_DEFUN(AC_SEARCH_LIBS1,
[AC_PREREQ([2.13])
AC_CACHE_CHECK([for library containing $1], [ac_cv_search_$1],
[ac_func_search_save_LIBS="$LIBS"
ac_cv_search_$1="no"
AC_TRY_LINK_FUNC([$1], [ac_cv_search_$1="none required"])
test "$ac_cv_search_$1" = "no" && for i in $2; do
LIBS="$i $5 $ac_func_search_save_LIBS"
AC_TRY_LINK_FUNC([$1],
[ac_cv_search_$1="$i"
break])
done
LIBS="$ac_func_search_save_LIBS"])
if test "$ac_cv_search_$1" != "no"; then
  test "$ac_cv_search_$1" = "none required" || LIBS="$ac_cv_search_$1 $LIBS"
  $3
else :
  $4
fi])

AC_DEFUN(AC_SNDLOWAT_WORKS,
[AC_CACHE_CHECK(for working SO_SNDLOWAT, ac_cv_so_sndlowat_works,
[LIBS="$LIBSOCKET"
AC_TRY_RUN([
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
int main(void) {
  int sd = -1, one = 268;
  if ((sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
    return 1;
  if (setsockopt(sd, SOL_SOCKET, SO_SNDLOWAT, (char *)&one, sizeof (one)) < 0)
    return 1;
  return 0;
}
], ac_cv_so_sndlowat_works=yes, ac_cv_so_sndlowat_works=no, ac_cv_so_sndlowat_works=yes)
LIBS=""])
if test $ac_cv_so_sndlowat_works = yes; then
  AC_DEFINE(HAS_SNDLOWAT)
fi
])

AC_DEFUN(AC_ARG_ENA1,
[AC_ARG_ENABLE($1,[[$2]],
[if test "$enableval" = yes -o "$enableval" = no; then
  echo "option --enable-$1 requires a mandatory argument" 2>&1 ; exit 1
fi
$3])])

AC_DEFUN(AC_ARG_ENA2,
[AC_ARG_ENABLE($1,[[$2]],
[if test "$enableval" = yes; then
$3
elif test "$enableval" != no; then
  echo "option --enable-$1 may not be used with an argument" 2>&1 ; exit 1
fi])])

AC_DEFUN(AC_ARG_ENA3,
[AC_ARG_ENABLE($1,[[$2]],
[if test "$enableval" = yes; then
$3
elif test "$enableval" != no; then
  echo "option --enable-$1 may not be used with an argument" 2>&1 ; exit 1
fi],[$3])])

dnl more robust than AC_EGREP_HEADER
AC_DEFUN(AC_CHECK_DEF,
[ac_hdrs=""
for ac_hdr in $1
do
ac_hdrs="$ac_hdrs
#include <$ac_hdr>"
done
for ac_obj in $2
do
ac_hdrs="$ac_hdrs
extern float $ac_obj;"
done
AC_TRY_COMPILE([$ac_hdrs],,[$4],[$3])])

AC_DEFUN(AC_REAPCHILD,
[AC_CACHE_CHECK(whether children are reaped automatically when SIGCHLD is ignored, ac_cv_reapchild,
[AC_TRY_RUN([
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>

int main(void) {
  pid_t pid;
  int status;

  if ((pid = fork()) < 0)
    exit(2); /* can't fork */
  else if (pid > 0) {
    /* parent */
    signal(SIGCHLD, SIG_IGN);
    sleep(3);
    pid = wait3(&status, WNOHANG, 0);
    if (pid <= 0)
      exit(0); /* no children */
    exit(1); /* reaped a child */
  } else {
    /* child */
    sleep(1);
    exit(0);
  }
}], ac_cv_reapchild=yes, ac_cv_reapchild=no, ac_cv_reapchild=no)])
if test $ac_cv_reapchild = yes; then
  AC_DEFINE(REAPCHILD)
fi])

AC_DEFUN(AC_REARMSIGNAL,
[AC_CACHE_CHECK(whether signal handlers need to be rearmed, ac_cv_rearmsignal,
[AC_TRY_RUN([
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <errno.h>

int num = 0;

void handler(int sig) {
  num++;
}

int main(void) {
  pid_t ppid=getpid(), cpid;
  int status;

  signal(SIGUSR1, handler);
  if ((cpid = fork()) < 0) {
    exit(2); /* can't fork */
  }
  else if (cpid > 0) {
    /* parent */
    while (wait3(&status, 0, 0) < 0 && errno == EINTR);
    exit(num!=2);
  } else {
    /* child */
    sleep(1);
    kill(ppid, SIGUSR1);
    sleep(1);
    kill(ppid, SIGUSR1);
    exit(0);
  }
}], ac_cv_rearmsignal=no, ac_cv_rearmsignal=yes, ac_cv_rearmsignal=yes)])
if test $ac_cv_rearmsignal = yes; then
  AC_DEFINE(REARMSIGNAL)
fi])
