#!/usr/bin/perl

# Copyright (c) 2000 Tasic (Taras Heychenko)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

use Pg;
use Sys::Syslog qw(:DEFAULT setlogsock);
setlogsock('unix');
use Date::Manip;

#    date
@ENV{LC_TIME}="C";

#       
$dbname = "tac_log";
$user = "logger";
#     $passwd    pgsql
# $passwd = "<password for user logger here>";
unless(defined $passwd) {
	print "You should define passwd for user logger\n";
	print "You should create user logger with permissions\n";
	print "on create table in DB tac_log. And you should create\n";
	print "tac_log DB in postgresql.\n";
	print " Or you may create another base and make another user\n";
	print "You should check \$dbname \$user \$passwd variables\n";
	print "in text of this script.\n";
	exit 0;
}

#          hash'
# ..      permission ALL to group staff  SELECT
# to group reader
# XXX ,  $user   ,   ALL
%permission = ('ALL','staff','SELECT','reader');

#   
@month[1..12] = ('Jan','Feb','Mar','Apr','May','Jun',
			'Jul','Aug','Sep','Oct','Nov','Dec');

#   TIMEZONE,     Date::Manip
$ENV{TZ}=`date "+%Z"`;
chomp $ENV{TZ};

# ,  utmp  UNIX
$utmp = "utmp";
#  
$utmp_seq = "utmp_id";
$wtmp_seq = "wtmp_id";

#     
$toread = "/usr/local/tac_plus/log/tac.acct";
#$toread = "/usr/homes/tasic/log/tac.acct";

#  pidfile    kill'
$pidfile = "/usr/local/tac_plus/log/log2psql.pid";
#$pidfile = "log2psql.pid";

#        
$laststrfile = "/usr/local/tac_plus/log/last.string";
#$laststrfile = "/usr/homes/tasic/log/last.string";

$inloop = 0; #         

#       ?
$unsafe_exit = 0;
if ( -e $pidfile ) {$unsafe_exit = 1;}

#  pid  $pidfile
open (PFILE, "> $pidfile") || bad_exit('emerg',"$!",100);
print PFILE $$;
close PFILE;

#    kill'
$SIG{TERM} = \&catch_term;
$SIG{KILL} = \&catch_term;
$SIG{INT} = \&catch_term;
$SIG{HUP} = \&catch_hup;
$SIG{USR1} = \&catch_hup;

#     
bad_exit('emerg','Can not find tacacs log file',101) unless ( -r $toread );

#    ,      
($childpid = open LFILE, "-|") || exec 'tail','+1','-F',$toread ;
#open LFILE, "< $toread";

#   
$conn = Pg::connectdb("dbname=$dbname user=$user password=$passwd");
#    ,     
bad_exit('emerg',"Can not connect to base: " . $conn->errorMessage,102)
			unless ( PGRES_CONNECTION_OK eq $conn->status );

#     utmp, utmp_id, wtmp_id
&check_init();

unlink($laststrfile) if ( $unsafe_exit && (-e $laststrfile ));

#         
$curtable = &get_curtable();

#      :     log'
#      .    , 
#        
if( -e $toread . ".0") {
	$toread0 = $toread . ".0";
	open (LFILE0, "< $toread0");
	$first_str0 = <LFILE0>;
	$first_date0 = &ParseDate(substr ($first_str0,0,24));
}

#        =>   
$first_str = <LFILE>;
#       logfile'
$first_date = &ParseDate(substr ($first_str,0,24));

#    by default       
#    ,    
$need_read_0 = 0;
$need_search = 0;

# ,       
if ( -r $laststrfile ) {
#    
	open ( RF, "< $laststrfile") || goto WRONG_LASTSTRFILE;
	$laststr = <RF>;
	close RF;
# ,      
	unlink($laststrfile);
	chomp $laststr;
	$lasttime = substr ($laststr,0,24);
#       
	$lastdate = &ParseDate($lasttime);
	&check_date();
} else {
#    .     
#   .    --    
#  (  ),     .   -
#       &check_date();   
#      

WRONG_LASTSTRFILE:
$checktable = $curtable;
REPEAT_PLS:
#       ?
	$result = $conn->exec("select * from pg_tables
							where tablename='$checktable'");
	if ( ($result->resultStatus eq PGRES_TUPLES_OK) && $result->ntuples ) {
#  =>    
		$result = $conn->exec("select count(*),max(event_time)
								from $checktable");
		if ( ($result->resultStatus eq PGRES_TUPLES_OK) &&
													$result->getvalue(0,0)) {
#    ,       
			$lasttime = $result->getvalue(0,1);
			$lasttime =~ s/\s+\S+\s*$//;
			$lasttime =~ s/^(.{8})0/$1 /;
#        
			$lastdate = &ParseDate($lasttime);
#       
			&check_date();
		} else {
			if ( $checktable eq $curtable ) {
				$checktable = lc $month[$mind-1] . $year;
				goto REPEAT_PLS;
			} else {
#    ,     ,  
#      
				$need_read_0 = 1 if defined $toread0;
				$need_search = 0;
			}
		}
	} else {
		if ( $checktable eq $curtable ) {
			$checktable = lc $month[$mind-1] . $year;
			goto REPEAT_PLS;
		} else {
			$need_read_0 = 1 if defined $toread0;
			$need_search = 0;
		}
	}
}

#   ,    ,     
if ($need_search) {
# ,    
	$nowtime = &ParseDate(`date`);
# ,    ,    => 
	bad_exit('emerg',"Date error: possible wrong system date",103)
			if( &Date_Cmp($lastdate,$nowtime) > 0 );
	if($need_read_0) {
	$fdrd = LFILE0;
#    
		$_ = $first_str0;
#      last  do-loop
LAB1:	{
		do {
			last LAB1 if ( /^\Q$lasttime\E/ );
		} while(<LFILE0>);
		}
#     ,        
#   
		bad_exit('emerg',"Can't find $lasttime in $toread0",104)
				unless ( defined $_ );
		if( defined $laststr ) {
			while(/^\Q$lasttime\E/) {
				last if (/\Q$laststr\E/);
				$_ = <LFILE0>;
			}
#    $laststr,        ,  
#   => -
			bad_exit('emerg',"Can't find $laststr in $toread0",105)
				unless (/^\Q$lasttime\E/);
		}
	} else {
	$fdrd = LFILE;
#          
#   ,  -  
		open (TFD, "tail -1 $toread |");
		$istr = <TFD>;
		close TFD;
		chomp $istr;

#    
		$_ = $first_str;
#      last  do-loop
LAB2:	{
		do {
			last LAB2 if( /\Q$lasttime\E/ );
#           => 
			bad_exit('emerg',"Can't find $lasttime in $toread",106)
				if( /$istr/ );
		} while(<LFILE>);
		}
		if( defined $laststr ) {
			while(/^\Q$lasttime\E/) {
				last if (/\Q$laststr\E/);
				$_ = <LFILE>;
			}
#    $laststr,        ,  
#   => -
			bad_exit('emerg',"Can't find $laststr in $toread0",112)
				unless (/^\Q$lasttime\E/);
		}
	}
} else {
	$_ = $first_str;
	$_ = $first_str0 if ($need_read_0);
}

#      ,   $_   ,  
#    ,    ,  
# 

$cmnt = substr $_,4,3;
$curtable = lc($cmnt) . $year;

#   $curtable --   
$result = $conn->exec("select * from pg_tables where tablename='$curtable'");
&create_table unless (
			($result->resultStatus eq PGRES_TUPLES_OK) && $result->ntuples );

$inloop = 1;

if ($need_search) {
#         
	do {
#        do{}
LAB3: {
#  $_  :       
		chomp;
		/^(\w+\s+\w+\s+\d+\s+\S+\s+\d+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\w+)\s+(.*)/;
		$date = $1;
		$nas_ip = $2;
		$username = $3;
		$nas_port = $4;
		$nac_address = $5;
		$ptype = $6;
		$other = $7;
		undef %avpair;
		foreach $i (split /\s+/,$other) {
			my($a, $b) = split /=/,$i;
			$avpair{$a} = $b;
		}
		$avpair{'service'} = "unknown" unless (defined $avpair{'service'});
#  switch   
		if( $ptype eq "start" ) {
# start 
# ,      ,    
			 $result = $conn->exec("SELECT count(*) FROM $utmp WHERE
					username='$username' AND nas_ip='$nas_ip' AND
					nas_port='$nas_port' AND nac_address='$nac_address'
					AND service='" . $avpair{'service'} . "' AND
					task_id='" . $avpair{'task_id'} . "' AND
					event_time='$date'");
			next LAB3 if ( ($result->resultStatus eq PGRES_TUPLES_OK) &&
					$result->getvalue(0,0));
#      ,   
			if (defined $avpair{'addr'}) {
				$result = $conn->exec("INSERT INTO $utmp (username,
					nas_ip,nas_port,nac_address,service,task_id,
					start_time,event_time,ip) VALUES ('$username',
					'$nas_ip','$nas_port','$nac_address',
					'" . $avpair{'service'} . "','" .
					$avpair{'task_id'} . "',(abstime('epoch'::timestamp)+
					'" . $avpair{'start_time'} . "sec'::timespan)::timestamp,
					'$date','" . $avpair{'addr'} . "')");
			} else {
				$result = $conn->exec("INSERT INTO $utmp (username,
					nas_ip,nas_port,nac_address,service,task_id,
					start_time,event_time) VALUES ('$username',
					'$nas_ip','$nas_port','$nac_address',
					'" . $avpair{'service'} . "','" .
					$avpair{'task_id'} . "',(abstime('epoch'::timestamp)+
					'" . $avpair{'start_time'} . "sec'::timespan)::timestamp,
					'$date')");
			}
		} elsif ( $ptype eq "update" ) {
# update 
#     update' $utmp,      
#    update,   $addr  
			if (defined $avpair{'addr'}) {
				$result = $conn->exec("UPDATE $utmp set update_time='$date',
						ip='" . $avpair{'service'} . "' where username=
						'$username' AND nas_ip='$nas_ip' AND nas_port=
						'$nas_port' AND task_id='" . $avpair{'task_id'} . "'");
			} else {
				$result = $conn->exec("UPDATE $utmp set update_time='$date'
						where username='$username' AND nas_ip=
						'$nas_ip' AND nas_port='$nas_port' AND task_id=
						'" . $avpair{'task_id'} . "'");
			}
		} elsif ( $ptype eq "stop" ) {
# stop 
# ,      ,    
			 $result = $conn->exec("SELECT count(*) FROM $curtable WHERE
						username='$username' AND nas_ip='$nas_ip' AND
						nas_port='$nas_port' AND nac_address='$nac_address'
						AND service='" . $avpair{'service'} . "' AND task_id=
						'" . $avpair{'task_id'} . "' AND event_time='$date'");
			next LAB3 if ( ($result->resultStatus eq PGRES_TUPLES_OK) &&
						$result->getvalue(0,0));
#    ,      $utmp   
# $curtable,   $addr
			unless (defined $avpair{'elapsed_time'}) {
				syslog('err',"Stop packet without elapsed_time $_");
				next LAB3;
			}
			$result = $conn->exec("DELETE FROM $utmp where username='$username'
						AND nas_ip='$nas_ip' AND nas_port='$nas_port' AND
						task_id='" . $avpair{'task_id'} . "' AND nac_address=
						'$nac_address' AND service='" . $avpair{'service'} .
						"'");
			if (defined $avpair{'addr'}) {
				$result = $conn->exec("INSERT INTO $curtable (username,
						nas_ip,nas_port,nac_address,service,task_id,
						start_time,event_time,elapsed_time,ip) VALUES
						('$username','$nas_ip','$nas_port','$nac_address',
						'" . $avpair{'service'} . "','" . $avpair{'task_id'} .
						"',(abstime('epoch'::timestamp)+'" .
						$avpair{'start_time'} . " sec'::timespan)::timestamp,
						'$date','" . $avpair{'elapsed_time'} . "','" .
						$avpair{'addr'} . "')");
			} else {
				$result = $conn->exec("INSERT INTO $curtable (username,
						nas_ip,nas_port,nac_address,service,task_id,
						start_time,event_time,elapsed_time) VALUES
						('$username','$nas_ip','$nas_port','$nac_address',
						'" . $avpair{'service'} . "','" . $avpair{'task_id'}
						. "', (abstime('epoch'::timestamp)+'" .
						$avpair{'start_time'} . " sec'::timespan)::timestamp,
						'$date','" . $avpair{'elapsed_time'} . "')");
			}
		} else {
			syslog('err',"Unknown packet type $_");
		}
		} # end of LAB3
	$_ = <$fdrd>;
	} while(/^\Q$lasttime\E/);
#     $_     , 
#  $lasttime
}

#          
&write_packet();

if($need_read_0) {
#      =>   kill 
#  
	while(<LFILE0>) {
		&write_packet();
	}
}
close LFILE0;

while(<LFILE>) {
	&write_packet();
}
unlink($pidfile);

sub write_packet {
	chomp;
	/^(\w+\s+\w+\s+\d+\s+\S+\s+\d+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\w+)\s+(.*)/;
	$date = $1;
	$nas_ip = $2;
	$username = $3;
	$nas_port = $4;
	$nac_address = $5;
	$ptype = $6;
	$other = $7;
	undef %avpair;
	foreach $i (split /\s+/,$other) {
		my($a, $b) = split /=/,$i;
		$avpair{$a} = $b;
	}
	$avpair{'service'} = "unknown" unless (defined $avpair{'service'});
	unless( defined $avpair{'start_time'} ) {
		$tmpdt = &ParseDate(substr ($_,0,24));
		$avpair{'start_time'} = &UnixDate($tmpdt,"%s");
	}
#  switch   
	if( $ptype eq "start" ) {
# start 
		if (defined $avpair{'addr'}) {
			$result = $conn->exec("INSERT INTO $utmp (username,
					nas_ip,nas_port,nac_address,service,task_id,
					start_time,event_time,ip) VALUES ('$username',
					'$nas_ip','$nas_port','$nac_address',
					'" . $avpair{'service'} . "','" .
					$avpair{'task_id'} . "',(abstime('epoch'::timestamp)+
					'" . $avpair{'start_time'} . "sec'::timespan)::timestamp,
					'$date','" . $avpair{'addr'} . "')");
		} else {
			$result = $conn->exec("INSERT INTO $utmp (username,
					nas_ip,nas_port,nac_address,service,task_id,
					start_time,event_time) VALUES ('$username',
					'$nas_ip','$nas_port','$nac_address',
					'" . $avpair{'service'} . "','" .
					$avpair{'task_id'} . "',(abstime('epoch'::timestamp)+
					'" . $avpair{'start_time'} . "sec'::timespan)::timestamp,
					'$date')");
		}
		if ($result->resultStatus ne PGRES_COMMAND_OK) {
			syslog
				('err',"$0:Error processing string $_ " . $conn->errorMessage);
			return;
		}
	} elsif ( $ptype eq "update" ) {
# update 
#     update' $utmp,   $addr  
		if (defined $avpair{'addr'}) {
			$result = $conn->exec("UPDATE $utmp set update_time='$date',
						ip='" . $avpair{'addr'} . "' where username='$username'
						AND nas_ip='$nas_ip' AND nas_port='$nas_port' AND
						task_id='" . $avpair{'task_id'} . "'");
		} else {
			$result = $conn->exec("UPDATE $utmp set update_time='$date'
						where username='$username' AND nas_ip=
						'$nas_ip' AND nas_port='$nas_port' AND task_id=
						'" . $avpair{'task_id'} . "'");
		}
		if ($result->cmdTuples eq "0") {
			$result = $conn->exec("INSERT INTO $utmp (username,
						nas_ip,nas_port,nac_address,service,task_id,
						start_time,event_time) VALUES ('$username',
						'$nas_ip','$nas_port','$nac_address',
						'" . $avpair{'service'} . "','" . $avpair{'task_id'}
						. "',(abstime('epoch'::timestamp)+'" .
						$avpair{'start_time'} . " sec'::timespan)::timestamp,
						'$date')");
			if ($result->resultStatus ne PGRES_COMMAND_OK) {
				syslog
				('err',"$0:Error processing string $_ " . $conn->errorMessage);
				return;
			}
		}
	} elsif ( $ptype eq "stop" ) {
# stop 
#     $utmp    $curtable,   $addr
		kill 'USR1',$$ unless ( $date =~ /^.{4}\Q$cmnt\E/ );
		unless (defined $avpair{'elapsed_time'}) {
			syslog('err',"Stop packet without elapsed_time $_");
			return;
		}
		$result = $conn->exec("DELETE FROM $utmp where username='$username'
				AND nas_ip='$nas_ip' AND nas_port='$nas_port' AND
				task_id='" . $avpair{'task_id'} . "' AND nac_address=
				'$nac_address' AND service='" . $avpair{'service'} . "'");
		if ($result->resultStatus ne PGRES_COMMAND_OK) {
			system ('warning',"Stop packet without start $_");
		}
		if (defined $avpair{'addr'}) {
			$result = $conn->exec("INSERT INTO $curtable (username,
					nas_ip,nas_port,nac_address,service,task_id,
					start_time,event_time,elapsed_time,ip) VALUES
					('$username','$nas_ip','$nas_port','$nac_address',
					'" . $avpair{'service'} . "','" . $avpair{'task_id'} . "',
					(abstime('epoch'::timestamp)+'" .
					$avpair{'start_time'} . " sec'::timespan)::timestamp,
					'$date','" . $avpair{'elapsed_time'} . "','" .
					$avpair{'addr'} . "')");
		} else {
			$result = $conn->exec("INSERT INTO $curtable (username,
					nas_ip,nas_port,nac_address,service,task_id,
					start_time,event_time,elapsed_time) VALUES
					('$username','$nas_ip','$nas_port','$nac_address',
					'" . $avpair{'service'} . "','" . $avpair{'task_id'} .
					"',(abstime('epoch'::timestamp)+'" .
					$avpair{'start_time'} . " sec'::timespan)::timestamp,
					'$date','" . $avpair{'elapsed_time'} . "')");
		}
		if ($result->resultStatus ne PGRES_COMMAND_OK) {
			syslog
				('err',"$0:Error processing string $_ " . $conn->errorMessage);
			return;
		}
	} else {
		syslog('err',"Unknown packet type $_");
	}
}

sub catch_term {
	if($inloop) {
		open LSF, "> $laststrfile";
		print LSF "$_\n";
		close LSF;
	}
	close LFILE0;
	kill 'TERM',$childpid;
	close LFILE;
	unlink $pidfile;
	exit 0;
}

sub catch_hup {
	return unless($inloop);
	return if( /^.{4}\Q$cmnt\E/);
	$cmnt = substr $_,4,3;
	$curtable = lc($cmnt) . $year;
	$result = $conn->exec
					("select * from pg_tables where tablename='$curtable'");
	&create_table unless (
			($result->resultStatus eq PGRES_TUPLES_OK) && $result->ntuples );
}

# ,  
sub get_curtable {
	($mind, $year) = split /\s+/,`date "+%m %Y"`;
	my $ct = lc $month[$mind] . $year;
	chomp $ct;
	return $ct;
}

sub create_table {
#  
	$result = $conn->exec("CREATE TABLE $curtable (
							username character(16) NOT NULL,
							nas_ip inet NOT NULL,
							nas_port character(15) NOT NULL,
							nac_address character(25) NOT NULL,
							service character(10) NOT NULL,
							task_id int4 NOT NULL,
							start_time timestamp NOT NULL,
							event_time timestamp NOT NULL,
							elapsed_time interval NOT NULL,
							ip inet,
							id int4 DEFAULT nextval ('$wtmp_seq')
							NOT NULL PRIMARY KEY)");
	bad_exit('emerg',"Error intialization $curtable",107)
			unless ($result->resultStatus eq PGRES_COMMAND_OK);
	$result = $conn->exec("CREATE INDEX ${curtable}_username ON
									$curtable (username)");
	bad_exit('emerg',"Error creating index ${curtable}_username",114)
			unless ($result->resultStatus eq PGRES_COMMAND_OK);
	foreach my $prm (keys %permission) {
		$result = $conn->exec("GRANT $prm ON $curtable TO GROUP
								$permission{$prm}");
	}
}

sub check_init {
#    sequence  utmp?
$result = $conn->exec("select * from pg_class where relname='$utmp_seq'
						and relkind='S'");
unless ( ($result->resultStatus eq PGRES_TUPLES_OK) && $result->ntuples ) {
#  -- 
	$result = $conn->exec("CREATE SEQUENCE $utmp_seq increment 1
							maxvalue 2147483647 minvalue 1  cache 1
							cycle");
	bad_exit('emerg',"Error intialization utmp sequence",108)
			unless ($result->resultStatus eq PGRES_COMMAND_OK);
	foreach my $prm (keys %permission) {
		$result = $conn->exec("GRANT $prm ON $utmp_seq TO GROUP
								$permission{$prm}");
	}
}
#    sequence    
$result = $conn->exec("select * from pg_class where relname='$wtmp_seq'
						and relkind='S'");
unless ( ($result->resultStatus eq PGRES_TUPLES_OK) && $result->ntuples ) {
#  -- 
	$result = $conn->exec("CREATE SEQUENCE $wtmp_seq increment 1
							maxvalue 2147483647 minvalue 1  cache 1
							cycle");
	bad_exit('emerg',"Error intialization wtmp sequence",109)
			unless ($result->resultStatus eq PGRES_COMMAND_OK);
	foreach my $prm (keys %permission) {
		$result = $conn->exec("GRANT $prm ON $wtmp_seq TO GROUP
								$permission{$prm}");
	}
}
#     utmp?
$result = $conn->exec("select * from pg_tables where tablename='$utmp'");
unless ( ($result->resultStatus eq PGRES_TUPLES_OK) && $result->ntuples ) {
#  -- 
	$result = $conn->exec("CREATE TABLE $utmp (
							username character(16) NOT NULL,
							nas_ip inet NOT NULL,
							nas_port character(15) NOT NULL,
							nac_address character(25) NOT NULL,
							service character(10) NOT NULL,
							task_id int4 NOT NULL,
							start_time timestamp NOT NULL,
							update_time timestamp,
							event_time timestamp NOT NULL,
							ip inet,
							id int4 DEFAULT nextval ('$utmp_seq')
							NOT NULL PRIMARY KEY)");
	bad_exit('emerg',"Error intialization $utmp",110)
		unless ($result->resultStatus eq PGRES_COMMAND_OK);
	$result = $conn->exec("CREATE INDEX ${utmp}_task_id
							ON $utmp (task_id)");
	bad_exit('emerg',"Error creating index ${utmp}_username",115)
			unless ($result->resultStatus eq PGRES_COMMAND_OK);
	foreach my $prm (keys %permission) {
		$result = $conn->exec("GRANT $prm ON $utmp TO GROUP
								$permission{$prm}");
	}
}
}

sub check_date {
#   ,   
	if( &Date_Cmp($lastdate,$first_date) < 0 ) {
		if (defined $first_date0) {
#      logfile,  :(
			bad_exit('emerg',"Do not know where to begin",111)
				if( &Date_Cmp($lastdate,$first_date0) < 0 );
# -       
			$need_read_0 = 1;
#    
			$need_search = 1;
		} else {
#   ,  
			$need_read_0 = 0;
#  (  )  ,     =>
#  ,       
			$need_search = 0;
		}
	} else {
#      
		$need_read_0 = 0;
#    
		$need_search = 1;
	}
}

sub bad_exit {
	$level = shift;
	$message = shift;
	$exitn = shift;
	syslog($level,"$0: $message");
	unlink($pidfile);
	exit $exitn;
}
