/*-
 * Copyright (c) 1983, 1988, 1989, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 *
 * Changed by Andy Igoshin <ai@vsu.ru>
 *
 */


/*
 * remote login server:
 *	\0
 *	remuser\0
 *	locuser\0
 *	terminal_type/speed\0
 *	data
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#ifdef HAVE_SYS_FILIO_H
#include <sys/filio.h>
#endif
#include <signal.h>
#include <termios.h>
#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#define _KMEMUSER
#ifdef HAVE_SYS_STREAM_H
#include <sys/stream.h>
#endif
#ifdef HAVE_SYS_TTY_H
#include <sys/tty.h>
#endif
#ifdef HAVE_SYS_POLL_H
#include <sys/poll.h>
#endif
#ifdef HAVE_SYS_STRSUBR_H
#include <sys/strsubr.h>
#endif
#ifdef HAVE_SYS_PTYVAR_H
#include <sys/ptyvar.h>
#endif

#include <sys/socket.h>
#include <netinet/in.h>
#ifdef HAVE_NETINET_IN_SYSTM_H
#include <netinet/in_systm.h>
#endif
#ifdef HAVE_NETINET_IP_H
#include <netinet/ip.h>
#endif
#include <arpa/inet.h>
#include <netdb.h>

#if 0
#include <netinet/ip.h>
#include <netinet/in_systm.h>
#include <fcntl.h>
#endif

#include <pwd.h>
#include <syslog.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <strings.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifndef TIOCPKT_WINDOW
#define TIOCPKT_WINDOW 0x80
#endif

/* `defaults' for tty settings.  */
#ifndef TTYDEF_IFLAG
#define	TTYDEF_IFLAG	(BRKINT | ISTRIP | ICRNL | IMAXBEL | IXON | IXANY)
#endif
#ifndef TTYDEF_OFLAG
#ifndef OXTABS
#define OXTABS 0
#endif
#define TTYDEF_OFLAG	(OPOST | ONLCR | OXTABS)
#endif
#ifndef TTYDEF_LFLAG
#define TTYDEF_LFLAG	(ECHO | ICANON | ISIG | IEXTEN | ECHOE|ECHOKE|ECHOCTL)
#endif

#ifndef PATH_DEV
#define PATH_DEV "/dev/"
#endif
#ifndef PATH_UUCICO
#define PATH_UUCICO "/usr/local/lib/uucp/uucico"
#endif

char	*env[2];
#define	NMAX 30
char	lusername[NMAX+1], rusername[NMAX+1];
static	char term[64] = "TERM=";
#define	ENVSIZE	(sizeof("TERM=")-1)	/* skip null for concatenation */
int	keepalive = 1;

void	doit (int, struct sockaddr_in *);
int	control (int, char *, int);
void	protocol (int, int);
void	cleanup (int);
void	fatal (int, char *, int);
int	do_rlogin (struct sockaddr_in *);
void	getstr (char *, int, char *);
void	setup_term (int);
void	usage (void);
int	local_domain (char *);
char	*topdomain (char *);

#if	defined(UUCP_PROXY)
void	uucp_proxy(char *, char *, int);
char	serv_addr[32] = "127.0.0.1";
int	serv_port = 0;
#endif

extern int forkpty(int *, char *, struct termios *, struct winsize *);

int
main(argc, argv)
	int argc;
	char *argv[];
{
	struct sockaddr_in from;
	int ch, on;
	char *pos;
	size_t fromlen;

	openlog("rluucpd", LOG_PID | LOG_CONS, LOG_AUTH);

	opterr = 0;
	while ((ch = getopt(argc, argv, "nS:")) != EOF)
		switch (ch) {
		case 'n':
			keepalive = 0;
			break;
#ifdef UUCP_PROXY
		case 'S':
			bzero(serv_addr, sizeof(serv_addr));
			if ( (pos = strchr(optarg, ':')) != NULL ) {
			  strncpy(serv_addr, optarg,
				  (pos-optarg < sizeof(serv_addr)-1) ? pos-optarg : sizeof(serv_addr)-1);
			  serv_port = atoi(pos+1);
			} else
			  strncpy(serv_addr, optarg, sizeof(serv_addr)-1);
			break;
#endif
		default:
			usage();
			break;
		}
	argc -= optind;
	argv += optind;

	fromlen = sizeof (from);
	if (getpeername(0, (struct sockaddr *)&from, &fromlen) < 0) {
		syslog(LOG_ERR,"Can't get peer name of remote host: %m");
		fatal(STDERR_FILENO, "Can't get peer name of remote host", 1);
	}

	on = 1;
	if (keepalive &&
	    setsockopt(0, SOL_SOCKET, SO_KEEPALIVE, (char *) &on,
		       sizeof (on)) < 0)
		syslog(LOG_WARNING, "setsockopt (SO_KEEPALIVE): %m");

#if defined (IP_TOS) && defined (IPPROTO_IP) && defined (IPTOS_LOWDELAY)
	on = IPTOS_LOWDELAY;
	if (setsockopt(0, IPPROTO_IP, IP_TOS, (char *)&on, sizeof(int)) < 0)
		syslog(LOG_WARNING, "setsockopt (IP_TOS): %m");
#endif

	doit(0, &from);
	return 0;
}

int	child;
int	netf;
char	line[1024];		/* XXX */
int	confirmed;

struct winsize win = { 0, 0, 0, 0 };


void
doit(f, fromp)
	int f;
	struct sockaddr_in *fromp;
{
	int master, pid, on = 1;
	register struct hostent *hp;
	char *hostname;
	const char *raw_hostname;
	char c;

	alarm(60);
	read(f, &c, 1);

	if (c != 0)
		exit(1);

	alarm(0);
	fromp->sin_port = ntohs((u_short)fromp->sin_port);
	hp = gethostbyaddr((char *)&fromp->sin_addr, sizeof(struct in_addr),
	    fromp->sin_family);
	if (hp)
		raw_hostname = hp->h_name;
	else
		raw_hostname = inet_ntoa(fromp->sin_addr);
	hostname = malloc (strlen (raw_hostname) + 1);
	if (! hostname)
		fatal (f, "Out of memory", 0);
	strcpy (hostname, raw_hostname);

	{
		if (fromp->sin_family != AF_INET ||
		    fromp->sin_port >= IPPORT_RESERVED ||
		    fromp->sin_port < IPPORT_RESERVED/2) {
			syslog(LOG_NOTICE, "Connection from %s on illegal port",
				inet_ntoa(fromp->sin_addr));
			fatal(f, "Permission denied", 0);
		}
#ifdef IP_OPTIONS
		{
		u_char optbuf[BUFSIZ/3], *cp;
		char lbuf[BUFSIZ], *lp;
		int ipproto;
		size_t optsize = sizeof(optbuf);
		struct protoent *ip;

		if ((ip = getprotobyname("ip")) != NULL)
			ipproto = ip->p_proto;
		else
			ipproto = IPPROTO_IP;
		if (getsockopt(0, ipproto, IP_OPTIONS, (char *)optbuf,
		    &optsize) == 0 && optsize != 0) {
			lp = lbuf;
			for (cp = optbuf; optsize > 0; cp++, optsize--, lp += 3)
				sprintf(lp, " %2.2x", *cp);
			syslog(LOG_NOTICE,
			    "Connection received using IP options (ignored):%s",
			    lbuf);
			if (setsockopt(0, ipproto, IP_OPTIONS,
			    (char *)NULL, optsize) != 0) {
				syslog(LOG_ERR,
				    "setsockopt IP_OPTIONS NULL: %m");
				exit(1);
			}
		}
		}
#endif
		do_rlogin(fromp);
	}
	if (confirmed == 0) {
		write(f, "", 1);
		confirmed = 1;		/* we sent the null! */
	}
	netf = f;

	pid = forkpty(&master, line, NULL, &win);
	if (pid < 0) {
		if (errno == ENOENT)
			fatal(f, "Out of ptys", 0);
		else
			fatal(f, "Forkpty", 1);
	}
	if (pid == 0) {
		if (f > 2)	/* f should always be 0, but... */
			(void) close(f);
		setup_term(0);
#if	defined(UUCP_PROXY)
		uucp_proxy(lusername, serv_addr, serv_port);
#else
		execle(PATH_UUCICO, "uucico", "-l", "-u", lusername, NULL, env);
		fatal(STDERR_FILENO, PATH_UUCICO, 1);
		/*NOTREACHED*/
#endif
	}
	ioctl(f, FIONBIO, &on);
	ioctl(master, FIONBIO, &on);
	ioctl(master, TIOCPKT, &on);
	signal(SIGCHLD, cleanup);
	protocol(f, master);
	signal(SIGCHLD, SIG_IGN);
	cleanup(0);
}

char	magic[2] = { 0377, 0377 };
char	oobdata[] = {TIOCPKT_WINDOW};

/*
 * Handle a "control" request (signaled by magic being present)
 * in the data stream.  For now, we are only willing to handle
 * window size changes.
 */
int
control(pty, cp, n)
	int pty;
	char *cp;
	int n;
{
	struct winsize w;

	if (n < 4+sizeof (w) || cp[2] != 's' || cp[3] != 's')
		return (0);
	oobdata[0] &= ~TIOCPKT_WINDOW;	/* we know he heard */
	memmove(&w, cp+4, sizeof(w));
	w.ws_row = ntohs(w.ws_row);
	w.ws_col = ntohs(w.ws_col);
	w.ws_xpixel = ntohs(w.ws_xpixel);
	w.ws_ypixel = ntohs(w.ws_ypixel);
	(void)ioctl(pty, TIOCSWINSZ, &w);
	return (4+sizeof (w));
}

/*
 * rlogin "protocol" machine.
 */
void
protocol(f, p)
	register int f, p;
{
	char pibuf[1024+1], fibuf[1024], *pbp, *fbp;
	register int pcc = 0, fcc = 0;
	int cc, nfd, n;
	char cntl;

	/*
	 * Must ignore SIGTTOU, otherwise we'll stop
	 * when we try and set slave pty's window shape
	 * (our controlling tty is the master pty).
	 */
	(void) signal(SIGTTOU, SIG_IGN);
	send(f, oobdata, 1, MSG_OOB);	/* indicate new rlogin */
	if (f > p)
		nfd = f + 1;
	else
		nfd = p + 1;
	if (nfd > FD_SETSIZE) {
		syslog(LOG_ERR, "select mask too small, increase FD_SETSIZE");
		fatal(f, "internal error (select mask too small)", 0);
	}
	for (;;) {
		fd_set ibits, obits, ebits, *omask;

		FD_ZERO(&ebits);
		FD_ZERO(&ibits);
		FD_ZERO(&obits);
		omask = (fd_set *)NULL;
		if (fcc) {
			FD_SET(p, &obits);
			omask = &obits;
		} else
			FD_SET(f, &ibits);
		if (pcc >= 0) {
			if (pcc) {
				FD_SET(f, &obits);
				omask = &obits;
			} else
				FD_SET(p, &ibits);
		}
		FD_SET(p, &ebits);
		if ((n = select(nfd, &ibits, omask, &ebits, 0)) < 0) {
			if (errno == EINTR)
				continue;
			fatal(f, "select", 1);
		}
		if (n == 0) {
			/* shouldn't happen... */
			sleep(5);
			continue;
		}
#define	pkcontrol(c)	((c)&(TIOCPKT_FLUSHWRITE|TIOCPKT_NOSTOP|TIOCPKT_DOSTOP))
		if (FD_ISSET(p, &ebits)) {
			cc = read(p, &cntl, 1);
			if (cc == 1 && pkcontrol(cntl)) {
				cntl |= oobdata[0];
				send(f, &cntl, 1, MSG_OOB);
				if (cntl & TIOCPKT_FLUSHWRITE) {
					pcc = 0;
					FD_CLR(p, &ibits);
				}
			}
		}
		if (FD_ISSET(f, &ibits)) {
				fcc = read(f, fibuf, sizeof(fibuf));
			if (fcc < 0 && errno == EWOULDBLOCK)
				fcc = 0;
			else {
				register char *cp;
				int left, n;

				if (fcc <= 0)
					break;
				fbp = fibuf;

			top:
				for (cp = fibuf; cp < fibuf+fcc-1; cp++)
					if (cp[0] == magic[0] &&
					    cp[1] == magic[1]) {
						left = fcc - (cp-fibuf);
						n = control(p, cp, left);
						if (n) {
							left -= n;
							if (left > 0)
								bcopy(cp+n, cp, left);
							fcc -= n;
							goto top; /* n^2 */
						}
					}
				FD_SET(p, &obits);		/* try write */
			}
		}

		if (FD_ISSET(p, &obits) && fcc > 0) {
			cc = write(p, fbp, fcc);
			if (cc > 0) {
				fcc -= cc;
				fbp += cc;
			}
		}

		if (FD_ISSET(p, &ibits)) {
			pcc = read(p, pibuf, sizeof (pibuf));
			pbp = pibuf;
			if (pcc < 0 && errno == EWOULDBLOCK)
				pcc = 0;
			else if (pcc <= 0)
				break;
			else if (pibuf[0] == 0) {
				pbp++, pcc--;
					FD_SET(f, &obits);	/* try write */
			} else {
				if (pkcontrol(pibuf[0])) {
					pibuf[0] |= oobdata[0];
					send(f, &pibuf[0], 1, MSG_OOB);
				}
				pcc = 0;
			}
		}
		if ((FD_ISSET(f, &obits)) && pcc > 0) {
				cc = write(f, pbp, pcc);
			if (cc < 0 && errno == EWOULDBLOCK) {
				/*
				 * This happens when we try write after read
				 * from p, but some old kernels balk at large
				 * writes even when select returns true.
				 */
				if (!FD_ISSET(p, &ibits))
					sleep(5);
				continue;
			}
			if (cc > 0) {
				pcc -= cc;
				pbp += cc;
			}
		}
	}
}

void
cleanup(signo)
	int signo;
{
	char *p;

	p = line + sizeof(PATH_DEV) - 1;
/*
	if (logout(p))
		logwtmp(p, "", "");
*/
	(void)chmod(line, 0666);
	(void)chown(line, 0, 0);
	*p = 'p';
	(void)chmod(line, 0666);
	(void)chown(line, 0, 0);
	shutdown(netf, 2);
	exit(1);
}

void
fatal(f, msg, syserr)
	int f;
	char *msg;
	int syserr;
{
	int len;
	char buf[BUFSIZ], *bp = buf;

	/*
	 * Prepend binary one to message if we haven't sent
	 * the magic null as confirmation.
	 */
	if (!confirmed)
		*bp++ = '\01';		/* error indicator */
	if (syserr)
		snprintf (bp, sizeof buf - (bp - buf),
				  "rluucpd: %s: %s.\r\n", msg, strerror(errno));
	else
		snprintf (bp, sizeof buf - (bp - buf), "rluucpd: %s.\r\n", msg);
	len = strlen (bp);
	(void) write(f, buf, bp + len - buf);
	exit(1);
}

int
do_rlogin(dest)
	struct sockaddr_in *dest;
{
	getstr(rusername, sizeof(rusername), "remuser too long");
	getstr(lusername, sizeof(lusername), "locuser too long");
	getstr(term+ENVSIZE, sizeof(term)-ENVSIZE, "Terminal type too long");

	return (-1);
}

void
getstr(buf, cnt, errmsg)
	char *buf;
	int cnt;
	char *errmsg;
{
	char c;

	do {
		if (read(0, &c, 1) != 1)
			exit(1);
		if (--cnt < 0)
			fatal(STDOUT_FILENO, errmsg, 0);
		*buf++ = c;
	} while (c != 0);
}

void
setup_term(fd)
	int fd;
{
	register char *cp = strchr (term+ENVSIZE, '/');
	char *speed;
	struct termios tt;

#if 1
	tcgetattr(fd, &tt);
	if (cp) {
		*cp++ = '\0';
		speed = cp;
		cp = strchr (speed, '/');
		if (cp)
			*cp++ = '\0';
#ifdef HAVE_CFSETSPEED
		cfsetspeed(&tt, atoi(speed));
#else
		cfsetispeed(&tt, atoi(speed));
		cfsetospeed(&tt, atoi(speed));
#endif
	}

	tt.c_iflag = TTYDEF_IFLAG;
	tt.c_oflag = TTYDEF_OFLAG;
	tt.c_lflag = TTYDEF_LFLAG;
	tcsetattr(fd, TCSAFLUSH, &tt);
#else
	if (cp) {
		*cp++ = '\0';
		speed = cp;
		cp = strchr (speed, '/');
		if (cp)
			*cp++ = '\0';
		tcgetattr(fd, &tt);
#ifdef HAVE_CFSETSPEED
		cfsetspeed(&tt, atoi(speed));
#else
		cfsetispeed(&tt, atoi(speed));
		cfsetospeed(&tt, atoi(speed));
#endif
		tcsetattr(fd, TCSAFLUSH, &tt);
	}
#endif

	env[0] = term;
	env[1] = 0;
}

void
usage()
{
	syslog(LOG_ERR, "usage: rluucpd [-n]");
}

/*
 * Check whether host h is in our local domain,
 * defined as sharing the last two components of the domain part,
 * or the entire domain part if the local domain has only one component.
 * If either name is unqualified (contains no '.'),
 * assume that the host is local, as it will be
 * interpreted as such.
 */
int
local_domain(h)
	char *h;
{
	extern char *localhost ();
	char *hostname = localhost ();

	if (! hostname)
		return 0;
	else {
		int is_local = 0;
		char *p1 = topdomain (hostname);
		char *p2 = topdomain (h);

		if (p1 == NULL || p2 == NULL || !strcasecmp(p1, p2))
			is_local = 1;

		free (hostname);

		return is_local;
	}
}

char *
topdomain(h)
	char *h;
{
	register char *p;
	char *maybe = NULL;
	int dots = 0;

	for (p = h + strlen(h); p >= h; p--) {
		if (*p == '.') {
			if (++dots == 2)
				return (p);
			maybe = p;
		}
	}
	return (maybe);
}

#if	defined(UUCP_PROXY)
void uucp_proxy(char *lusername, char *serv_addr, int serv_port) {
  int uucp_so = -1;
  struct sockaddr_in uucp_sa;
  struct servent *uucp_s;
  char uucp_bu[32768];
  int cou_r, cou_w, iter, point, flags, one = 1;
  fd_set ready;

    bzero(&uucp_sa, sizeof(uucp_sa));
    uucp_sa.sin_family = AF_INET;
    uucp_sa.sin_addr.s_addr = inet_addr(serv_addr);
    if ( serv_port <= 0 || serv_port > 65535) {
	if ( (uucp_s = getservbyname("uucp", "tcp")) == NULL )
	    serv_port = 540;
	else
	    serv_port = uucp_s->s_port;
    }
    uucp_sa.sin_port = htons(serv_port);

    if ( (uucp_so = socket(PF_INET, SOCK_STREAM, 0)) == -1 ) {
	syslog(LOG_ERR, "uucp_proxy(): Can't create socket: %m");
	return;
    }

    if ( connect(uucp_so, (struct sockaddr *)&uucp_sa, sizeof(uucp_sa)) == -1 ) {
	syslog(LOG_ERR, "uucp_proxy(): Can't connect to `%s:%d': %m", serv_addr, serv_port);
	goto done;
    }

    if ( read(uucp_so, uucp_bu, sizeof(uucp_bu)) == -1 ) {
	point = 1;
	goto err;
    }
    if ( !strcasecmp(uucp_bu, "login: ") ) {
	if ( write(uucp_so, lusername, strlen(lusername)+1) == -1 ) {
	    point = 2;
	    goto err;
	}
    } else {
	point = 3;
	goto err;
    }

#if 0
    flags = fcntl(uucp_so, F_GETFL, 0);
    fcntl(uucp_so, F_SETFL, flags | O_NDELAY | O_NONBLOCK); /* FNDELAY */
    flags = fcntl(0, F_GETFL, 0);
    fcntl(0, F_SETFL, flags | O_NDELAY | O_NONBLOCK);

    if (setsockopt(0, SOL_SOCKET, SO_SNDLOWAT, (char *) &one, sizeof(one)) < 0)
      syslog(LOG_ERR, "setsockopt cisco SNDLOWAT: %m");
    if (setsockopt(0, SOL_SOCKET, SO_RCVLOWAT, (char *) &one, sizeof(one)) < 0)
      syslog(LOG_ERR, "setsockopt cisco RCVLOWAT: %m");
    if (setsockopt(uucp_so, SOL_SOCKET, SO_SNDLOWAT, (char *) &one, sizeof(one)) < 0)
      syslog(LOG_ERR, "setsockopt uucp SNDLOWAT: %m");
    if (setsockopt(uucp_so, SOL_SOCKET, SO_RCVLOWAT, (char *) &one, sizeof(one)) < 0)
      syslog(LOG_ERR, "setsockopt uucp RCVLOWAT: %m");
#endif

#if 0
    for (iter = 0;;iter++) {
	if ( (cou_r = read(uucp_so, uucp_bu, sizeof(uucp_bu))) == -1 ) {
	    if (errno == EAGAIN) goto l1; /* FIXME EWOULDBLOCK */
	    point = 4;
	    goto err;
	}
	syslog(LOG_ERR, "vvv: cou_r1=%i", cou_r);
	if ( cou_r == 0 ) goto done;
	uucp_bu[cou_r]=0; syslog(LOG_ERR, "vvv: uucp_bu=`%s'", uucp_bu);
	if ( (cou_w = write(0, uucp_bu, cou_r)) == -1 ) {
	    point = 5;
	    goto err;
	}
	syslog(LOG_ERR, "vvv: cou_w1=%i", cou_w);
    l1:
	if ( (cou_r = read(0, uucp_bu, sizeof(uucp_bu))) == -1 ) {
	    if (errno == EAGAIN) continue; /* FIXME EWOULDBLOCK */
	    point = 6;
	    goto err;
	}
	syslog(LOG_ERR, "vvv: cou_r2=%i", cou_r);
	if ( cou_r == 0 ) goto done;
	uucp_bu[cou_r]=0; syslog(LOG_ERR, "vvv: uucp_bu=`%s'", uucp_bu);
	if ( (cou_w = write(uucp_so, uucp_bu, cou_r)) == -1 ) {
	    point = 7;
	    goto err;
	}
	syslog(LOG_ERR, "vvv: cou_w2=%i", cou_w);
    }
#endif
    
#if 1
    for (iter=0;;iter++) {
      int n;
      FD_ZERO(&ready);
      FD_SET(uucp_so, &ready);
      FD_SET(0, &ready);
      if ((n = select(uucp_so + 1, &ready, (fd_set *)0, (fd_set *)0, (struct timeval *)0)) < 0) {
	syslog(LOG_INFO, "errno=%i %m", errno);
	if (errno == EINTR) continue;
	goto err;
      }
      syslog(LOG_ERR, "select n=%i", n);
      if (FD_ISSET(uucp_so, &ready)) {
	if ( (cou_r = read(uucp_so, uucp_bu, sizeof(uucp_bu))) == -1 ) {
	  point = 1;
	  goto err;
	}
	syslog(LOG_ERR, "vvv: read %i bytes from uucico", cou_r);
	if (cou_r == 0) goto done;
	uucp_bu[cou_r]=0; syslog(LOG_ERR, "vvv: uucp_bu=`%s'", uucp_bu);
	if ( (cou_w = write(0, uucp_bu, cou_r)) == -1 ) {
	  point = 2;
	  goto err;
	}
	syslog(LOG_ERR, "vvv: wrote %i bytes to cisco", cou_w);
      } else
      if (FD_ISSET(0, &ready)) {
	if ( (cou_r = read(0, uucp_bu, sizeof(uucp_bu))) == -1 ) {
	  point = 3;
	  goto err;
	}
	syslog(LOG_ERR, "vvv: read %i bytes from cisco", cou_r);
	if (cou_r == 0) continue /*goto done*/;
	uucp_bu[cou_r]=0; syslog(LOG_ERR, "vvv: uucp_bu=`%s'", uucp_bu);
	if ( (cou_w = write(uucp_so, uucp_bu, cou_r)) == -1 ) {
	  point = 4;
	  goto err;
	}
	syslog(LOG_ERR, "vvv: wrote %i bytes to uucico", cou_w);
      }
    }
#endif

err:
    syslog(LOG_ERR, "uucp_proxy(): point %d, iter = %d: %m", point, iter);

done:
    if ( uucp_so != -1 ) close(uucp_so);

    exit(0);
}
#endif
