/* pwcheck_dce.c -- check passwords using DCE
 *
 * DCE authentication code
 *
 * Paul Henson <henson@acm.org>
 * California State Polytechnic University, Pomona
 *
 * Copyright (c) 1997-99 Paul Henson
 *
 * Changed by Andy Igoshin <ai@vsu.ru>
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <dce/dce_error.h>
#include <dce/sec_login.h>
#include "tac_plus.h"


int pwcheck_dce(char *userid, char *passwd)
{
  error_status_t dce_st;
  sec_login_handle_t login_context;
  dce_error_string_t dce_error;
  int dce_error_st;
  sec_login_auth_src_t auth_src;
  sec_passwd_rec_t pw_entry;
  sec_passwd_str_t dce_pw;
  boolean32 reset_passwd;

  if (strcmp(userid, "") == 0 || strcmp(passwd, "") == 0) {
        report(LOG_ERR, "pwcheck_dce: userid='%s', passwd='%s'", userid, passwd);
	return 0;
    }

  if (!sec_login_setup_identity(userid, sec_login_no_flags, &login_context, &dce_st)) {
	dce_error_inq_text(dce_st, dce_error, &dce_error_st);
	report(LOG_ERR, "pwcheck_dce: setup identity failed for %s - %s", userid, dce_error);
	sec_login_purge_context(&login_context, &dce_st);
	return 0;
    }

  pw_entry.version_number = sec_passwd_c_version_none;
  pw_entry.pepper = NULL;
  pw_entry.key.key_type = sec_passwd_plain;

  strncpy( (char *)dce_pw, passwd, sec_passwd_str_max_len);
  dce_pw[sec_passwd_str_max_len] = '\0';
  pw_entry.key.tagged_union.plain = &(dce_pw[0]);

  if (!sec_login_valid_and_cert_ident(login_context, &pw_entry, &reset_passwd,
					&auth_src, &dce_st)) {
	dce_error_inq_text(dce_st, dce_error, &dce_error_st);
	report(LOG_ERR, "pwcheck_dce: valid_and_cert identity failed for %s - %s", userid, dce_error);
	sec_login_purge_context(&login_context, &dce_st);
	return 0;
    }

  if (auth_src != sec_login_auth_src_network) {
	sec_login_purge_context(&login_context, &dce_st);
	report(LOG_ERR, "pwcheck_dce: no network creds for %s", userid);
	return 0;
    }

  sec_login_set_context(login_context, &dce_st);
  if (dce_st) {
	dce_error_inq_text(dce_st, dce_error, &dce_error_st);
	report(LOG_ERR, "pwcheck_dce: set context failed for %s - %s", userid, dce_error);
	sec_login_purge_context(&login_context, &dce_st);
	return 0;
    }

    sec_login_purge_context(&login_context, &dce_st);
    return 1;
}
