/*
 * Copyright (c) 1990, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <ctype.h>
#include <err.h>
#include <errno.h>
#include "pwdt.h"
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <db.h>
#include <fcntl.h>

#include "libpasswd.h"
#include "extern.h"

#ifndef LONGPWD
 #define LONGPWD 1
#endif


/*
static uid_t uid;
*/

char   *tempname;

static unsigned char itoa64[] =		/* 0 ... 63 => ascii - 64 */
	"./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

void
to64(s, v, n)
	char *s;
	long v;
	int n;
{
	while (--n >= 0) {
		*s++ = itoa64[v&0x3f];
		v >>= 6;
	}
}

char *
getnewpasswd(pw, lflag)
	struct passwd *pw;
	int lflag;
{
	char *p;
	char buf[_PASSWORD_LEN+1], salt[9];
	int wp = (lflag) ? !LONGPWD : LONGPWD;

	printf("Changing tacacs+ password for %s.\n", pw->pw_name);

	for (buf[0] = '\0';;) {
		snprintf(buf, sizeof(buf),
		    "New password (%d significant characters):",
		    wp ? _PASSWORD_LEN : 8);
		p = getpass(buf);
		if (!*p)
			errx(0, "password unchanged");
		strcpy(buf, p);
		if (!strcmp(buf, getpass("Retype new password:")))
			break;
		printf("Mismatch; try again, EOF to quit.\n");
	}
	/* grab a random printable character that isn't a colon */
	srandom((int)time((time_t *)NULL));
	if (wp) {
		salt[0] = _PASSWORD_EFMT1;
		to64(&salt[1], (long)(29 * 25), 4);
		to64(&salt[5], random(), 4);
	} else
		to64(&salt[0], random(), 2);

	return (crypt(buf, salt));
}

int
local_passwdt(uname, fname, lflag)
	char *uname;
	char *fname;
	int lflag;
{
	struct passwd *new, *old;
        uid_t uid;
	gid_t gid;

	if((uid = geteuid()) != 0)
	 {
#ifdef TAC_PLUS_USERID
          if(uid != TAC_PLUS_USERID)
#endif
	     errx(1, "%s", strerror(EACCES));
	 }
#ifdef TAC_PLUS_USERID
        else
          seteuid(TAC_PLUS_USERID);
#endif

	if((gid = getegid()) != 0)
	 {
#ifdef TAC_PLUS_GROUPID
          if(gid != TAC_PLUS_GROUPID)
#endif
	     errx(1, "%s", strerror(EACCES));
	 }
#ifdef TAC_PLUS_GROUPID
        else
          setegid(TAC_PLUS_GROUPID);
#endif

	setpwfilet(fname);
	if (!(new = getpwnamt(uname)))
		errx(1, "unknown user %s in file %s", uname, _PATH_PASSWDT);
	endpwentt();		/* Close up the database */

	new = pw_copy(new);	/* Make a copy */
	old = pw_copy(new);	/* Make a copy */
	if (new == NULL || old == NULL)
		errx(1, NULL);

	new->pw_passwd = getnewpasswd(new, lflag);
	new->pw_change = 0;
	return (pw_update(old, new, 0));
}
