/*
 * Copyright (c) 1990, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/types.h>
#include <sys/time.h>
#include <ctype.h>
#include <err.h>
#include <errno.h>
#include "pwdt.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "pw_copy.h"
#include "pw_util.h"
#include "extern.h"

#ifndef LONGPWD
 #define LONGPWD 1
#endif

#ifndef __bsdi__
int randinit = 0;
#endif

char   *tempname;

static unsigned char itoa64[] =		/* 0 ... 63 => ascii - 64 */
	"./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

void
to64(s, v, n)
	char *s;
	long v;
	int n;
{
	while (--n >= 0) {
		*s++ = itoa64[v&0x3f];
		v >>= 6;
	}
}

char *
getnewpasswd(pw, lflag)
	struct passwd *pw;
	int lflag;
{
	char *p;
	char buf[_PASSWORD_LEN+1], salt[10];
	int wp = (lflag) ? !LONGPWD : LONGPWD;

        printf("Changing tacacs+ password for %s.\n", pw->pw_name);

	for (buf[0] = '\0';;) {
		snprintf(buf, sizeof(buf),
		    "New password (%d significant characters):",
		    wp ? _PASSWORD_LEN : 8);
		p = getpass(buf);
		if (!*p) {
			printf("Password unchanged.\n");
			pw_error(NULL, 0, 0);
		}
		strcpy(buf, p);
		if (!strcmp(buf, getpass("Retype new password:")))
			break;
		printf("Mismatch; try again, EOF to quit.\n");
	}


#ifdef __bsdi__
	srandom((int)time((time_t *)NULL));
#else
	/* grab a random printable character that isn't a colon */
	if (!randinit) {
		randinit = 1;
		srandomdev();
	}
#endif

        if(wp) {
		salt[0] = _PASSWORD_EFMT1;
		to64(&salt[1], (long)(29 * 25), 4);
		to64(&salt[5], random(), 4);
		salt[9] = '\0';
        } else {
		to64(&salt[0], random(), 2);
                salt[2] = '\0';
        }
	return (crypt(buf, salt));
}

int
local_passwdt(uname, fname, lflag)
	char *uname;
	char *fname;
	int lflag;
{
	struct passwd *pw;
	int pfd, tfd;
        uid_t uid;
	gid_t gid;

	if((uid = geteuid()) != 0)
	 {
#ifdef TAC_PLUS_USERID
          if(uid != TAC_PLUS_USERID)
#endif
	     errx(1, "%s", strerror(EACCES));
	 }
#ifdef TAC_PLUS_USERID
        else
          seteuid(TAC_PLUS_USERID);
#endif

	if((gid = getegid()) != 0)
	 {
#ifdef TAC_PLUS_GROUPID
          if(gid != TAC_PLUS_GROUPID)
#endif
	     errx(1, "%s", strerror(EACCES));
	 }
#ifdef TAC_PLUS_GROUPID
        else
          setegid(TAC_PLUS_GROUPID);
#endif

        setpwfilet(fname);
	if (!(pw = getpwnamt(uname)))
		errx(1, "unknown user %s in file %s", uname, _PATH_PASSWDT);
        endpwentt();

	pw_init();

	/*
	 * Get the new password.  Reset passwd change time to zero by
	 * default. If the user has a valid login class (or the default
	 * fallback exists), then the next password change date is set
	 * by getnewpasswd() according to the "passwordtime" capability
	 * if one has been specified.
	 */
	pw->pw_change = 0;
	pw->pw_passwd = getnewpasswd(pw, lflag);

	pfd = pw_lock();
	tfd = pw_tmp();
	pw_copy(pfd, tfd, pw);

	if (!pw_mkdb(uname))
		pw_error((char *)NULL, 0, 1);
	return (0);
}
