#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/types.h>
#include "radius.h"
#include "radclient.h"
#include "header.h"

/*
 *	Super simple wildcard matcher.
 */
static int matches(char *name, char *wild, char *matchpart)
{
	int len, wlen;
	int ret = 0;

	len = strlen(name);
	wlen = strlen(wild);

	if (len == 0 || wlen == 0) return 0;

	if (wild[0] == '*') {
		wild++;
		wlen--;
		if (wlen <= len && strcmp(name + (len - wlen), wild) == 0) {
			strcpy(matchpart, name);
			matchpart[wlen] = 0;
			ret = 1;
		}
	} else if (wild[wlen - 1] == '*') {
		wild[--wlen] = 0;
		if (wlen <= len && strncmp(name, wild, wlen) == 0) {
			strcpy(matchpart, name + wlen);
			ret = 1;
		}
	}

	return ret;
}

/*
 *	Map a tacacs port to a RADIUS port (serial0:19 --> s19, etc)
 */
int tac_map(unsigned int nas_addr, char *tac_port)
{
  FILE *fp;
  char buf[256], buf2[128];
  char *s, *t;
  int thishost = 0;
  int port = -1;

  if ((fp = fopen(MAPFILE, "r")) == NULL) {
	nsyslog(LOG_ERR, "%s: %m", MAPFILE);
	return -1;
  }

  while(fgets(buf, 256, fp)) {
	if (buf[0] == '#' || buf[0] == '\n')
		continue;
	s = strtok(buf,  " \t\n");
	t = strtok(NULL, " \t\n");
	if (t == NULL) continue;

	if (strcmp(s, "host") == 0)
		thishost = inet_addr(t);

	if (thishost != 0 && thishost != nas_addr)
		continue;

	if (!matches(tac_port, s, buf2))
		continue;

	if (t[0] == '+')
		port = atoi(t + 1) + atoi(buf2);
	else
		port = atoi(t);
	break;
  }
  fclose(fp);

  if (port < 0) {
	nsyslog(LOG_ERR, "tac_map: port %s (NAS %08x) no match",
		tac_port, nas_addr);
	port = 0;
  }

  return port;
}

/*
 *	Print debugging output.
 */
void DEBUG(char *fmt, ...)
{
  va_list ap;
  char buf[256];

  va_start(ap, fmt);
  vsprintf(buf, fmt, ap);
  va_end(ap);
  nsyslog(LOG_DEBUG,buf);
}

/*
 *	Our own syslog version.
 */
void nsyslog(int lvl, char *fmt, ...)
{
  va_list ap;
  char buf1[256], buf2[256], *s = buf1, *pe;
  int l, le;
  int e = errno;

  strncpy(buf1, fmt, sizeof(buf1)-1);
  while((s = strstr(s, "%m")) != NULL) {
    pe = strerror(e);
    le = strlen(pe);
    l = strlen(s);
    if ( ((s - buf1) + le + l) < sizeof(buf1) ) {
      memmove(s + le, s + 2, l - 2);
      memcpy(s, pe, le);
    }
  }

  va_start(ap, fmt);
  vsprintf(buf2, buf1, ap);
  va_end(ap);

  report(lvl, "%s", buf2);
}
