/*
 * Copyright (c) 1989, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <ttyent.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include "own_wtmp.h"
#include "libutil.h"

#define	MAXLINELENGTH	256

static char zapchar;
static FILE *tf;

static char *skip(char *);
static char *value(char *);

struct ttyent *
t_getttyent(status)
int status;
{
	static struct ttyent tty;
	register int c;
	register char *p, *s;
	static char line[MAXLINELENGTH];

	if (!tf && !setttyent())
		return (NULL);
	for (;;) {
		if (!fgets(p = line, sizeof(line), tf))
			return (NULL);
		/* skip lines that are too big */
		if (!strchr(p, '\n')) {
			while ((c = getc(tf)) != '\n' && c != EOF)
				;
			continue;
		}
		while (isspace((int)*p))
			++p;
		if (*p && *p != '#')
			break;
	}
	zapchar = 0;
	tty.ty_name = p;		/* tty	*/
	p = skip(p);
	tty.ty_getty = p;		/* host */
	p = skip(p);

	tty.ty_status = status;

/* --------------------------------------------------------------------
	tty.ty_window = NULL;

#ifdef __FreeBSD__
        tty.ty_group = NULL;
#else
#ifndef __linux__
        tty.ty_auth = NULL;
#endif
#endif
 ------------------------------------------------------------------------ */

#define	scmp(e)	(!strncmp(p, e, sizeof(e) - 1) && isspace(p[sizeof(e) - 1]))
#define	vcmp(e)	(!strncmp(p, e, sizeof(e) - 1) && p[sizeof(e) - 1] == '=')
#define mcmp(e) (!strncmp(s, e, sizeof(e) - 1))
	for (; *p; p = skip(p)) {

                if(!(s = value(p)))		/* no value - no keyword */
                  break;

		     if(vcmp(_TTYS_CHECK))
		 {
		       if(mcmp(_TTYS_YES))
		         tty.ty_status |= A_FLAG_CHECK;
		  else if(mcmp(_TTYS_NO))
		         tty.ty_status &= ~A_FLAG_CHECK;
		 }
		else if(vcmp(_TTYS_PCHECK))
		 {
		       if(mcmp(_TTYS_YES))
		         tty.ty_status |= A_FLAG_PCHECK;
		  else if(mcmp(_TTYS_NO))
		         tty.ty_status &= ~A_FLAG_PCHECK;
		 }
		else if(vcmp(_TTYS_PUTMP) || vcmp(_TTYS_PWTMP))
		 {
		       if(mcmp(_TTYS_ALIAS))
		         tty.ty_status |= A_FLAG_PUTMP_AL;
		  else if(mcmp(_TTYS_TRUE))
		         tty.ty_status &= ~A_FLAG_PUTMP_AL;
		 }
		else if(vcmp(_TTYS_PACCT))
		 {
		       if(mcmp(_TTYS_ALIAS))
		         tty.ty_status |= A_FLAG_PACCT_AL;
		  else if(mcmp(_TTYS_TRUE))
		         tty.ty_status &= ~A_FLAG_PACCT_AL;
		 }
		else if(vcmp(_TTYS_UTMP) || vcmp(_TTYS_WTMP))
		 {
		       if(mcmp(_TTYS_ALIAS))
		        {
		         tty.ty_status &= ~A_FLAG_UTMP_X;
		         tty.ty_status |= A_FLAG_UTMP_AL;
		        }
		  else if(mcmp(_TTYS_TRUE))
		        {
		         tty.ty_status &= ~A_FLAG_UTMP_X;
		         tty.ty_status |= A_FLAG_UTMP_TR;
		        }
		  else if(mcmp(_TTYS_ASIS))
		        {
		         tty.ty_status &= ~A_FLAG_UTMP_X;
		         tty.ty_status |= A_FLAG_UTMP_AS;
		        }
		  else if(mcmp(_TTYS_NO))
		         tty.ty_status &= ~A_FLAG_UTMP_X;
		 }
		else if(vcmp(_TTYS_ACCT))
		 {
		       if(mcmp(_TTYS_ALIAS))
		        {
		         tty.ty_status &= ~A_FLAG_ACCT_X;
		         tty.ty_status |= A_FLAG_ACCT_AL;
		        }
		  else if(mcmp(_TTYS_TRUE))
		        {
		         tty.ty_status &= ~A_FLAG_ACCT_X;
		         tty.ty_status |= A_FLAG_ACCT_TR;
		        }
		  else if(mcmp(_TTYS_ASIS))
		        {
		         tty.ty_status &= ~A_FLAG_ACCT_X;
		         tty.ty_status |= A_FLAG_ACCT_AS;
		        }
		  else if(mcmp(_TTYS_NO))
		         tty.ty_status &= ~A_FLAG_ACCT_X;
		 }
/* --------------------------------------------------------------------
		if (scmp(_TTYS_OFF))
			tty.ty_status &= ~TTY_ON;
		else if (scmp(_TTYS_ON))
			tty.ty_status |= TTY_ON;
#ifdef __FreeBSD__
                else if (scmp(_TTYS_DIALUP))
                        tty.ty_status |= TTY_DIALUP;
                else if (scmp(_TTYS_SECURE))
                        tty.ty_status |= TTY_SECURE;
                else if (vcmp(_TTYS_WINDOW))
                        tty.ty_window = value(p);
#else
#ifdef __linux__
                else if (scmp(_TTYS_SECURE))
                        tty.ty_status |= TTY_SECURE;
                else if (vcmp(_TTYS_WINDOW))
                        tty.ty_window = value(p);
#else
		else if (scmp(_TTYS_DIALOUT))
			tty.ty_status |= TTY_DIALOUT;
		else if (scmp(_TTYS_DIALIN))
			tty.ty_status |= TTY_DIALIN;
		else if (scmp(_TTYS_BIDIR))
			tty.ty_status |= TTY_BIDIR;
		else if (scmp(_TTYS_SECURE))
			tty.ty_status |= TTY_SECURE;
		else if (vcmp(_TTYS_WINDOW))
			tty.ty_window = value(p);
		else if (vcmp(_TTYS_AUTH))
			tty.ty_auth = value(p);
#endif
#endif
 ------------------------------------------------------------------------ */
		else
			break;
	}

	if (zapchar == '#' || *p == '#' || *p == 0 || *p == '\n')
         tty.ty_comment = 0;
        else
         {
	  tty.ty_comment = p;
	  if ((p = strchr(p, '\n')))
		*p = '\0';
         }
	return (&tty);
}

#define	QUOTED	1

/*
 * Skip over the current field, removing quotes, and return a pointer to
 * the next field.
 */
static char *
skip(p)
	register char *p;
{
	register char *t;
	register int c, q;

	for (q = 0, t = p; (c = *p) != '\0'; p++) {
		if (c == '"') {
			q ^= QUOTED;	/* obscure, but nice */
			continue;
		}
		if (q == QUOTED && *p == '\\' && *(p+1) == '"')
			p++;
		*t++ = *p;
		if (q == QUOTED)
			continue;
		if (c == '#') {
			zapchar = c;
			*p = 0;
			break;
		}
		if (c == '\t' || c == ' ' || c == '\n') {
			zapchar = c;
			*p++ = 0;
			while ((c = *p) == '\t' || c == ' ' || c == '\n')
				p++;
			break;
		}
	}
	*--t = '\0';
	return (p);
}

static char *
value(p)
	register char *p;
{

	return ((p = strchr(p, '=')) ? ++p : NULL);
}

int
setttyent()
{

	if (tf) {
		rewind(tf);
		return (1);
	} else if ((tf = fopen(TAC_PLUS_TTYSFILE, "r")))
		return (1);
	/* report(LOG_ERR, "fopen(%s): %s", TAC_PLUS_TTYSFILE, strerror(errno)); */
	return (0);
}

int
endttyent()
{
	int rval;

	if (tf) {
		rval = !(fclose(tf) == EOF);
		tf = NULL;
		return (rval);
	}
	return (1);
}

struct ttyent *
getttyent()
{
 return t_getttyent(0);
}
