/*
 * Copyright (c) 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/param.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "own_wtmp.h"
#include "libutil.h"
#define ACCT_TYPE_UPDATE 3

void logintac(char *name, long time, char *host, char *alias, char *port, char *palias,
	int slot, int status, struct in_addr addr, int rec_type) {
    register int fd;
    char *s;
    UTMP_T u, u2;

    /* invalid slot or no utmp logging needed */
    if(slot <= 0  || !(status & A_FLAG_UTMP_X))
	return;

    u.ut_time = time;
    strncpy(u.ut_user, name, UT_NAMESIZE);

    switch(status & A_FLAG_UTMP_X) {
    case A_FLAG_UTMP_TR:
	s = host;
	break;
    case A_FLAG_UTMP_AL:
	s = alias;
	break;
    default:
	s = (status & A_FLAG_FOUND_AL) ? alias : host;
	break;
    }

    strncpy(u.ut_host, s, UT_HOSTSIZE);
    strncpy(u.ut_line, (status & A_FLAG_PUTMP_AL) ? palias : port, UT_LINESIZE);
    u.ut_addr.s_addr = addr.s_addr;

    if ((fd = open(TAC_PLUS_UTMPFILE, O_RDWR|O_CREAT, 0644)) >= 0) {
	if (rec_type == ACCT_TYPE_UPDATE) {
	    lseek(fd, (off_t)(slot * sizeof(UTMP_T)), SEEK_SET);
	    read(fd, &u2, sizeof(UTMP_T));
	    u.ut_time = u2.ut_time;
	}
	lseek(fd, (off_t)(slot * sizeof(UTMP_T)), SEEK_SET);
	write(fd, &u, sizeof(UTMP_T));
	close(fd);
    }
    if ((rec_type != ACCT_TYPE_UPDATE) && ((fd = open(TAC_PLUS_WTMPFILE, O_WRONLY|O_APPEND, 0)) >= 0)) {
	write(fd, &u, sizeof(UTMP_T));
	close(fd);
    }
}
