/*
 * Copyright (c) 1989, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Michael Fischbein.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/time.h>
#include <pwd.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "own_wtmp.h"
#include "libutil.h"
static void output(UTMP_T *);

int main(int argc, char **argv) {
	register char *p;
	UTMP_T usr;
	struct passwd *pw;
	FILE *ufp, *file();
	char *t;
	time_t time();

	switch (argc) {
	case 1:					/* who */
		ufp = file(TAC_PLUS_UTMPFILE);
		fseek(ufp, sizeof (usr), SEEK_SET);
		/* only entries with both name and line fields */
		while (fread((char *)&usr, sizeof(usr), 1, ufp) == 1)
			if (*usr.ut_user && *usr.ut_line)
				output(&usr);
		break;
	case 2:					/* who utmp_file */
		ufp = file(argv[1]);
		fseek(ufp, sizeof (usr), SEEK_SET);
		/* all entries */
		while (fread((char *)&usr, sizeof(usr), 1, ufp) == 1)
			output(&usr);
		break;
	case 3:					/* who am i */
		ufp = file(TAC_PLUS_UTMPFILE);
		fseek(ufp, sizeof (usr), SEEK_SET);

		/* search through the utmp and find an entry for this tty */
		if ((p = ttyname(0))) {
			/* strip any directory component */
			if ((t = strrchr(p, '/')))
				p = t + 1;
			while (fread((char *)&usr, sizeof(usr), 1, ufp) == 1)
				if (usr.ut_user && !strncmp(usr.ut_line, p, UT_LINESIZE)) {
					output(&usr);
					exit(0);
				}
			/* well, at least we know what the tty is */
			strncpy(usr.ut_line, p, UT_LINESIZE);
		} else
			strcpy(usr.ut_line, "tty??");
		pw = getpwuid(getuid());
		strncpy(usr.ut_user, pw ? pw->pw_name : "?", UT_NAMESIZE);
		time(&usr.ut_time);
		*usr.ut_host = '\0';
		output(&usr);
		break;
	default:
		fprintf(stderr, "usage: who [ file ]\n       who am i\n");
		exit(1);
	}
	exit(0);
}

void output(up)
	UTMP_T *up;
{
	char *ctime();

	printf("%-*.*s %-*.*s", (int)UT_NAMESIZE, (int)UT_NAMESIZE, up->ut_user,
	    (int)UT_LINESIZE, (int)UT_LINESIZE, up->ut_line);
	printf("%.12s", ctime(&up->ut_time) + 4);
	if (*up->ut_host)
		printf("\t(%.*s)", (int)UT_HOSTSIZE, up->ut_host);
	if ((up->ut_addr.s_addr) != INADDR_ANY)
	    printf("\t%s", inet_ntoa(up->ut_addr));
	putchar('\n');
}

FILE *
file(name)
	char *name;
{
	extern int errno;
	FILE *ufp;
	char *strerror();

	if (!(ufp = fopen(name, "r"))) {
		fprintf(stderr, "who: %s: %s.\n", name, strerror(errno));
		exit(1);
	}
	return(ufp);
}
