/* 
   Copyright (c) 1995-2000 by Cisco systems, Inc.

   Permission to use, copy, modify, and distribute modified and
   unmodified copies of this software for any purpose and without fee is
   hereby granted, provided that (a) this copyright and permission notice
   appear on all copies of the software and supporting documentation, (b)
   the name of Cisco Systems, Inc. not be used in advertising or
   publicity pertaining to distribution of the program without specific
   prior permission, and (c) notice be given in supporting documentation
   that use, modification, copying and distribution is by permission of
   Cisco Systems, Inc.

   Cisco Systems, Inc. makes no representations about the suitability
   of this software for any purpose.  THIS SOFTWARE IS PROVIDED ``AS
   IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
   WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "tac_plus.h"

#ifdef RAD_ROAMING
#include "radius.h"
#include "radclient.h"
extern void cfg_get_rad_roaming_data(char *, int, struct radius **);
#endif

/*
 *  Come here when we receive an Start Accounting packet
 */

void account();

void
accounting(pak)
u_char *pak;
{
    struct acct *acct_pak;
    u_char *p;
    HDR *hdr;
    u_char *read_packet();
    int i, len;

    if (debug & DEBUG_ACCT_FLAG)
	report(LOG_DEBUG, "Start accounting request");

    hdr = (HDR *) pak;
    acct_pak = (struct acct *) (pak + TAC_PLUS_HDR_SIZE);

    /* Do some sanity checking on the packet */

    /* arg counts start here */
    p = pak + TAC_PLUS_HDR_SIZE + TAC_ACCT_REQ_FIXED_FIELDS_SIZE;

    /* Length checks */
    len = TAC_ACCT_REQ_FIXED_FIELDS_SIZE;
    len += acct_pak->user_len + acct_pak->port_len + 
	acct_pak->rem_addr_len + acct_pak->arg_cnt;
    for (i = 0; i < (int)acct_pak->arg_cnt; i++) {
	len += p[i];
    }

    if (len != ntohl(hdr->datalength)) {
	send_error_reply(TAC_PLUS_ACCT, NULL);
	return;
    }

    account(pak);

    free(pak);
}

void
account(pak)
u_char *pak;
{
    struct acct *acct_pak;
    u_char *p, *argsizep;
    struct acct_rec rec;
    struct identity identity;
    char **cmd_argp;
    int i, errors, status;

    acct_pak = (struct acct *) (pak + TAC_PLUS_HDR_SIZE);

    /* Fill out accounting record structure */
    bzero(&rec, sizeof(struct acct_rec));

    if (acct_pak->flags & TAC_PLUS_ACCT_FLAG_WATCHDOG)
	rec.acct_type = ACCT_TYPE_UPDATE;
    if (acct_pak->flags & TAC_PLUS_ACCT_FLAG_START)
	rec.acct_type = ACCT_TYPE_START;
    if (acct_pak->flags & TAC_PLUS_ACCT_FLAG_STOP)
	rec.acct_type = ACCT_TYPE_STOP;

    rec.authen_method  = acct_pak->authen_method;
    rec.authen_type    = acct_pak->authen_type;
    rec.authen_service = acct_pak->authen_service;

    /* start of variable length data is here */
    p = pak + TAC_PLUS_HDR_SIZE + TAC_ACCT_REQ_FIXED_FIELDS_SIZE;

    /* skip arg cnts */
    p += acct_pak->arg_cnt;

    /* zero out identity struct */
    bzero(&identity, sizeof(struct identity));

    identity.username = tac_make_string(p, (int)acct_pak->user_len);
    p += acct_pak->user_len;

    identity.NAS_name = tac_strdup(session.peer);

#ifdef ALIASES
    identity.NAS_alias = tac_strdup(session.peeralias);
    identity.NAS_flags = session.flags;
#else
 #ifdef TTYS
    identity.NAS_flags = B_FLAG_INIT;
 #endif
#endif

    identity.NAS_port = tac_make_string(p, (int)acct_pak->port_len);
    p += acct_pak->port_len;
    if (acct_pak->port_len <= 0) {
	strcpy(session.port, "unknown-port");
    } else {
	strcpy(session.port, identity.NAS_port);
    }

#ifdef TTYS
    identity.NAS_palias = get_port_alias(identity.NAS_port, identity.NAS_name,
#ifdef ALIASES
                                         identity.NAS_alias,
#else
                                         NULL,
#endif
                                         &identity.NAS_flags, &identity.NAS_slot);
#endif

    identity.NAC_address = tac_make_string(p, (int)acct_pak->rem_addr_len);
    p += acct_pak->rem_addr_len;

    identity.priv_lvl = acct_pak->priv_lvl;

    rec.identity = &identity;

    /* Now process cmd args */
    argsizep = pak + TAC_PLUS_HDR_SIZE + TAC_ACCT_REQ_FIXED_FIELDS_SIZE;

    cmd_argp = (char **) tac_malloc(acct_pak->arg_cnt * sizeof(char *));

    for (i = 0; i < (int)acct_pak->arg_cnt; i++) {
	cmd_argp[i] = tac_make_string(p, *argsizep);
	p += *argsizep++;
    }

    rec.args = cmd_argp;
    rec.num_args = acct_pak->arg_cnt;

#ifdef MAXSESS
    /*
     * Tally for MAXSESS counting
     */
    loguser(&rec);
#endif

#ifdef DYNA /* send logout/login notification to dynipd */
    if (identity.username[0])
      switch (rec.acct_type) {
      case ACCT_TYPE_STOP:
	for (i = 0; i < (int)acct_pak->arg_cnt; i++)
	  if (strncmp(cmd_argp[i], "addr=", 5) == 0) { /* FIXME support address= too? */
	    if (debug & DEBUG_ACCT_FLAG)
	      report(LOG_INFO, "sending logout notification for %s (%s)", identity.username, cmd_argp[i] + 5);
	    dyna_req(identity.username, cmd_argp[i] + 5, NULL, 1);
	    break;
	  }
	break;
      case ACCT_TYPE_UPDATE:
      case ACCT_TYPE_START:
	for (i = 0; i < (int)acct_pak->arg_cnt; i++)
	  if (strncmp(cmd_argp[i], "addr=", 5) == 0) { /* FIXME support address= too? */
	    char nasinfo[32];
	    strcpy(nasinfo, inet_ntoa(session.nasaddr));
	    strcat(nasinfo, "/");
	    strncat(nasinfo, identity.NAS_port, 16);
	    if (debug & DEBUG_ACCT_FLAG)
	      report(LOG_INFO, "sending login notification for %s (%s)", identity.username, cmd_argp[i] + 5);
	    dyna_req(identity.username, cmd_argp[i] + 5, nasinfo, 2);
	    break;
	  }
	break;
      }
#endif

    if (identity.username[0]) {
      char *cmd = NULL;
      switch (rec.acct_type) {
      case ACCT_TYPE_START:
	cmd = cfg_get_pvalue(identity.username, TAC_IS_USER, S_acct_start, TAC_PLUS_RECURSE);
	if ((cmd != NULL) && (debug & DEBUG_ACCT_FLAG)) report(LOG_DEBUG, "acct_start call: %s", cmd);
	break;
      case ACCT_TYPE_STOP:
	cmd = cfg_get_pvalue(identity.username, TAC_IS_USER, S_acct_stop, TAC_PLUS_RECURSE);
	if ((cmd != NULL) && (debug & DEBUG_ACCT_FLAG)) report(LOG_DEBUG, "acct_stop call: %s", cmd);
	break;
      case ACCT_TYPE_UPDATE:
	cmd = cfg_get_pvalue(identity.username, TAC_IS_USER, S_acct_update, TAC_PLUS_RECURSE);
	if ((cmd != NULL) && (debug & DEBUG_ACCT_FLAG)) report(LOG_DEBUG, "acct_update call: %s", cmd);
	break;
      }
      if (cmd != NULL) call_extern_acct(cmd, &rec);
    }

    /*
     * Do accounting.
     */
    if (wtmpfile) {
	errors = do_wtmp(&rec);
    } else {
	errors = do_acct(&rec);
    }

    if (errors) {
	status = TAC_PLUS_ACCT_STATUS_ERROR;
    } else {
	status = TAC_PLUS_ACCT_STATUS_SUCCESS;
    }
    send_acct_reply(status, rec.msg, rec.admin_msg);

#ifdef TAC_ROAMING
    {
      char *cfg_passwd;
      char *tmpkey=session.key;
      int tmpsock=session.sock;
      int tmpseq=session.seq_no;
#ifdef PERUSER_ENABLE
      if(rec.authen_service == TAC_PLUS_AUTHEN_SVC_ENUSER)
	cfg_passwd = (char*)cfg_get_enable(identity.username, TAC_PLUS_RECURSE);
      else {
#endif
	if (rec.authen_type == TAC_PLUS_AUTHEN_TYPE_PAP)
	  cfg_passwd = cfg_get_pap_secret(identity.username, TAC_PLUS_RECURSE);
	else
	  cfg_passwd = cfg_get_login_secret(identity.username, TAC_PLUS_RECURSE);
	if (!cfg_passwd)
	  cfg_passwd = cfg_get_global_secret(identity.username, TAC_PLUS_RECURSE);
#ifdef PERUSER_ENABLE
      }
#endif
      if (cfg_passwd && (STREQ("tacacs+", cfg_passwd))) {
	char *roamingaddr, *roamingkey;
	struct sockaddr_in tac_port;
	int rsock;
	u_char *tmppak;

	cfg_get_tac_roaming_data(identity.username, TAC_PLUS_RECURSE, &roamingaddr, &roamingkey);
	tac_port.sin_family = AF_INET;
	tac_port.sin_addr.s_addr = inet_addr(roamingaddr);
	/* FIXME support port specification in roamingaddr: host.name:port */
	/* FIXME support address resolving of roamingaddr via gethostbyname() */
	tac_port.sin_port = htons(49);
	if ((rsock = socket(AF_INET, SOCK_STREAM, 0))<0) {
	  report(LOG_ERR, "roaming: socket(): %s", strerror(errno));
	  goto out;
	}
	if (connect(rsock, (struct sockaddr *) &tac_port, sizeof tac_port) < 0) {
	  report(LOG_ERR, "roaming: unable to connect to tacacs+ server %s: %s", roamingaddr, strerror(errno));
	  goto out;
	}
	session.key=roamingkey;
	session.sock=rsock;
	session.seq_no=1;
	if (write_packet(pak)) {
	  report(LOG_ERR, "roaming: error writing a packet to %s: %s", roamingaddr, strerror(errno));
	  goto out;
	}
	if ((tmppak = read_packet()) == NULL) {
	  report(LOG_ERR, "roaming: error reading a packet from %s: %s", roamingaddr, strerror(errno));
	  goto out;
	}
	report(LOG_INFO, "roaming: accounting passed at %s for %s", roamingaddr, identity.username);
      }
out:  session.key=tmpkey;
      session.sock=tmpsock;
      session.seq_no=tmpseq;
    }
#endif

#ifdef RAD_ROAMING
    {
      char *cfg_passwd;
#ifdef PERUSER_ENABLE
      if(rec.authen_service == TAC_PLUS_AUTHEN_SVC_ENUSER)
	cfg_passwd = (char*)cfg_get_enable(identity.username, TAC_PLUS_RECURSE);
      else {
#endif
	if (rec.authen_type == TAC_PLUS_AUTHEN_TYPE_PAP)
	  cfg_passwd = cfg_get_pap_secret(identity.username, TAC_PLUS_RECURSE);
	else
	  cfg_passwd = cfg_get_login_secret(identity.username, TAC_PLUS_RECURSE);
	if (!cfg_passwd)
	  cfg_passwd = cfg_get_global_secret(identity.username, TAC_PLUS_RECURSE);
#ifdef PERUSER_ENABLE
      }
#endif
      if (cfg_passwd && (STREQ("radius", cfg_passwd))) {
	struct radius *radius_conf;
	struct acctinfo acct;
	int sendit;

	cfg_get_rad_roaming_data(identity.username, TAC_PLUS_RECURSE, &radius_conf);

  /* We have to send a RADIUS accounting packet here. */
  if ((acct_pak->flags & (TAC_PLUS_ACCT_FLAG_START | TAC_PLUS_ACCT_FLAG_STOP | TAC_PLUS_ACCT_FLAG_WATCHDOG)) != 0) {
    memset(&acct, 0, sizeof(acct));
    strncpy(acct.login, identity.username, sizeof(acct.login));
    acct.nas_ip = inet_addr(identity.NAS_name); /* pkt->sin_addr.s_addr */
    acct.nas_port = tac_map(acct.nas_ip, identity.NAS_port);
    if (acct_pak->flags & (TAC_PLUS_ACCT_FLAG_START | TAC_PLUS_ACCT_FLAG_WATCHDOG)) acct.islogin = 1;

    for (i = 0; i < (int)acct_pak->arg_cnt; i++) {
	if (strncmp(cmd_argp[i], "task_id=", 8) == 0)
		sprintf(acct.acctsessionid, "%08ld", atol(cmd_argp[i] + 8));

	if (strncmp(cmd_argp[i], "bytes_in=", 9) == 0)
		acct.bytes_in = atol(cmd_argp[i] + 9);

	if (strncmp(cmd_argp[i], "bytes_out=", 10) == 0)
		acct.bytes_out = atol(cmd_argp[i] + 10);

	if (strncmp(cmd_argp[i], "paks_in=", 8) == 0)
		acct.paks_in = atol(cmd_argp[i] + 8);

	if (strncmp(cmd_argp[i], "paks_out=", 9) == 0)
		acct.paks_out = atol(cmd_argp[i] + 9);

	if (strncmp(cmd_argp[i], "elapsed_time=", 13) == 0)
		acct.elapsed = atol(cmd_argp[i] + 13);

	if (strncmp(cmd_argp[i], "addr=", 5) == 0) /* FIXME support address= too? */
		acct.address = inet_addr(cmd_argp[i] + 5);

	if (strncmp(cmd_argp[i], "service=", 8) == 0) {
		char *sp = cmd_argp[i] + 8;
		if (strcmp(sp, "shell") == 0) {
			/* FIXME: check cmd* for telnet, rlogin */
			acct.proto = P_CONSOLE;
		}
		if (strcmp(sp, "ppp") == 0)
			acct.proto = P_PPP;
	}
    }

    /*
     *	If this is a framed connection, only send the info
     *	if we have the IP address too!
     */
    sendit = 1;
    if (acct.proto == P_PPP || acct.proto == P_SLIP || acct.proto == P_CSLIP)
	if (acct.address == 0) sendit = 0;

    if (sendit) {
	report(LOG_INFO, "Sending accounting info to RADIUS server for %s", identity.username);
	rad_acct(radius_conf, &acct);
    } else
	DEBUG("NOT sending accounting info to RADIUS server (no IPno) for %s", identity.username);
  }}}
#endif

    free(identity.username);
    free(identity.NAS_name);
#ifdef ALIASES
    free(identity.NAS_alias);
#endif
    free(identity.NAS_port);
#ifdef TTYS
    free(identity.NAS_palias);
#endif
    free(identity.NAC_address);

    for (i = 0; i < (int)acct_pak->arg_cnt; i++) {
	free(cmd_argp[i]);
    }    
    free(cmd_argp);
    
    if (rec.msg)
	free(rec.msg);
    if (rec.admin_msg)
	free(rec.admin_msg);
}
