/*
 * advanced tacacs+ server based on Cisco Systems tac_plus code.
 * For Cisco copyright notice, see LICENSE.CISCO
 *
 * This is a copyrighted commercial software; see LICENSE for details.
 */

#define ACL_DENY        0
#define ACL_PERMIT      1

/* Access group - defined per user/group */
struct agroup {
	char           *name;           /* ACL name                     */
	void           *hash;           /* (for hash)                   */
	char           *msg;            /* reason message (if any)      */
	int             defacc;         /* default access               */
	struct aexp    *aexps;          /* List of acl expressions      */
	struct aexp    *atail;          /* (for parser)                 */
};

struct aexp {                           /* Access Expression            */
	int             op;             /* Operation                    */
	char           *agroup;         /* recursive agroup             */
	regex_t        *group;          /* username group regexp        */
	regex_t        *user;           /* username regexp              */
	regex_t        *from;           /* "from" regexp                */
	regex_t        *tty;            /* terminal line regexp         */
#ifdef DXS1
	regex_t        *dxs1;           /* dsx1 regexp                  */
#endif
	struct nasgrp  *nasg;           /* NAS group/list               */
	struct tgrade  *tg;             /* Time grade                   */
	struct aexp    *next;           /* link                         */
};

struct nasgrp {                         /* NAS group definition         */
	char           *name;           /* NAS group name               */
	struct nasl    *naslist;        /* appropriate NAS list         */
	struct nasl    *tail;           /* (for parser)                 */
	struct nasgrp  *next;           /* link                         */
};

struct nasl {                           /* NAS list                     */
	u_long          ipaddr;         /* NAS ip address               */
	u_long          imask;          /* NAS ip address wildcard      */
	struct nasl    *next;           /* next element in list         */
};

struct tgrade {                         /* Timegrade definition         */
	int             wkfrom;         /* Weekday - from               */
	int             wkto;           /* Weekday - until              */
	int             mfrom;          /* mins - from                  */
	int             mto;            /* mins - to                    */
	int             rev;            /* negation                     */
	struct tgrade  *next;           /* next element in list         */
};

struct  nasgrp  *acl_getnas(char *);
struct  tgrade  *acl_parse_tgrade(char *);
int             acl_check(char *, struct identity *, char **);
void            acl_inherit_ag(struct agroup *, struct agroup *);
int             acl_parse_nas(char *, char *);
int             acl_naschk(struct nasgrp *, char *);
void            acl_free_nasgrp();
void            acl_free_agroup(struct agroup *);
void            acl_settime();
int             acl_timeok(struct tgrade *);
#ifdef DXS1
int             acl_dxs1chk();
#endif
