/*
   Copyright (c) 1995-2000 by Cisco systems, Inc.

   Permission to use, copy, modify, and distribute modified and
   unmodified copies of this software for any purpose and without fee is
   hereby granted, provided that (a) this copyright and permission notice
   appear on all copies of the software and supporting documentation, (b)
   the name of Cisco Systems, Inc. not be used in advertising or
   publicity pertaining to distribution of the program without specific
   prior permission, and (c) notice be given in supporting documentation
   that use, modification, copying and distribution is by permission of
   Cisco Systems, Inc.

   Cisco Systems, Inc. makes no representations about the suitability
   of this software for any purpose.  THIS SOFTWARE IS PROVIDED ``AS
   IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
   WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "tac_plus.h"
#include <stdio.h>
#include <errno.h>

#if defined(DENY) || defined(CRON)
 #include <sys/stat.h>
#endif

#if (defined(AUTO_MEMBERS) && !defined(AM_REGEXP)) || \
    (defined(NAMEISREGEXP) && !defined(NAM_REGEXP)) || \
    (defined(CRON) && !defined(CRON_REGEXP)) || defined(UDB)
 #include <fnmatch.h>
#endif

#ifdef CRON
 #define MAIN_PROGRAM
 #include <pwd.h>
 #include "cron.h"
#endif

#include REGEX_H
#ifndef REG_EXTENDED
#define REG_EXTENDED 0
#endif

#if defined(ALIASES) || defined(TTYS)
 #include "own_wtmp.h"
#endif
#include "allow.h"

#ifndef STDIO_SNPRINTF
extern int snprintf(char *, size_t, const char *, ...);
#endif

extern unsigned long bindaddress;

#ifndef INADDR_NONE
#define INADDR_NONE (unsigned long)0xffffffff
#endif

#ifdef RAD_ROAMING
#include "radius.h"
#include "radclient.h" /* struct radius */
#endif

/*
   <config>         := <decl>*

   <decl>           := <top_level_decl> | <user_decl>

   <top_level_decl> := <authen_default> |
                       accounting file = <string>
                       default authorization = permit |
                       key = <string> |
		       <nas_decl> |
                       prompt = <string> |
                       reject = <string> |
                       admdeny = <string> |
                       crondeny = <string> |
		       allow-from = <ipaddr> |
		       listen-on = <ipaddr>

   <authen_default> := default authentication = file <filename>

   <nas_decl>       := nas = <name> {
                         nasaddr = <ipaddr>
			 key = <string>
			 snmp-community = <string> (if ENABLE_SNMP)
			 rsh-user = <string>
		       }

   <permission>     := permit | deny

   <filename>       := <string>

   <password>       := <string>

   <user_decl>      := user = <string> {
                        [ default service = [ permit | deny ] ]
                        <user_attr>*
                        <svc>*
                   }

   <password_spec>  := file <filename> |
		       skey |
		       cleartext <password> |
		       des <password> |
		       extern <path> | (if extern authentication)
		       nopassword |
		       permit |
		       deny |
		       none

   <enable_spec>    := file <filename> |
		       cleartext <password> |
                       des <password> |
		       extern <path> | (if extern authentication)
		       nopassword |
                       permit |
                       deny |
                       none |
                       login

   <user_attr>      :=   name     = <string> |
                         login    = <password_spec> |
                         enable   = <enable_spec> |
        	         users    = <string> |
        	         maxsess  = <string> |
                         member   = <string> |
        	         expires  = <string> |
                	 arap     = cleartext <string> |
	                 chap     = cleartext <string> |
	                 chap     = extern <string> |    (if extern authentication)
#ifdef MSCHAP
	                 ms-chap  = cleartext <string> |
#endif
	                 pap      = cleartext <string> | (if no encrypted-pap)
	                 pap      = des <string> |
	                 pap      = <password_spec> |    (if encrypted-pap)
	                 pap      = extern <string> |    (if extern authentication)
	                 opap     = cleartext <string> |
	                 global   = cleartext <string> |
	                 global   = extern <string> |    (if extern authentication)
        	         message  = <string> |
        	         emessage = <string> |
        	         reject = <string> |
        	         admdeny = <string> |
        	         crondeny = <string> |
			 acct_start = <string> |
			 acct_stop = <string> |
			 acct_update = <string> |
			 before authorization <string> |
			 after authorization <string>

   <svc>            := <svc_auth> | <cmd_auth>

   <flags_spec>     := none |
                       [acc] [wtmp] [passwd] [unique]

   <cmd_auth>       := cmd = <string> {
                        <cmd-match>*
                    }

   <cmd-match>      := <permission> <string>

   <svc_auth>       := service = ( exec | arap | slip | ppp protocol = <string> {
                        [ default attribute = permit ]
                        <attr_value_pair>*
                    }

   <attr_value_pair> := [ optional ] <string> = <string>

*/

static char sym_buf[MAX_INPUT_LINE_LEN];	/* parse buffer */
static int sym_pos=0;           /* current place in sym_buf */
static int sym_ch;		/* current parse character */
static int sym_code;		/* parser output */
static int sym_line = 1;	/* current line number for parsing */
static FILE *cf = NULL;		/* config file pointer */
static int sym_error = 0;	/* a parsing error has occurred */
static int no_user_dflt = 0;	/* default if user doesn't exist */
static char *authen_default = NULL;	/* top level authentication default */

#ifdef UNIX_LOGIN
 static char u_prompt[] = UNIX_PROMPT;
 static char u_reject[] = UNIX_REJECT;
 static char *prompt_default = u_prompt;
 static char *reject_default = u_reject;
#endif

#ifdef CRON
 static char u_cron[] = CRONDENY;
 static char *cron_default = u_cron;
#endif

#ifdef DENY
 static char u_deny[] = ADMDENY;
 static char *deny_default = u_deny;
#endif

/* A host definition structure. Currently unused, but when we start
   configuring host-specific information e.g. per-host keys, this is
   where it should be kept.

   The first 2 fields (name and hash) are used by the hash table
   routines to hash this structure into a table.  Do not (re)move them */

struct host {
    char *name;			/* host name */
    void *hash;			/* hash table next pointer */
    int line;			/* line number defined on */
};

/* A user or group definition

   The first 2 fields (name and hash) are used by the hash table
   routines to hash this structure into a table.  Move them at your
   peril */

struct user {
    char *name;			/* username */
    void *hash;			/* hash table next pointer */
    int line;			/* line number defined on */
    long flags;			/* flags field */

#define FLAG_ISUSER  1		/* this structure represents a user */
#define FLAG_ISGROUP 2		/* this structure represents a group */
#define FLAG_SEEN    4		/* for circular definition detection */

    char *full_name;		/* users full name */
    char *login;		/* Login password */
#ifdef TAC_SMB_AUTH
    char *smbdomain;
    char *smbprimary;
    char *smbbackup;
#endif
#ifdef TAC_LDAP_AUTH
    char *ldapserver;
    char *ldapbasedn;
    char *ldapbinddn;
    char *ldapbindpw;
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_SASL)
    char *ldapsaslmech;
    char *ldapsaslcred;
#endif
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_NS_SSL)
    char *ldapsslpath;
#endif
#endif
#ifdef TAC_ROAMING
    char *traddr;
    char *trkey;
#endif
#ifdef RAD_ROAMING
    struct radius radius_conf;
#endif
#ifdef PERUSER_ENABLE
    char *enable;               /* enable spec */
#endif
    char *global;		/* password to use if none set */
    char *member;		/* group we are a member of */
#ifdef AUTO_MEMBERS
    char *users;		/* users, who are a members of us */
#endif
#ifdef NAMEISREGEXP
 #ifdef NAM_REGEXP
    regex_t *name1;		/* compiled regexp name */
 #else
    char  *name1;		/* regexp name */
 #endif
#endif
#if defined(AUTO_MEMBERS) && defined(AM_REGEXP)
    char *newuser;
#endif
    char *expires;		/* expiration date */
    char *arap;			/* our arap secret */
    char *pap;			/* our pap secret */
    char *opap;			/* our outbound pap secret */
    char *chap;			/* our chap secret */
#ifdef MSCHAP
    char *mschap;		/* our mschap secret */
#endif /* MSCHAP */
    char *msg;			/* a message for this user */
#ifdef EMESSAGE
    char *emsg;			/* a enable message for this user */
#endif
#ifdef UNIX_LOGIN
    char *rmsg;			/* a reject message */
#endif
#ifdef CRON
    char *cdeny;		/* a deny message for cron reason */
#endif
#ifdef DENY
    char *adeny;		/* a deny message for admin reason */
#endif
    char *before_author;	/* command to run before authorization */
    char *after_author;		/* command to run after authorization */
    char *acct_start;		/* command to run for accounting start packet */
    char *acct_stop;		/* command to run for accounting stop packet */
    char *acct_update;		/* command to run for accounting update packet */
    int svc_dflt;		/* default authorization behaviour for svc or
				 * cmd */
    NODE *svcs;			/* pointer to svc nodes */
#ifdef MAXSESS
    int maxsess;		/* Max sessions/user */
#endif /* MAXSESS */
#ifdef UDB
    DB *udb;                    /* subst database handle */
#if defined(AUTO_MEMBERS) && defined(AM_REGEXP) && defined(MORE_SUBST)
    char *db_user;              /* key subst for username */
#endif
#endif
#ifdef ACL
    char *agroup;               /* access-group name */
#endif
#ifdef DYNA
    char *apool;                /* default IP pool name */
#endif
};

typedef struct user USER;

#ifdef UDB
DB     *default_udb = (DB *) 0;     /* database for default auth */
DB     *cfg_get_udb(char *);
char   *cfg_getmember(USER *, char *);
#define GETMEMBER(user, name) (member = cfg_getmember(user, name))
#else
#define GETMEMBER(user, name) (member = user->member)
#endif

/* Only the first 2 fields (name and hash) are used by the hash table
   routines to hashh structures into a table.
*/

union hash {
    struct user u;
    struct host h;
};

typedef union hash HASH;

void *grouptable[HASH_TAB_SIZE];/* Table of group declarations */
void *usertable[HASH_TAB_SIZE];	/* Table of user declarations */

/* void *hosttable[HASH_TAB_SIZE];	Table of host declarations */

#ifdef ACL
void *acltable[HASH_TAB_SIZE];    /* Access-groups definitions */
int  default_enable = S_permit;   /* Does 'enable' permitted by default ? */
#endif

#ifdef AUTO_MEMBERS
static USER *amtable[HASH_TAB_SIZE];	/* Table of pointers to groups */
					/* with automembering */
static int amcounter = 0;		/* counter */
#endif

#ifdef NAMEISREGEXP
static USER *regtable[HASH_TAB_SIZE];	/* Table of pointers to regexp-users */
static int regcounter = 0;		/* counter */
static USER *reguser = NULL;		/* last found user */
static char *regusername = NULL;	/* and his name */
#endif

#if defined (AUTO_MEMBERS)
static USER amuser;			/* tmp am user */
#endif

static void
 sym_get();
static int save_sym2;

naslist *nas_list = NULL; /* (naslist *) 0; */
naslist *nas_new;
char *defaultkey;
static int keyline;

#ifdef __STDC__
#include <stdarg.h>		/* ANSI C, variable length args */
static void
parse_error(char *fmt,...)
#else
#include <varargs.h>		/* has 'vararg' definitions */
/* VARARGS2 */
static void
parse_error(fmt, va_alist)
char *fmt;

va_dcl				/* no terminating semi-colon */
#endif
{
    char msg[256];		/* temporary string */
    va_list ap;

#ifdef __STDC__
    va_start(ap, fmt);
#else
    va_start(ap);
#endif
    vsprintf(msg, fmt, ap);
    va_end(ap);

    report(LOG_ERR, "%s", msg);
    fprintf(stderr, "Error: %s\n", msg);
    tac_exit(1);
}

char *
cfg_nodestring(type)
    int type;
{
    switch (type) {
    default:
	return ("unknown node type");
    case N_arg:
	return ("N_arg");
    case N_optarg:
	return ("N_optarg");
    case N_svc:
	return ("N_svc");
    case N_svc_exec:
	return ("N_svc_exec");
    case N_svc_slip:
	return ("N_svc_slip");
    case N_svc_ppp:
	return ("N_svc_ppp");
    case N_svc_arap:
	return ("N_svc_arap");
    case N_svc_cmd:
	return ("N_svc_cmd");
    case N_permit:
	return ("N_permit");
    case N_deny:
	return ("N_deny");
    }
}

static void
free_attrs(node)
NODE *node;
{
    NODE *next;

    while (node) {
	switch (node->type) {
	case N_optarg:
	case N_arg:
	    if (debug & DEBUG_CLEAN_FLAG)
		report(LOG_DEBUG, "free_cmd_match %s %s",
		       cfg_nodestring(node->type),
		       node->value);
	    break;
	default:
	    report(LOG_ERR, "Illegal node type %s for free_attrs", 
		   cfg_nodestring(node->type));
	    return;
	}

	free(node->value);
	next = node->next;
	free(node);
	node = next;
    }
}

static void
free_cmd_matches(node)
NODE *node;
{
    NODE *next;

    while (node) {
	if (debug & DEBUG_CLEAN_FLAG)
	    report(LOG_DEBUG, "free_cmd_match %s %s",
		   cfg_nodestring(node->type),
		   node->value);

	free(node->value);	/* text */
	free(node->value1);	/* regexp compiled text */
	next = node->next;
	free(node);
	node = next;
    }
}

static void
free_svcs(node)
NODE *node;
{
    NODE *next;

    while (node) {

	switch (node->type) {
	case N_svc_cmd:
	    if (debug & DEBUG_CLEAN_FLAG)
		report(LOG_DEBUG, "free %s %s",
		       cfg_nodestring(node->type), node->value);
	    free(node->value);	/* cmd name */
	    free_cmd_matches(node->value1);
	    next = node->next;
	    free(node);
	    node = next;
	    continue;

	case N_svc:
	case N_svc_ppp:
	    free(node->value1);
	    /* FALL-THROUGH */
	case N_svc_exec:
	case N_svc_arap:
	case N_svc_slip:
	    if (debug & DEBUG_CLEAN_FLAG)
		report(LOG_DEBUG, "free %s", cfg_nodestring(node->type));
	    free_attrs(node->value);
	    next = node->next;
	    free(node);
	    node = next;
	    continue;

	default:
	    report(LOG_ERR, "Illegal node type %d for free_svcs", node->type);
	    return;
	}
    }
}

static void
free_userstruct(user)
USER *user;
{
    if (debug & DEBUG_CLEAN_FLAG)
	report(LOG_DEBUG, "free %s %s",
	       (user->flags & FLAG_ISUSER) ? "user" : "group",
	       user->name);

    if (user->name)
	free(user->name);
    if (user->full_name)
	free(user->full_name);
    if (user->login)
	free(user->login);
#ifdef TAC_SMB_AUTH
    if (user->smbdomain)
	free(user->smbdomain);
    if (user->smbprimary)
	free(user->smbprimary);
    if (user->smbbackup)
	free(user->smbbackup);
#endif
#ifdef TAC_LDAP_AUTH
    if (user->ldapserver)
	free(user->ldapserver);
    if (user->ldapbasedn)
	free(user->ldapbasedn);
    if (user->ldapbinddn)
	free(user->ldapbinddn);
    if (user->ldapbindpw)
	free(user->ldapbindpw);
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_SASL)
    if (user->ldapsaslmech)
	free(user->ldapsaslmech);
    if (user->ldapsaslcred)
	free(user->ldapsaslcred);
#endif
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_NS_SSL)
    if (user->ldapsslpath)
	free(user->ldapsslpath);
#endif
#endif
#ifdef TAC_ROAMING
    if (user->traddr)
        free(user->traddr);
    if (user->trkey)
        free(user->trkey);
#endif
#ifdef PERUSER_ENABLE
    if (user->enable)
      free(user->enable);
#endif
    if (user->member)
	free(user->member);
#ifdef AUTO_MEMBERS
    if (user->users)
	free(user->users);
#endif
#ifdef NAMEISREGEXP
    if (user->name1)
	free(user->name1);
#endif
#if defined(AUTO_MEMBERS) && defined(AM_REGEXP)
    if(user->newuser)
      free(user->newuser);
#endif
    if (user->expires)
	free(user->expires);
    if (user->arap)
	free(user->arap);
    if (user->chap)
	free(user->chap);
#ifdef MSCHAP
    if (user->mschap)
	free(user->mschap);
#endif /* MSCHAP */
    if (user->pap)
	free(user->pap);
    if (user->opap)
	free(user->opap);
    if (user->global)
	free(user->global);
    if (user->msg)
	free(user->msg);
#ifdef EMESSAGE
    if (user->emsg)
	free(user->emsg);
#endif
#ifdef UNIX_LOGIN
    if (user->rmsg)
	free(user->rmsg);
#endif
#ifdef CRON
    if (user->cdeny)
	free(user->cdeny);
#endif
#ifdef DENY
    if (user->adeny)
	free(user->adeny);
#endif
#if defined(UDB) && defined(AUTO_MEMBERS) && defined(AM_REGEXP) && defined(MORE_SUBST)
    if (user->db_user)
	free(user->db_user);
#endif
    if (user->before_author)
	free(user->before_author);
    if (user->after_author)
	free(user->after_author);
    if (user->acct_start)
	free(user->acct_start);
    if (user->acct_stop)
	free(user->acct_stop);
    if (user->acct_update)
	free(user->acct_update);
#ifdef ACL
    if (user->agroup)
	free(user->agroup);
#endif
    free_svcs(user->svcs);
}

#if defined(AUTO_MEMBERS) || defined(NAMEISREGEXP) || defined(DEFUSER_ALWAYS)

static USER *get_hash_user(char *name) {
#if defined(AUTO_MEMBERS) || defined(NAMEISREGEXP)
int i;
#endif
USER *user;

 					/* normal user? 	*/
  if(
     ((user = (USER *) hash_lookup(usertable, name)) != NULL)
#if defined(NAMEISREGEXP) && defined(REGEXPONLY)
     && !user->name1			/* and not regexp user!	*/
#endif
     )
    return user;

#ifdef NAMEISREGEXP
					/* old regexp user? */
 if(regusername && STREQ(regusername, name))
   return reguser;
   					/* new regexp user?	*/
  for(i = 0; i < regcounter; i++)
   {
    if(
#ifdef NAM_REGEXP
       !regexec(regtable[i]->name1, name, 0, NULL, 0)
#else
       !fnmatch(regtable[i]->name1, name, 0)
#endif
      )					/* yes!		*/
     {
      reguser = regtable[i];
      if(regusername)
        free(regusername);
      regusername = tac_strdup(name);
      return reguser;
     }
   }

#endif


#ifdef AUTO_MEMBERS
   					/* good old am user?	*/
 if(amuser.name && STREQ(amuser.name, name))
   return &amuser;
   					/* new am user?		*/
  for(i = 0; i < amcounter; i++)
   {
    if(
#ifdef AM_REGEXP
       !regexec((regex_t *)(amtable[i]->users), name, 0, NULL, 0)
#else
       !fnmatch(amtable[i]->users, name, 0)
#endif
      )					/* yes!		*/
     {
      if(amuser.name)			/* clear old am user	*/
       {
         free_userstruct(&amuser);
         bzero(&amuser, sizeof(amuser));
       }
      					/* fill up		*/
       amuser.name = tac_strdup(name);
       amuser.line = amtable[i]->line;
       amuser.flags |= FLAG_ISUSER;
       amuser.member = tac_strdup(amtable[i]->name);
       return &amuser;
     }
   }

#endif

#ifdef DEFUSER_ALWAYS
  return (USER *) hash_lookup(usertable, DEFAULT_USERNAME);
#else
  return NULL;
#endif
}

#else
 #define get_hash_user(name) ((USER *)hash_lookup(usertable, name))
#endif	/* AUTO_MEMBERS || NAMEISREGEXP */


static void clean_n(string)
 char *string;
{
 char *s = string;
 while((s = strstr(s, "\\n")) != NULL)
  {
   if(s > string && *(s-1) == '\\')
    {
     strcpy(s, s + 1);
     s++;
     continue;
    }
   *(s++) = '\n';
   strcpy(s, s + 1);
  }
}

/*
 * Exported routines
 */

/* Free all allocated structures preparatory to re-reading the config file */
void
cfg_clean_config()
{
    int i;
    USER *entry, *next;

#ifdef UDB
    db_closeall();
    default_udb = (DB *) 0;
#endif

    if (authen_default) {
	free(authen_default);
	authen_default = NULL;
    }

#ifdef UNIX_LOGIN
    if (prompt_default && prompt_default != u_prompt) {
	free(prompt_default);
	prompt_default = u_prompt;
    }

    if (reject_default && reject_default != u_reject) {
	free(reject_default);
	reject_default = u_reject;
    }
#endif

#ifdef CRON
    if (cron_default && cron_default != u_cron) {
	free(cron_default);
	cron_default = u_cron;
    }
#endif

#ifdef DENY
    if (deny_default && deny_default != u_deny) {
	free(deny_default);
	deny_default = u_deny;
    }
#endif

    if (defaultkey) {
	free(defaultkey);
	defaultkey = NULL;
    }

    while (nas_list) {
      nas_new = nas_list->next;
      if (nas_list->key) free(nas_list->key);
#ifdef ENABLE_SNMP
      if (nas_list->snmpcomm) free(nas_list->snmpcomm);
#endif
      free(nas_list);
      nas_list = nas_new;
    }

    if (session.acctfile) {
	free(session.acctfile);
	session.acctfile = NULL;
    }

    /* clean the hosttable -- currently a no-op */

    /* the grouptable */
    for (i = 0; i < HASH_TAB_SIZE; i++) {
	entry = (USER *) grouptable[i];
	while (entry) {
	    next = entry->hash;
	    free_userstruct(entry);
	    free(entry);
	    entry = next;
	}
	grouptable[i] = NULL;
    }

    /* the usertable */
    for (i = 0; i < HASH_TAB_SIZE; i++) {
	entry = (USER *) usertable[i];
	while (entry) {
	    next = entry->hash;
	    free_userstruct(entry);
	    free(entry);
	    entry = next;
	}
	usertable[i] = NULL;
    }

#ifdef ACL
    /* nas grouptable */
    (void) acl_free_nasgrp();

    /* ACL table */
    for (i = 0; i < HASH_TAB_SIZE; i++) {
	entry = (USER *) acltable[i];
	while (entry) {
	    next = entry->hash;
	    acl_free_agroup((struct agroup *)entry);
	    free(entry);
	    entry = next;
	}
	acltable[i] = NULL;
    }
#endif

#ifdef AUTO_MEMBERS
    bzero(amtable, sizeof(amtable));
    amcounter = 0;
#endif

#ifdef NAMEISREGEXP
    bzero(regtable, sizeof(regtable));
    regcounter = 0;
    reguser = NULL;
    if(regusername)
     {
      free(regusername);
      regusername = NULL;
     }
#endif

#if defined (AUTO_MEMBERS)
    free_userstruct(&amuser);
    bzero(&amuser, sizeof(amuser));
#endif
}

static int
parse_permission()
{
    int symbol = sym_code;

    if (sym_code != S_permit && sym_code != S_deny) {
	parse_error("expecting permit or deny but found '%s' on line %d",
		    sym_buf, sym_line);
	return (0);
    }
    sym_get();

    return (symbol);
}

static int
parse(symbol)
int symbol;

{
    if (sym_code != symbol) {
	parse_error("expecting '%s' but found '%s' on line %d",
		    (symbol == S_string ? "string" : codestring(symbol)),
		    sym_buf, sym_line);
	return (1);
    }
    sym_get();
    return (0);
}

static int
parse_opt_svc_default()
{
    if (sym_code != S_default) {
	return (0);
    }

    parse(S_default);
    parse(S_svc);
    parse(S_separator);
    if (sym_code == S_permit) {
	parse(S_permit);
	return (S_permit);
    }
    parse(S_deny);
    return (S_deny);
}

static int
parse_opt_attr_default()
{
    if (sym_code != S_default)
	return (S_deny);

    parse(S_default);
    parse(S_attr);
    parse(S_separator);
    parse(S_permit);
    return (S_permit);
}

/*
static void
parse_host()
{
}
*/

static int parse_user();

#ifdef TAC_SMB_AUTH
static void parse_smb(USER *user) {
    if (save_sym2 == S_login)
	user->login = tac_strdup(sym_buf);
    else
	user->pap = tac_strdup(sym_buf);
    sym_get(); parse(S_openbra);

    while (1) {
	switch (sym_code) {
	case S_closebra:
	    return;
	case S_smbprimary:
	    sym_get();
	    parse(S_separator);
	    user->smbprimary = tac_strdup(sym_buf);
	    sym_get();
	    break;
	case S_smbbackup:
	    sym_get();
	    parse(S_separator);
	    user->smbbackup = tac_strdup(sym_buf);
	    sym_get();
	    break;
        case S_smbdomain:
	    sym_get();
	    parse(S_separator);
	    user->smbdomain = tac_strdup(sym_buf);
	    sym_get();
	    break;
	default:
	    parse_error("parsing smb directive: line %d, %s, %s", sym_line, codestring(sym_code), sym_buf);
	}
    }
}
#endif

#ifdef TAC_LDAP_AUTH
static void parse_ldap(USER *user) {
    if (save_sym2 == S_login)
	user->login = tac_strdup(sym_buf);
    else
	user->pap = tac_strdup(sym_buf);
    sym_get(); parse(S_openbra);

    user->ldapserver = user->ldapbasedn = user->ldapbinddn = user->ldapbindpw
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_SASL)
	= user->ldapsaslmech = user->ldapsaslcred
#endif
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_NS_SSL)
	= user->ldapsslpath
#endif
	= "";

    while (1) {
	switch (sym_code) {
	case S_closebra:
	    return;
	case S_ldapserver:
	    sym_get();
	    parse(S_separator);
	    user->ldapserver = tac_strdup(sym_buf);
	    sym_get();
	    break;
	case S_ldapbasedn:
	    sym_get();
	    parse(S_separator);
	    user->ldapbasedn = tac_strdup(sym_buf);
	    sym_get();
	    break;
	case S_ldapbinddn:
	    sym_get();
	    parse(S_separator);
	    user->ldapbinddn = tac_strdup(sym_buf);
	    sym_get();
	    break;
	case S_ldapbindpw:
	    sym_get();
	    parse(S_separator);
	    user->ldapbindpw = tac_strdup(sym_buf);
	    sym_get();
	    break;
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_SASL)
	case S_ldapsaslmech:
	    sym_get();
	    parse(S_separator);
	    user->ldapsaslmech = tac_strdup(sym_buf);
	    sym_get();
	    break;
	case S_ldapsaslcred:
	    sym_get();
	    parse(S_separator);
	    user->ldapsaslcred = tac_strdup(sym_buf);
	    sym_get();
	    break;
#endif
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_NS_SSL)
	case S_ldapsslpath:
	    sym_get();
	    parse(S_separator);
	    user->ldapsslpath = tac_strdup(sym_buf);
	    sym_get();
	    break;
#endif
	default:
	    parse_error("parsing ldap directive: line %d, %s, %s", sym_line, codestring(sym_code), sym_buf);
	}
    }
}
#endif

#ifdef TAC_ROAMING
static void parse_tac_roaming(USER *user) {
    if (save_sym2 == S_login)
	user->login = tac_strdup(sym_buf);
    else
	user->pap = tac_strdup(sym_buf);
    sym_get(); parse(S_openbra);

    while (1) {
	switch (sym_code) {
	case S_closebra:
	    return;
	case S_traddr:
	    sym_get();
	    parse(S_separator);
	    user->traddr = tac_strdup(sym_buf);
	    sym_get();
	    break;
	case S_key:
	    sym_get();
	    parse(S_separator);
	    user->trkey = tac_strdup(sym_buf);
	    sym_get();
	    break;
	default:
	    parse_error("parsing tacacs+ directive: line %d, %s, %s", sym_line, codestring(sym_code), sym_buf);
	}
    }
}
#endif

#ifdef RAD_ROAMING
static void parse_rad_roaming(USER *user) {
  if (save_sym2 == S_login)
    user->login = tac_strdup(sym_buf);
  else
    user->pap = tac_strdup(sym_buf);
  sym_get(); parse(S_openbra);

  user->radius_conf.authhost1 = user->radius_conf.authhost2
    = user->radius_conf.accthost1 = user->radius_conf.accthost2 = 0; /* radclient.c does not use INADDR_NONE */
  user->radius_conf.secret[0] = 0;

  while (1) {
    switch (sym_code) {
    case S_closebra:
      if (user->radius_conf.authhost1 == 0 /* INADDR_NONE */ ||
	  user->radius_conf.accthost1 == 0 /* INADDR_NONE */ ||
	  user->radius_conf.secret[0] == 0)
	parse_error("radius: missing authost, accthost or secret at line %d", sym_line);
      return;
    case S_authhost1:
      sym_get();
      parse(S_separator);
      user->radius_conf.authhost1 = inet_addr(sym_buf);
      sym_get();
      break;
    case S_authhost2:
      sym_get();
      parse(S_separator);
      user->radius_conf.authhost2 = inet_addr(sym_buf);
      sym_get();
      break;
    case S_accthost1:
      sym_get();
      parse(S_separator);
      user->radius_conf.accthost1 = inet_addr(sym_buf);
      sym_get();
      break;
    case S_accthost2:
      sym_get();
      parse(S_separator);
      user->radius_conf.accthost2 = inet_addr(sym_buf);
      sym_get();
      break;
    case S_secret:
      sym_get();
      parse(S_separator);
      strncpy(user->radius_conf.secret, sym_buf, sizeof(user->radius_conf.secret));
      sym_get();
      break;
    default:
      parse_error("parsing radius directive: line %d, %s, %s", sym_line, codestring(sym_code), sym_buf);
    }
  }
}
#endif

static void
 rch();

/*
   Parse lines in the config file, creating data structures
   Return 1 on error, otherwise 0 */

static int
parse_decls()
{
    no_user_dflt = 0; /* default if user doesn't exist */
    bindaddress = htonl(INADDR_ANY);

    sym_code = 0;
    rch();

    bzero(grouptable, sizeof(grouptable));
    bzero(usertable, sizeof(usertable));
    /* bzero(hosttable, sizeof(hosttable)); */
#ifdef ACL
    bzero(acltable, sizeof(acltable));
#endif

#ifdef AUTO_MEMBERS
    bzero(amtable, sizeof(amtable));
    amcounter = 0;
#endif
#ifdef NAMEISREGEXP
    bzero(regtable, sizeof(regtable));
    regcounter = 0;
    regusername = NULL;
#endif
#if defined (AUTO_MEMBERS)
    bzero(&amuser, sizeof(amuser));
#endif

    sym_get();

    /* Top level of parser */
    while (1) {

	switch (sym_code) {
	case S_eof:
	    return (0);
#ifdef ACL
	case S_agroup:                  /* Access Group Definition */
	    {
		struct  agroup  *ap;
		int save_sym;

		parse(S_agroup);
		parse(S_separator);
		ap = (struct agroup *) tac_malloc(sizeof(*ap));
		ap->name = tac_strdup(sym_buf);
		ap->aexps = ap->atail = (struct aexp *) 0;
		ap->defacc = ACL_DENY;
		ap->msg = NULL;
		sym_get();
		parse(S_openbra);
		while(sym_code != S_closebra) {
			switch (sym_code) {

			    case S_default:
				sym_get();
				parse(S_authentication);
				parse(S_separator);
				if (sym_code == S_deny)
					ap->defacc = ACL_DENY;
				else if (sym_code == S_permit)
					ap->defacc = ACL_PERMIT;
				else
					goto parse1;
				sym_get();
				continue;

			    case S_permit:
			    case S_deny:
				save_sym = sym_code;
				sym_get();
				if (sym_code == S_message) {
				    if (save_sym != S_deny) {
					parse_error("only 'deny message' supported");
					goto parse1;
				    }
				    parse(S_message);
				    parse(S_separator);
				    ap->msg = tac_strdup(sym_buf);
				    sym_get();
				    continue;
				}
				if (acl_parse(save_sym, ap))
					return 1;
				continue;

			    default:
parse1:
				parse_error("syntax error when defining access-group, line %d\n",
					sym_line);
				return (1);

			}
		}
		parse(S_closebra);
		if (hash_add_entry(acltable, ap)) {
			parse_error("redefining access-group %s, line %d\n",
				ap->name, sym_line);
			return (1);
		}
		continue;
	    }

	case S_nasgrp:                  /* define NAS group */
	    {
		char    *nname;

		parse(S_nasgrp);
		parse(S_separator);
		nname = tac_strdup(sym_buf);
		sym_get();
		parse(S_openbra);
		while (sym_code != S_closebra) {
			if (acl_parse_nas(nname, sym_buf)) {
				parse_error("bad NAS spec on line %d\n",
					sym_line);
				return (1);
			}
			sym_get();
		}
		parse(S_closebra);
		(void) free(nname);
		continue;
	    }
#endif
#ifdef DYNA
        case S_dynip:
            sym_get();
            parse(S_separator);
            dyna_sethost(sym_buf);
            sym_get();
            continue;
#endif
	case S_accounting:
	    sym_get();
	    parse(S_file);
	    parse(S_separator);
	    if (session.acctfile) 
		free(session.acctfile);
	    session.acctfile = tac_strdup(sym_buf);
	    sym_get();
	    continue;

	case S_allow_from:
	    sym_get();
	    parse(S_separator);
	    if (put_new_allow_addr(sym_buf))
	      parse_error("bad ip address %s\n", sym_buf);
	    sym_get();
	    continue;

	case S_listen_on:
	    sym_get();
	    parse(S_separator);
	    bindaddress = inet_addr(sym_buf);
	    if (bindaddress == INADDR_NONE)
	      parse_error("bad ip address %s\n", sym_buf);
	    sym_get();
	    continue;

#ifdef UNIX_LOGIN
	case S_prompt:
	    sym_get();
	    parse(S_separator);
	    if (prompt_default && prompt_default != u_prompt)
	       free(prompt_default);
	    prompt_default = tac_strdup(sym_buf);
	    clean_n(prompt_default);
	    sym_get();
	    continue;

	case S_reject:
	    sym_get();
	    parse(S_separator);
	    if (reject_default && reject_default != u_reject)
	       free(reject_default);
	    reject_default = tac_strdup(sym_buf);
	    clean_n(reject_default);
	    sym_get();
	    continue;
#endif

#ifdef CRON
	case S_crondeny:
	    sym_get();
	    parse(S_separator);
	    if (cron_default && cron_default != u_cron)
	       free(cron_default);
	    cron_default = tac_strdup(sym_buf);
	    clean_n(cron_default);
	    sym_get();
	    continue;
#endif

#ifdef DENY
	case S_admdeny:
	    sym_get();
	    parse(S_separator);
	    if (deny_default && deny_default != u_deny)
	       free(deny_default);
	    deny_default = tac_strdup(sym_buf);
	    clean_n(deny_default);
	    sym_get();
	    continue;
#endif

	case S_default:
	    sym_get();
	    switch (sym_code) {
	    default:
		parse_error(
	        "Expecting default authorization/authentication on lines %d",
			    sym_line);
		return (1);

	    case S_authentication:
		if (authen_default) {
		    parse_error(
		    "Multiply defined authentication default on line %d",
				sym_line);
		    return (1);
		}
		parse(S_authentication);
		parse(S_separator);
#ifdef UDB
		if (sym_code == S_file) {
			parse(S_file);
			authen_default = tac_strdup(sym_buf);
		} else if (sym_code == S_db) {
			parse(S_db);
			authen_default = tac_strdup("db");
			default_udb    = (DB *) db_register(sym_buf);
			if (!default_udb)
				return(1);
		} else {
			parse_error("should be (db|file) filename\n");
			return(1);
		}
#else
		parse(S_file);
		authen_default = tac_strdup(sym_buf);
#endif
		sym_get();
		continue;

	    case S_authorization:
		parse(S_authorization);
		parse(S_separator);
		parse(S_permit);
		no_user_dflt = S_permit;
		report(LOG_INFO, 
		       "default authorization = permit is now deprecated. Please use user = DEFAULT instead");
		continue;
	    }

	case S_key:
	    /* Process a key declaration. */
	    sym_get();
	    parse(S_separator);
	    if (defaultkey) {
		parse_error("multiply defined key on lines %d and %d",
			    keyline, sym_line);
		return (1);
	    }
	    defaultkey = tac_strdup(sym_buf);
	    keyline = sym_line;
	    sym_get();
	    continue;

	case S_nas:
	  sym_get();
	  parse(S_separator);
	  /* parse(S_string); */
	  sym_get(); /* nas_name -- ignored */
	  parse(S_openbra);
	  nas_new = tac_malloc(sizeof(naslist));
	  nas_new->next = nas_list;
	  nas_list = nas_new;
	  nas_new->key = NULL;
	  nas_new->rshuser = NULL;
#ifdef ENABLE_SNMP
	  nas_new->snmpcomm = NULL;
#endif
	  while (1) {
	    switch (sym_code) {
	    case S_closebra:
	      sym_get();
	      goto nas_out;
	    case S_nasaddr:
	      sym_get();
	      parse(S_separator);
	      if (!(nas_new->addr.s_addr = inet_addr(sym_buf)))
		parse_error("bad ip address %s\n", sym_buf);
	      sym_get();
	      break;
	    case S_key:
	      sym_get();
	      parse(S_separator);
	      nas_new->key = tac_strdup(sym_buf);
	      sym_get();
	      break;
	    case S_rshuser:
	      sym_get();
	      parse(S_separator);
	      nas_new->rshuser = tac_strdup(sym_buf);
	      sym_get();
	      break;
#ifdef ENABLE_SNMP
	    case S_snmpcomm:
	      sym_get();
	      parse(S_separator);
	      nas_new->snmpcomm = tac_strdup(sym_buf);
	      sym_get();
	      break;
#endif
	    default:
	      parse_error("parsing nas directive: line %d, %s, %s", sym_line, codestring(sym_code), sym_buf);
	    }
	  }
nas_out:  continue;

	case S_user:
	case S_group:
	    parse_user();
	    continue;

	    /* case S_host: parse_host(); continue; */

	default:
	    parse_error("Unrecognised token %s on line %d", sym_buf, sym_line);
	    return (1);
	}
    }
}

static NODE *parse_svcs();

/* Assign a value to a field. Issue an error message and return 1 if
   it's already been assigned. This is a macro because I was sick of
   repeating the same code fragment over and over */

#define ASSIGN(field) \
sym_get(); parse(S_separator); if (field) { \
	parse_error("Duplicate value for %s %s and %s on line %d", \
		    codestring(sym_code), field, sym_buf, sym_line); \
        tac_exit(1); \
    } \
    field = tac_strdup(sym_buf);

static int
parse_user()
{
    USER *n;
    int isuser;
    USER *user = (USER *) tac_malloc(sizeof(USER));
    int save_sym;
    char **fieldp;
    char buf[MAX_INPUT_LINE_LEN];

    bzero(user, sizeof(USER));

    isuser = (sym_code == S_user);

    sym_get();
    parse(S_separator);
    user->name = tac_strdup(sym_buf);

#ifdef NAMEISREGEXP

#ifdef NAM_REGEXP
  #define        META    "^$.[()|?+*\\"
#else
  #define        META    "!*?["
#endif

    if(isuser && strpbrk(user->name, META))
     {
      if (regcounter >= sizeof(regtable)/sizeof(regtable[0]))
         parse_error("regexp users limit reached on line %d",
			 sym_line);

#ifdef NAM_REGEXP
      user->name1 = (void *)tac_malloc(sizeof(regex_t));
      if(regcomp(user->name1, user->name, REG_EXTENDED))
        parse_error("bad regular expression %s on line %d",
                    sym_buf, sym_line);
#else
      user->name1 = tac_strdup(user->name);
#endif

      regtable[regcounter++] = user;
     }
#endif /* NAMEISREGEXP */

    user->line = sym_line;

    if (isuser) {
	user->flags |= FLAG_ISUSER;
	n = hash_add_entry(usertable, (void *) user);
    } else {
	user->flags |= FLAG_ISGROUP;
	n = hash_add_entry(grouptable, (void *) user);
    }

    if (n) {
	parse_error("multiply defined %s %s on lines %d and %d",
		    isuser ? "user" : "group",
		    user->name, n->line, sym_line);
	return (1);
    }
    sym_get();
    parse(S_openbra);

    /* Is the default deny for svcs or cmds to be overridden? */
    user->svc_dflt = parse_opt_svc_default();

    while (1) {
	switch (sym_code) {
	case S_eof:
	    return (0);
#ifdef ACL
	case S_agroup:
	    ASSIGN(user->agroup);
	    sym_get();
	    continue;
#endif
#ifdef UDB
	case S_db:
	    sym_get();
	    parse(S_separator);
	    user->udb = (DB *) db_register(sym_buf);
	    sym_get();
	    continue;
#if defined(AUTO_MEMBERS) && defined(AM_REGEXP) && defined(MORE_SUBST)
	case S_db_user:
	    sym_get();
	    parse(S_separator);
	    if (user->db_user)
	      parse_error("Duplicate value for %s on line %d", codestring(sym_code), sym_line);
	    user->db_user = tac_strdup(sym_buf);
	    sym_get();
	    continue;
#endif
#endif

#ifdef DYNA
        case S_apool:
            ASSIGN(user->apool);
            sym_get();
            continue;
#endif /* DYNA */
	case S_before:
	    sym_get();
	    parse(S_authorization);
	    if (user->before_author)
		free(user->before_author);
	    user->before_author = tac_strdup(sym_buf);
	    sym_get();
	    continue;

	case S_after:
	    sym_get();
	    parse(S_authorization);
	    if (user->after_author)
		free(user->after_author);
	    user->after_author = tac_strdup(sym_buf);
	    sym_get();
	    continue;

	case S_acct_start:
	  if (user->acct_start)
	    parse_error("Duplicate value for %s on line %d", codestring(sym_code), sym_line);
	  sym_get();
	  user->acct_start = tac_strdup(sym_buf);
	  sym_get();
	  continue;

	case S_acct_stop:
	  if (user->acct_stop)
	    parse_error("Duplicate value for %s on line %d", codestring(sym_code), sym_line);
	  sym_get();
	  user->acct_stop = tac_strdup(sym_buf);
	  sym_get();
	  continue;

	case S_acct_update:
	  if (user->acct_update)
	    parse_error("Duplicate value for %s on line %d", codestring(sym_code), sym_line);
	  sym_get();
	  user->acct_update = tac_strdup(sym_buf);
	  sym_get();
	  continue;

	case S_svc:
	case S_cmd:
	    
	    if (user->svcs) {
		/* 
		 * Already parsed some services/commands. Thanks to Gabor Kiss
		 * who found this bug.
		 */
		NODE *p;
		for (p=user->svcs; p->next; p=p->next) 
		    /* NULL STMT */;
		p->next = parse_svcs();
	    } else {
		user->svcs = parse_svcs();
	    }
	    continue;

	case S_pap:
	case S_login:
	    if ((user->login) && (sym_code == S_login))
		parse_error("Duplicate value for %s %s and %s on line %d",
			    codestring(sym_code), user->login,
			    sym_buf, sym_line);
	    if ((user->pap) && (sym_code == S_pap))
		parse_error("Duplicate value for %s %s and %s on line %d",
			    codestring(sym_code), user->pap,
			    sym_buf, sym_line);
	    save_sym2 = sym_code;
	    sym_get();
	    parse(S_separator);
	    switch(sym_code) {
#ifdef NONE_LOGIN
	    case S_none:
	    case S_permit:
	    case S_nopasswd:
	    case S_deny:
#endif

	    case S_skey:
		if (save_sym2 == S_login)
		    user->login = tac_strdup(sym_buf);
		else
		    user->pap = tac_strdup(sym_buf);
		break;

	    case S_file:
	    case S_cleartext:
	    case S_des:
#ifdef EXTERN_AUTHEN
	    case S_extern:
#endif
		sprintf(buf, "%s ", sym_buf);
#ifdef UDB
		/* fall thru */
	    case S_db:
		if (sym_code == S_db) {
			sym_get();
			if (save_sym2 == S_login)
			  user->login = tac_strdup("db ");
			else
			  user->pap = tac_strdup("db ");
			user->udb = (DB *) db_register(sym_buf);
			break;
		}
#endif
		sym_get();
		strcat(buf, sym_buf);
		if (save_sym2 == S_login)
		    user->login = tac_strdup(buf);
		else
		    user->pap = tac_strdup(buf);
		break;

#ifdef TAC_SMB_AUTH
	    case S_smb:
		parse_smb(user);
		break;
#endif

#ifdef TAC_LDAP_AUTH
	    case S_ldap:
		parse_ldap(user);
		break;
#endif

#ifdef TAC_DCE_AUTH
	    case S_dce:
		if (save_sym2 == S_login)
		    user->login = tac_strdup(sym_buf);
		else
		    user->pap = tac_strdup(sym_buf);
		break;
#endif

#ifdef TAC_ROAMING
	    case S_tacacs:
		parse_tac_roaming(user);
		break;
#endif

#ifdef RAD_ROAMING
	    case S_radius:
		parse_rad_roaming(user);
		break;
#endif

	    default:
		parse_error("expecting 'file',"
#ifdef NONE_LOGIN
 " 'none', 'permit', 'nopassword', 'deny',"
#endif
 " 'skey', 'cleartext' or 'des' keyword after '(login|pap) =' on line %d",
			    sym_line);
	    }
	    sym_get();
	    continue;

#ifdef PERUSER_ENABLE
	case S_enable:
	    if (user->enable)
	      parse_error("Duplicate value for %s %s and %s on line %d",
			  codestring(sym_code), user->enable,
			  sym_buf, sym_line);
	    sym_get();
	    parse(S_separator);
	    switch(sym_code) {

	    case S_none:
	    case S_permit:
	    case S_nopasswd:
	    case S_deny:
	    case S_login:
		user->enable = tac_strdup(sym_buf);
		break;

	    case S_file:
	    case S_cleartext:
	    case S_des:
#ifdef EXTERN_AUTHEN
	    case S_extern:
#endif
		sprintf(buf, "%s ", sym_buf);
		sym_get();
		strcat(buf, sym_buf);
		user->enable = tac_strdup(buf);
		break;

	    default:
		parse_error(
 "expecting 'none', 'permit', 'nopassword', 'deny', 'login', 'file', 'cleartext' or 'des' keyword after 'enable =' on line %d",
			    sym_line);
	    }
	    sym_get();
	    continue;
#endif

	case S_name:
	    ASSIGN(user->full_name);
	    sym_get();
	    continue;

	case S_member:
	    ASSIGN(user->member);
	    sym_get();
	    continue;

#ifdef AUTO_MEMBERS
	case S_users:
            if (isuser)
	      parse_error("Unrecognised keyword %s for user on line %d",
			  sym_buf, sym_line);
            if (amcounter >= sizeof(amtable)/sizeof(amtable[0]))
	      parse_error("Auto members limit reached on line %d",
			  sym_line);
	    ASSIGN(user->users);
#ifdef AM_REGEXP
            {
             regex_t *s=tac_malloc(sizeof(regex_t));
             if(regcomp(s, user->users, REG_EXTENDED))
               parse_error("bad regular expression %s on line %d",
			   sym_buf, sym_line);
             free(user->users);
             user->users = (char *)s;
            }
#endif
	    sym_get();
	    amtable[amcounter++] = user;
	    continue;
#endif

#if defined(AUTO_MEMBERS) && defined(AM_REGEXP)
	case S_newuser:
	  ASSIGN(user->newuser);
	  sym_get();
	  continue;
#endif

	case S_expires:
	    ASSIGN(user->expires);
	    sym_get();
	    continue;

	case S_message:
	    ASSIGN(user->msg);
	    clean_n(user->msg);
	    sym_get();
	    continue;

#ifdef EMESSAGE
	case S_emessage:
	    ASSIGN(user->emsg);
	    clean_n(user->emsg);
	    sym_get();
	    continue;
#endif

#ifdef UNIX_LOGIN
	case S_reject:
	    ASSIGN(user->rmsg);
	    clean_n(user->rmsg);
	    sym_get();
	    continue;
#endif

#ifdef CRON
	case S_crondeny:
	    ASSIGN(user->cdeny);
	    clean_n(user->cdeny);
	    sym_get();
	    continue;
#endif

#ifdef DENY
	case S_admdeny:
	    ASSIGN(user->adeny);
	    clean_n(user->adeny);
	    sym_get();
	    continue;
#endif

	case S_arap:
	case S_chap:
#ifdef MSCHAP
	case S_mschap:
#endif /* MSCHAP */
	case S_opap:
	case S_global:
	    save_sym = sym_code;
	    sym_get();
	    parse(S_separator);
	    switch (sym_code) {
	    case S_cleartext:
	    case S_file:
#ifdef UDB
	    case S_db:
#endif
#ifdef EXTERN_AUTHEN
	    case S_extern:
#endif
		sprintf(buf, "%s ", sym_buf);
		sym_get();
		strcat(buf, sym_buf);
		break;
#ifdef NONE_LOGIN
	    case S_none:
	    case S_permit:
	    case S_nopasswd:
	    case S_deny:
		sprintf(buf,"%s",sym_buf);
		break;
#endif
	    default:
		parse_error(
#ifdef EXTEN_AUTHEN
 "expecting 'cleartext' or 'extern' keyword after 'pap =' on line %d",
#else
 "expecting 'cleartext' keyword after 'pap =' on line %d",
#endif
			    sym_line);
	    }

	    if (save_sym == S_arap)
		fieldp = &user->arap;
	    if (save_sym == S_chap)
		fieldp = &user->chap;
#ifdef MSCHAP
	    if (save_sym == S_mschap)
		fieldp = &user->mschap;
#endif /* MSCHAP */
	    if (save_sym == S_opap)
		fieldp = &user->opap;
	    if (save_sym == S_global)
		fieldp = &user->global;

	    if (*fieldp) {
		parse_error("Duplicate value for %s %s and %s on line %d",
			    codestring(save_sym), *fieldp, sym_buf, sym_line);
		tac_exit(1);
	    }
	    *fieldp = tac_strdup(buf);
	    sym_get();
	    continue;

	case S_closebra:
	    parse(S_closebra);
	    return (0);

#ifdef MAXSESS
	case S_maxsess:
	    sym_get(); 
	    parse(S_separator);
	    if (sscanf(sym_buf, "%d", &user->maxsess) != 1) {
		parse_error("expecting integer, found '%s' on line %d",
		    sym_buf, sym_line);
	    }
	    sym_get();
	    continue;
#endif /* MAXSESS */

	default:
	    if (STREQ(sym_buf, "password")) {
		fprintf(stderr,
			"\npassword = <string> is obsolete. Use login = des <string>\n");
	    }
	    parse_error("Unrecognised keyword %s for user on line %d",
			sym_buf, sym_line);

	    return (0);
	}
    }
}

static NODE *parse_attrs();
static NODE *parse_cmd_matches();

static NODE *
parse_svcs()
{
    NODE *result;

    switch (sym_code) {
    default:
	return (NULL);
    case S_svc:
    case S_cmd:
	break;
    }

    result = (NODE *) tac_malloc(sizeof(NODE));

    bzero(result, sizeof(NODE));
    result->line = sym_line;

    /* cmd declaration */
    if (sym_code == S_cmd) {
	parse(S_cmd);
	parse(S_separator);
	result->value = tac_strdup(sym_buf);

	sym_get();
	parse(S_openbra);

	result->value1 = parse_cmd_matches();
	result->type = N_svc_cmd;

	parse(S_closebra);
	result->next = parse_svcs();
	return (result);
    }

    /* svc declaration */
    parse(S_svc);
    parse(S_separator);
    switch (sym_code) {
    default:
	parse_error("expecting service type but found %s on line %d",
		    sym_buf, sym_line);
	return (NULL);

    case S_string:
	result->type = N_svc;
	/* should perhaps check that this is an allowable service name */
	result->value1 = tac_strdup(sym_buf);
	break;
    case S_exec:
	result->type = N_svc_exec;
	break;
    case S_arap:
	result->type = N_svc_arap;
	break;
    case S_slip:
	result->type = N_svc_slip;
	break;
    case S_ppp:
	result->type = N_svc_ppp;
	parse(S_ppp);
	parse(S_protocol);
	parse(S_separator);
	/* Should perhaps check that this is a known PPP protocol name */
	result->value1 = tac_strdup(sym_buf);
	break;
    }
    sym_get();
    parse(S_openbra);
    result->dflt = parse_opt_attr_default();
    result->value = parse_attrs();
    parse(S_closebra);
    result->next = parse_svcs();
    return (result);
}

/*  <cmd-match>	 := <permission> <string> */

static NODE *
parse_cmd_matches()
{
    NODE *result;

    if (sym_code != S_permit && sym_code != S_deny) {
	return (NULL);
    }
    result = (NODE *) tac_malloc(sizeof(NODE));

    bzero(result, sizeof(NODE));
    result->line = sym_line;

    result->type = (parse_permission() == S_permit) ? N_permit : N_deny;
    result->value = tac_strdup(sym_buf);

    result->value1 = tac_malloc(sizeof(regex_t));
    if (regcomp((void *)result->value1, result->value, REG_EXTENDED)) {
	report(LOG_ERR, "in regular expression %s on line %d",
	       sym_buf, sym_line);
	tac_exit(1);
    }
    sym_get();

    result->next = parse_cmd_matches();

    return (result);
}

static NODE *
parse_attrs()
{
    NODE *result;
    char buf[MAX_INPUT_LINE_LEN];
    int optional = 0;

    if (sym_code == S_closebra) {
	return (NULL);
    }
    result = (NODE *) tac_malloc(sizeof(NODE));

    bzero(result, sizeof(NODE));
    result->line = sym_line;

    if (sym_code == S_optional) {
	optional++;
	sym_get();
    }
    result->type = optional ? N_optarg : N_arg;

    strcpy(buf, sym_buf);
    parse(S_string);
#ifdef UDB
    if (*buf != '%') {
#endif
    strcat(buf, sym_buf);
    parse(S_separator);
    strcat(buf, sym_buf);
    parse(S_string);
#ifdef UDB
    }
#endif

    result->value = tac_strdup(buf);
    result->next = parse_attrs();
    return (result);
}


static void
 getsym();

static void
sym_get()
{
    getsym();

    if (debug & DEBUG_PARSE_FLAG) {
	report(LOG_DEBUG, "line=%d sym=%s code=%d buf='%s'",
	       sym_line, codestring(sym_code), sym_code, sym_buf);
    }
}

static char *
sym_buf_add(c)
char c;
{
    if (sym_pos >= MAX_INPUT_LINE_LEN) {
	sym_buf[MAX_INPUT_LINE_LEN-1] = '\0';
	if (debug & DEBUG_PARSE_FLAG) {
	    report(LOG_DEBUG, "line too long: line=%d sym=%s code=%d buf='%s'",
		   sym_line, codestring(sym_code), sym_code, sym_buf);
	}
	return(NULL);
    }

    sym_buf[sym_pos++] = c;
    return(sym_buf);
}
    
static void
getsym()
{

next:
    switch (sym_ch) {

    case EOF:
	sym_code = S_eof;
	return;

    case '\n':
	sym_line++;
	rch();
	goto next;

    case '\t':
    case ' ':
	while (sym_ch == ' ' || sym_ch == '\t')
	    rch();
	goto next;

    case '=':
	strcpy(sym_buf, "=");
	sym_code = S_separator;
	rch();
	return;

    case '{':
	strcpy(sym_buf, "{");
	sym_code = S_openbra;
	rch();
	return;

    case '}':
	strcpy(sym_buf, "}");
	sym_code = S_closebra;
	rch();
	return;

    case '#':
	while ((sym_ch != '\n') && (sym_ch != EOF))
	    rch();
	goto next;

    case '"':
	rch();
	sym_pos = 0;
	while (1) {

	    if (sym_ch == '"') {
		break;
	    }

	    /* backslash-double-quote is supported inside strings */
	    /* also allow \<char> */
	    if (sym_ch == '\\') {
		rch();
		switch (sym_ch) {
		default:
		    /* preserve the slash for \<char> */
		    if (!sym_buf_add('\\')) {
			sym_code = S_unknown;
			rch();
			return;
		    }
		    
		    /* fall through */
		case '"':
		    if (!sym_buf_add(sym_ch)) {
			sym_code = S_unknown;
			rch();
			return;
		    }
		    rch();
		    continue;
		}
	    }
	    if (!sym_buf_add(sym_ch)) {
		sym_code = S_unknown;
		rch();
		return;
	    }
	    rch();
	}
	rch();

	if (!sym_buf_add('\0')) {
	    sym_code = S_unknown;
	    rch();
	    return;
	}
	sym_code = S_string;
	return;

    default:
	sym_pos = 0;
	while (sym_ch != '\t' && sym_ch != ' ' && sym_ch != '='
	       && sym_ch != '\n') {

	    if (!sym_buf_add(sym_ch)) {
		sym_code = S_unknown;
		rch();
		return;
	    }
	    rch();
	}

	if (!sym_buf_add('\0')) {
	    sym_code = S_unknown;
	    rch();
	    return;
	}
	sym_code = keycode(sym_buf);
	if (sym_code == S_unknown)
	    sym_code = S_string;
	return;
    }
}

static void
rch()
{
    if (sym_error) {
	sym_ch = EOF;
	return;
    }
    sym_ch = getc(cf);

    if (parse_only && sym_ch != EOF)
	fprintf(stderr, "%c", sym_ch);
}


/* For a user or group, find the value of a field. Does not recurse. */
VALUE
get_value(user, field)
USER *user;
int field;
{
    VALUE v;

    v.intval = 0;

    if (!user) {
	parse_error("get_value: illegal user");
	return (v);
    }
    switch (field) {

    case S_name:
	v.pval = user->name;
	break;

    case S_login:
	v.pval = user->login;
	break;

#ifdef PERUSER_ENABLE
    case S_enable:
      v.pval = user->enable;
      break;
#endif

    case S_global:
	v.pval = user->global;
	break;

    case S_member:
	v.pval = user->member;
	break;

#ifdef AUTO_MEMBERS
    case S_users:
	v.pval = user->users;
	break;
#endif

#if defined(AUTO_MEMBERS) && defined(AM_REGEXP)
    case S_newuser:
      v.pval = user->newuser;
      break;
#endif

    case S_expires:
	v.pval = user->expires;
	break;

    case S_arap:
	v.pval = user->arap;
	break;

    case S_chap:
	v.pval = user->chap;
	break;

#ifdef MSCHAP
    case S_mschap:
	v.pval = user->mschap;
	break;
#endif /* MSCHAP */

    case S_pap:
	v.pval = user->pap;
	break;

    case S_opap:
	v.pval = user->opap;
	break;

    case S_message:
	v.pval = user->msg;
	break;

#ifdef EMESSAGE
    case S_emessage:
	v.pval = user->emsg;
	break;
#endif

#ifdef UNIX_LOGIN
    case S_reject:
	v.pval = user->rmsg;
	break;
#endif

#ifdef CRON
    case S_crondeny:
	v.pval = user->cdeny;
	break;
#endif

#ifdef DENY
    case S_admdeny:
	v.pval = user->adeny;
	break;
#endif

    case S_svc:
	v.pval = user->svcs;
	break;

    case S_before:
	v.pval = user->before_author;
	break;

    case S_after:
	v.pval = user->after_author;
	break;

    case S_acct_start:
	v.pval = user->acct_start;
	break;

    case S_acct_stop:
	v.pval = user->acct_stop;
	break;

    case S_acct_update:
	v.pval = user->acct_update;
	break;

    case S_svc_dflt:
	v.intval = user->svc_dflt;
	break;

#ifdef MAXSESS
    case S_maxsess:
	v.intval = user->maxsess;
	break;
#endif 

#ifdef UDB
    case S_db:
	v.pval = user->udb;
	break;
#if defined(AUTO_MEMBERS) && defined(AM_REGEXP) && defined(MORE_SUBST)
    case S_db_user:
	v.pval = user->db_user;
	break;
#endif
#endif

#ifdef ACL
    case S_agroup:
	v.pval = user->agroup;
	break;
#endif

#ifdef DYNA
    case S_apool:
        v.pval = user->apool;
        break;
#endif

#ifdef TAC_SMB_AUTH
    case S_smbprimary:
	v.pval = user->smbprimary;
	break;
    case S_smbbackup:
	v.pval = user->smbbackup;
	break;
    case S_smbdomain:
	v.pval = user->smbdomain;
	break;
#endif

#ifdef TAC_LDAP_AUTH
    case S_ldapserver:
	v.pval = user->ldapserver;
	break;
    case S_ldapbasedn:
	v.pval = user->ldapbasedn;
	break;
    case S_ldapbinddn:
	v.pval = user->ldapbinddn;
	break;
    case S_ldapbindpw:
	v.pval = user->ldapbindpw;
	break;
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_SASL)
    case S_ldapsaslmech:
	v.pval = user->ldapsaslmech;
	break;
    case S_ldapsaslcred:
	v.pval = user->ldapsaslcred;
	break;
#endif
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_NS_SSL)
    case S_ldapsslpath:
	v.pval = user->ldapsslpath;
	break;
#endif
#endif

#ifdef TAC_ROAMING
    case S_traddr:
	v.pval = user->traddr;
	break;
    case S_key:
	v.pval = user->trkey;
	break;
#endif

#ifdef RAD_ROAMING
    case S_radius:
	if (user->radius_conf.secret[0])
	  v.pval = &(user->radius_conf);
	else
	  v.pval = NULL;
	break;
#endif
	
    default:
	report(LOG_ERR, "get_value: unknown field %d", field);
	break;
    }
    return (v);
}

/* For each user, check she doesn't circularly reference a
   group. Return 1 if it does */
static int
circularity_check()
{
    USER *user, *entry, *group;
    USER **users = (USER **) hash_get_entries(usertable);
    USER **groups = (USER **) hash_get_entries(grouptable);
    USER **p, **q;
    char *groupname;

#if defined(AUTO_MEMBERS)
    int i;
#endif

    /* users */
    for (p = users; *p; p++) {
	user = *p;

	if (debug & DEBUG_PARSE_FLAG)
	    report(LOG_DEBUG, "circularity_check: user=%s", user->name);

	/* Initialise all groups "seen" flags to zero */
	for (q = groups; *q; q++) {
	    group = *q;
	    group->flags &= ~FLAG_SEEN;
	}

	entry = user;

	while (entry) {
	    /* check groups we are a member of */
	    groupname = entry->member;

	    if (debug & DEBUG_PARSE_FLAG)
		report(LOG_DEBUG, "\tmember of group %s",
		       groupname ? groupname : "<none>");

#if defined(AUTO_MEMBERS) && defined (AM_INFILE)
		/* no member for user defined, check auto */
            if(!groupname && (entry == *p)
#if defined(NAMEISREGEXP) && defined(REGEXPONLY)
               && !entry->name1
#endif
              )
             {
              for (i = 0; i < amcounter; i++) {
                user = amtable[i];

	        if (debug & DEBUG_PARSE_FLAG)
	          report(LOG_DEBUG, "auto membering check: user=%s group=%s",
	                             entry->name, user->name);

                if(
#ifdef AM_REGEXP
                   !regexec(user->users, entry->name, 0, NULL, 0)
#else
                   !fnmatch(user->users, entry->name, 0)
#endif
                   )		/* found as auto member		*/
                 {
	          if (debug & DEBUG_PARSE_FLAG)
	            report(LOG_DEBUG, "auto membering found: user=%s group=%s",
	                             entry->name, user->name);
	          entry->member = tac_strdup(user->name);
	          break;
                 }

              }
              /* no groupname reinit - all am groups will checked later */
             }
#endif

	    /* if not a member of any groups, go on to next user */
	    if (!groupname
#ifdef UDB
		|| *groupname == '%'
#endif
		)
		break;

	    group = (USER *) hash_lookup(grouptable, groupname);
	    if (!group) {
		report(LOG_ERR, "%s=%s, group %s does not exist",
		       (entry->flags & FLAG_ISUSER) ? "user" : "group",
		       entry->name, groupname);
		free(users);
		free(groups);
		return (1);
	    }
	    if (group->flags & FLAG_SEEN) {
		report(LOG_ERR, "recursively defined groups");

		/* print all seen "groups" */
		for (q = groups; *q; q++) {
		    group = *q;
		    if (group->flags & FLAG_SEEN)
			report(LOG_ERR, "%s", group->name);
		}
		free(users);
		free(groups);
		return (1);
	    }
	    group->flags |= FLAG_SEEN;	/* mark group as seen */
	    entry = group;
	}
    }

#ifdef AUTO_MEMBERS

    /* groups with am */
    for (i = 0; i < amcounter; i++) {
        user = amtable[i];

	if (debug & DEBUG_PARSE_FLAG)
	    report(LOG_DEBUG, "auto membering circularity_check: group=%s", user->name);

	/* Initialise all groups "seen" flags to zero */
	for (q = groups; *q; q++) {
	    group = *q;
	    group->flags &= ~FLAG_SEEN;
	}

	entry = user;

	while (entry) {
	    /* check groups we are a member of */
	    groupname = entry->member;

	    if (debug & DEBUG_PARSE_FLAG)
		report(LOG_DEBUG, "\tmember of group %s",
		       groupname ? groupname : "<none>");


	    /* if not a member of any groups, go on to next user */
	    if (!groupname
#ifdef UDB
		|| *groupname == '%'
#endif
		)
		break;

	    group = (USER *) hash_lookup(grouptable, groupname);
	    if (!group) {
		report(LOG_ERR, "group=%s, group %s does not exist",
		       entry->name, groupname);
		free(users);
		free(groups);
		return (1);
	    }
	    if (group->flags & FLAG_SEEN) {
		report(LOG_ERR, "recursively defined groups");

		/* print all seen "groups" */
		for (q = groups; *q; q++) {
		    group = *q;
		    if (group->flags & FLAG_SEEN)
			report(LOG_ERR, "%s", group->name);
		}
		free(users);
		free(groups);
		return (1);
	    }
	    group->flags |= FLAG_SEEN;	/* mark group as seen */
	    entry = group;
	}
    }

#endif

    free(users);
    free(groups);
    return (0);
}


/* Return a value for a group or user (isuser says if
   this name is a group or a user name).

   If no value exists, and recurse is true, also check groups we are a
   member of, recursively.

   Returns void * because it can return a string or a node pointer
   (should really return a union pointer).
*/
static VALUE
cfg_get_value(name, isuser, attr, recurse)
char *name;
int isuser, attr, recurse;
{
    USER *user, *group;
    VALUE value;
    char        *member;
    int reclim = 0;

    value.pval = NULL;

    if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, "cfg_get_value: name=%s isuser=%d attr=%s rec=%d",
	       name, isuser, codestring(attr), recurse);

    /* find the user/group entry */

    user = (isuser) ? get_hash_user(name) :
		    (USER *) hash_lookup(grouptable, name);

    if (!user) {
	if (debug & DEBUG_CONFIG_FLAG)
	    report(LOG_DEBUG, "cfg_get_value: no user/group named %s", name);
	return (value);
    }

    /* found the entry. Lookup value from attr=value */
    value = get_value(user, attr);

    if (value.pval || !recurse) {
	return (value);
    }
    /* no value. Check containing group */
    if (GETMEMBER(user, name))
	group = (USER *) hash_lookup(grouptable, member);
    else
	group = NULL;

    while (group) {
	if (++reclim > 32) { report (LOG_ERR, "group recurse limit exceeded for %s", name); break; }
	if (debug & DEBUG_CONFIG_FLAG)
	    report(LOG_DEBUG, "cfg_get_value: recurse group = %s",
		   group->name);

	value = get_value(group, attr);

	if (value.pval) {
	    return (value);
	}
	/* still nothing. Check containing group and so on */

	if (GETMEMBER(group, name))
	    group = (USER *) hash_lookup(grouptable, member);
	else
	    group = NULL;
    }

    /* no value for this user or her containing groups */
    value.pval = NULL;
    return (value);
}


/* Wrappers for cfg_get_value */
int
cfg_get_intvalue(name, isuser, attr, recurse)
char *name;
int isuser, attr, recurse;
{
    int val = cfg_get_value(name, isuser, attr, recurse).intval;

    if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, "cfg_get_intvalue: returns %d", val);
    return(val);
}

char *
cfg_get_pvalue(name, isuser, attr, recurse)
char *name;
int isuser, attr, recurse;
{
    char *p = cfg_get_value(name, isuser, attr, recurse).pval;

#ifdef UDB
    if (p && !fnmatch("*%[0-9]*", p, 0)) {
      DB *db = cfg_get_udb(name);
      if (db) p = db_replace(name, db, p);
      else report(LOG_ERR, ": no db defined for user %s", name);
    }
#endif

    if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, "cfg_get_pvalue: returns %s", 
	       p ? p : "NULL");
    return(p);
}
/*
   Read the config file and do some basic sanity checking on
   it. Return 1 if we find any errors. */

cfg_read_config(cfile)
char *cfile;
{
    sym_line = 1;

    if ((cf = fopen(cfile, "r")) == NULL) {
	report(LOG_ERR, "read_config: fopen() error for file %s %s, exiting",
	       cfile, sys_errlist[errno]);
	return (1);
    }
    if (parse_decls() || sym_error) {
	fclose(cf);
	return (1);
    }

    if (circularity_check()) {
	fclose(cf);
	return (1);
    }

    fclose(cf);
    return (0);
}

/* return 1 if user exists, 0 otherwise */
int
cfg_user_exists(username)
char *username;
{
    USER *user = get_hash_user(username);

    return (user != NULL);
}

/* return expiry string of user. If none, try groups she is a member
   on, and so on, recursively if recurse is non-zero */
char *
cfg_get_expires(username, recurse)
char *username;

{
    return (cfg_get_pvalue(username, TAC_IS_USER, S_expires, recurse));
}

/* return password string of user. If none, try groups she is a member
   on, and so on, recursively if recurse is non-zero */
char *
cfg_get_login_secret(user, recurse)
char *user;

{
    return (cfg_get_pvalue(user, TAC_IS_USER, S_login, recurse));
}

#ifdef TAC_SMB_AUTH
void cfg_get_smb_data(char *user, int recurse, char **smbprimary, char **smbbackup, char **smbdomain) {
    *smbprimary = cfg_get_pvalue(user, TAC_IS_USER, S_smbprimary, recurse);
    *smbbackup = cfg_get_pvalue(user, TAC_IS_USER, S_smbbackup, recurse);
    *smbdomain = cfg_get_pvalue(user, TAC_IS_USER, S_smbdomain, recurse);
}
#endif

#ifdef TAC_LDAP_AUTH
void cfg_get_ldap_data(char *user, int recurse,
    char **ldapserver, char **ldapbinddn, char **ldapbindpw, char **ldapbasedn
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_SASL)
    , char **ldapsaslmech, char **ldapsaslcred
#endif
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_NS_SSL)
    , char **ldapsslpath
#endif
) {
    *ldapserver = cfg_get_pvalue(user, TAC_IS_USER, S_ldapserver, recurse);
    *ldapbinddn = cfg_get_pvalue(user, TAC_IS_USER, S_ldapbinddn, recurse);
    *ldapbindpw = cfg_get_pvalue(user, TAC_IS_USER, S_ldapbindpw, recurse);
    *ldapbasedn = cfg_get_pvalue(user, TAC_IS_USER, S_ldapbasedn, recurse);
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_SASL)
    *ldapsaslmech = cfg_get_pvalue(user, TAC_IS_USER, S_ldapsaslmech, recurse);
    *ldapsaslcred = cfg_get_pvalue(user, TAC_IS_USER, S_ldapsaslcred, recurse);
#endif
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_NS_SSL)
    *ldapsslpath = cfg_get_pvalue(user, TAC_IS_USER, S_ldapsslpath, recurse);
#endif
}
#endif

#ifdef TAC_ROAMING
void cfg_get_tac_roaming_data(char *user, int recurse, char **traddr, char **trkey) {
    *traddr = cfg_get_pvalue(user, TAC_IS_USER, S_traddr, recurse);
    *trkey = cfg_get_pvalue(user, TAC_IS_USER, S_key, recurse);
}
#endif

#ifdef RAD_ROAMING
void cfg_get_rad_roaming_data(char *user, int recurse, struct radius **radius_conf) {
    *radius_conf = cfg_get_value(user, TAC_IS_USER, S_radius, recurse).pval;
}
#endif

/* return user's secret. If none, try groups she is a member
   on, and so on, recursively if recurse is non-zero */
char *
cfg_get_arap_secret(user, recurse)
char *user;
int recurse;

{
    return (cfg_get_pvalue(user, TAC_IS_USER, S_arap, recurse));
}

char *
cfg_get_chap_secret(user, recurse)
char *user;
int recurse;

{
    return (cfg_get_pvalue(user, TAC_IS_USER, S_chap, recurse));
}

#ifdef MSCHAP
char *
cfg_get_mschap_secret(user, recurse)
char *user;

{
    return (cfg_get_pvalue(user, TAC_IS_USER, S_mschap, recurse));
}
#endif /* MSCHAP */

char *
cfg_get_pap_secret(user, recurse)
char *user;
int recurse;
{
    return (cfg_get_pvalue(user, TAC_IS_USER, S_pap, recurse));
}

char *
cfg_get_opap_secret(user, recurse)
char *user;
int recurse;
{
    return (cfg_get_pvalue(user, TAC_IS_USER, S_opap, recurse));
}

/* return the global password for the user (or the group, etc.) */

char *
cfg_get_global_secret(user, recurse)
char *user;
int recurse;
{
    return (cfg_get_pvalue(user, TAC_IS_USER, S_global, recurse));
}

#ifdef PERUSER_ENABLE
/* return enable string of user. If none, try groups she is a member
   on, and so on, recursively if recurse is non-zero */
char *
cfg_get_enable(user, recurse)
char *user;
int recurse;
{
    return (cfg_get_pvalue(user, TAC_IS_USER, S_enable, recurse)); 
}

#endif

/* Return a pointer to a node representing a given service
   authorization, taking care of recursion issues correctly. Protocol
   is only read if the type is N_svc_ppp. svcname is only read if type
   is N_svc.
*/

NODE *
cfg_get_svc_node(username, type, protocol, svcname, recurse)
char *username;
int type;
char *protocol, *svcname;
int recurse;
{
    USER *user, *group;
    NODE *svc;
    char        *member;
    int reclim = 0;

    if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, 
	       "cfg_get_svc_node: username=%s %s proto=%s svcname=%s rec=%d",
	       username, 
	       cfg_nodestring(type), 
	       protocol ? protocol : "", 
	       svcname ? svcname : "", 
	       recurse);

    /* find the user/group entry */
    user = get_hash_user(username);

    if (!user) {
	if (debug & DEBUG_CONFIG_FLAG)
	    report(LOG_DEBUG, "cfg_get_svc_node: no user named %s", username);
	return (NULL);
    }

    /* found the user entry. Find svc node */
    for(svc = (NODE *) get_value(user, S_svc).pval; svc; svc = svc->next) {

	if (svc->type != type) 
	    continue;

	if (type == N_svc_ppp && !STREQ(svc->value1, protocol)) {
	    continue;
	}

	if (type == N_svc && !STREQ(svc->value1, svcname)) {
	    continue;
	}

	if (debug & DEBUG_CONFIG_FLAG)
	    report(LOG_DEBUG, 
		   "cfg_get_svc_node: found %s proto=%s svcname=%s",
		   cfg_nodestring(type), 
		   protocol ? protocol : "", 
		   svcname ? svcname : "");

	return(svc);
    }

    if (!recurse) {
	if (debug & DEBUG_CONFIG_FLAG)
	    report(LOG_DEBUG, "cfg_get_svc_node: returns NULL");
	return (NULL);
    }

    /* no matching node. Check containing group */
    if (GETMEMBER(user, username))
	group = (USER *) hash_lookup(grouptable, member);
    else
	group = NULL;

    while (group) {
	if (++reclim > 32) { report (LOG_ERR, "group recurse limit exceeded for %s", username); break; }
	if (debug & DEBUG_CONFIG_FLAG)
	    report(LOG_DEBUG, "cfg_get_svc_node: recurse group = %s",
		   group->name);

	for(svc = (NODE *) get_value(group, S_svc).pval; svc; svc = svc->next) {

	    if (svc->type != type) 
		continue;

	    if (type == N_svc_ppp && !STREQ(svc->value1, protocol)) {
		continue;
	    }

	    if (type == N_svc && !STREQ(svc->value1, svcname)) {
		continue;
	    }

	    if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, 
		       "cfg_get_svc_node: found %s proto=%s svcname=%s",
		       cfg_nodestring(type), 
		       protocol ? protocol : "", 
		       svcname ? svcname : "");

	    return(svc);
	}

	/* still nothing. Check containing group and so on */

	if (GETMEMBER(group, username))
	    group = (USER *) hash_lookup(grouptable, member);
	else
	    group = NULL;
    }

    if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, "cfg_get_svc_node: returns NULL");

    /* no matching svc node for this user or her containing groups */
    return (NULL);
}

/* Return a pointer to the node representing a set of command regexp
   matches for a user and command, handling recursion issues correctly */
NODE *
cfg_get_cmd_node(name, cmdname, recurse)
char *name, *cmdname;
int recurse;

{
    USER *user, *group;
    NODE *svc;
    char        *member;
    int reclim = 0;

    if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, "cfg_get_cmd_node: name=%s cmdname=%s rec=%d",
	       name, cmdname, recurse);

    /* find the user/group entry */
    user = get_hash_user(name);

    if (!user) {
	if (debug & DEBUG_CONFIG_FLAG)
	    report(LOG_DEBUG, "cfg_get_cmd_node: no user named %s", name);
	return (NULL);
    }
    /* found the user entry. Find svc node */
    svc = (NODE *) get_value(user, S_svc).pval;

    while (svc) {
	if (svc->type == N_svc_cmd && STREQ(svc->value, cmdname)) {
	    if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, "cfg_get_cmd_node: found cmd %s %s node",
		       cmdname, cfg_nodestring(svc->type));
	    return (svc);
	}
	svc = svc->next;
    }

    if (!recurse) {
	if (debug & DEBUG_CONFIG_FLAG)
	    report(LOG_DEBUG, "cfg_get_cmd_node: returns NULL");
	return (NULL);
    }
    /* no matching node. Check containing group */
    if (GETMEMBER(user, name))
	group = (USER *) hash_lookup(grouptable, member);
    else
	group = NULL;

    while (group) {
	if (++reclim > 32) { report (LOG_ERR, "group recurse limit exceeded for %s", name); break; }
	if (debug & DEBUG_CONFIG_FLAG)
	    report(LOG_DEBUG, "cfg_get_cmd_node: recurse group = %s",
		   group->name);

	svc = get_value(group, S_svc).pval;

	while (svc) {
	    if (svc->type == N_svc_cmd && STREQ(svc->value, cmdname)) {
		if (debug & DEBUG_CONFIG_FLAG)
		    report(LOG_DEBUG, "cfg_get_cmd_node: found cmd %s node %s",
			   cmdname, cfg_nodestring(svc->type));
		return (svc);
	    }
	    svc = svc->next;
	}

	/* still nothing. Check containing group and so on */

	if (GETMEMBER(group, name))
	    group = (USER *) hash_lookup(grouptable, member);
	else
	    group = NULL;
    }

    if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, "cfg_get_cmd_node: returns NULL");

    /* no matching cmd node for this user or her containing groups */
    return (NULL);
}

/* Return an array of character strings representing configured AV
 * pairs, given a username and a service node. 
 *
 * In the AV strings returned, manipulate the separator character to
 * indicate which args are optional and which are mandatory.
 *
 * Lastly, indicate what default permission was configured by setting
 * denyp */

char **
cfg_get_svc_attrs(svcnode, denyp)
NODE *svcnode;
int *denyp;
{
    int i;
    NODE *node;
    char **args;

    *denyp = 1;

    if (!svcnode)
	return (NULL);

    *denyp = (svcnode->dflt == S_deny);

    i = 0;
    for (node = svcnode->value; node; node = node->next)
	i++;

    args = (char **) tac_malloc(sizeof(char *) * (i + 1));

    i = 0;
    for (node = svcnode->value; node; node = node->next) {
	char *arg = tac_strdup(node->value);
	char *p = index(arg, '=');

	if (p && node->type == N_optarg)
	    *p = '*';
	args[i++] = arg;
    }
    args[i] = NULL;
    return (args);
}


int
cfg_user_svc_default_is_permit(user)
char *user;

{
    int permit = cfg_get_intvalue(user, TAC_IS_USER, S_svc_dflt,
			       TAC_PLUS_RECURSE);

    switch (permit) {
    default:			/* default is deny */
    case S_deny:
	return (0);
    case S_permit:
	return (1);
    }
}

int
cfg_no_user_permitted()
{
    if (no_user_dflt == S_permit)
	return (1);
    return (0);
}


char *
cfg_get_authen_default()
{
    return (authen_default);
}

#ifdef UNIX_LOGIN

char *
cfg_get_prompt_default()
{
    return prompt_default;
}

char *
cfg_get_reject_default(user, recurse)
char *user;
int recurse;
{
 char *p = cfg_get_pvalue(user, TAC_IS_USER, S_reject, recurse);
 return (p) ? p : reject_default;
}

#endif	/* UNIX_LOGIN */

#ifdef CRON

char *
cfg_get_cron_default(user, recurse)
char *user;
int recurse;
{
 char *p = cfg_get_pvalue(user, TAC_IS_USER, S_crondeny, recurse);
 return (p) ? p : cron_default;
}

#endif

#ifdef DENY

char *
cfg_get_deny_default(user, recurse)
char *user;
int recurse;
{
 char *p = cfg_get_pvalue(user, TAC_IS_USER, S_admdeny, recurse);
 return (p) ? p : deny_default;
}

#endif

/* Return 1 if this user has any ppp services configured. Used for
   authorizing ppp/lcp requests */
int
cfg_ppp_is_configured(username, recurse)
    char *username;
    int recurse;
{
    USER *user, *group;
    NODE *svc;
    char        *member;
    int reclim = 0;

    if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, "cfg_ppp_is_configured: username=%s rec=%d",
	       username, recurse);

    /* find the user/group entry */
    user = get_hash_user(username);

    if (!user) {
	if (debug & DEBUG_CONFIG_FLAG)
	    report(LOG_DEBUG, "cfg_ppp_is_configured: no user named %s", 
		   username);
	return (0);
    }

    /* found the user entry. Find svc node */
    for(svc = (NODE *) get_value(user, S_svc).pval; svc; svc = svc->next) {

	if (svc->type != N_svc_ppp) 
	    continue;

	if (debug & DEBUG_CONFIG_FLAG)
	    report(LOG_DEBUG, "cfg_ppp_is_configured: found svc ppp %s node",
		   svc->value1);
	
	return(1);
    }

    if (!recurse) {
	if (debug & DEBUG_CONFIG_FLAG)
	    report(LOG_DEBUG, "cfg_ppp_is_configured: returns 0");
	return (0);
    }

    /* no matching node. Check containing group */
    if (GETMEMBER(user, username))
	group = (USER *) hash_lookup(grouptable, member);
    else
	group = NULL;

    while (group) {
	if (++reclim > 32) { report (LOG_ERR, "group recurse limit exceeded for %s", username); break; }
	if (debug & DEBUG_CONFIG_FLAG)
	    report(LOG_DEBUG, "cfg_ppp_is_configured: recurse group = %s",
		   group->name);

	for(svc = (NODE *) get_value(group, S_svc).pval; svc; svc = svc->next) {

	    if (svc->type != N_svc_ppp)
		continue;

	    if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, "cfg_ppp_is_configured: found svc ppp %s node",
		       svc->value1);
	
	    return(1);
	}

	/* still nothing. Check containing group and so on */

	if (GETMEMBER(group, username))
	    group = (USER *) hash_lookup(grouptable, member);
	else
	    group = NULL;
    }

    if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, "cfg_ppp_is_configured: returns 0");

    /* no PPP svc nodes for this user or her containing groups */
    return (0);
}

#ifdef UDB
/*
 * Get database handle for some particular user
 */
DB      *cfg_get_udb(name)
    char    *name;
{
	DB      *db;

	if (!(db = (DB *) cfg_get_pvalue(name, TAC_IS_USER,
		S_db, TAC_PLUS_RECURSE)))
			db = default_udb;
	return (db);
}

/*
 * Extract group name from database, if applies
 */
char    *cfg_getmember(user, name)
	USER            *user;
	char            *name;
{
	DB              *db;
	char            *p, *member;

	member = user->member;
	if (!member || *member != '%') {
		if (debug & DEBUG_AUTHEN_FLAG)
			report(LOG_DEBUG, "cfg_getmember rets %s",
				member ? member : "NULL");
		return (member);
	}
	db = user->udb ? user->udb : default_udb;
	/* FIXME cfg_get_udb(name) instead of default_udb here leads to infinite recursion:
	   cfg_getmember -> cfg_get_udb -> cfg_get_pvalue -> cfg_get_value -> cfg_getmember */
	if (!db) {
		report (LOG_ERR, "no db defined for user %s",
			user->name);
		return (NULL);
	}
	p = db_replace(name, db, member);
	if (*p == '%') {
		report(LOG_ERR, ": no such field in db: %s (getmember)",
			member);
		return (NULL);
	}
	if (debug & DEBUG_AUTHEN_FLAG)
		report(LOG_DEBUG, "cfg_getmember rets %s", p);
	return (p);
}
#endif

#ifdef DENY

static int check_allow_file(name)
char *name;
{
 struct stat st;
 char *fn;
 int retval;

 fn = tac_malloc(sizeof(TAC_PLUS_ALLOWDIR) + strlen(name) + 1);
 strcpy(fn, TAC_PLUS_ALLOWDIR);
 strcat(fn, "/");
 strcat(fn, name);

 retval = (stat(fn, &st) >= 0);

 free(fn);
 return retval;
}

static char *check_deny_file(name, realname, data, recurse)
char *name;
char *realname;
struct authen_data *data;
int recurse;
{
 struct stat st;
 char *fn;
 int f;
 char *retval = NULL;

 fn = tac_malloc(sizeof(TAC_PLUS_DENYDIR) + strlen(name) + 1);
 strcpy(fn, TAC_PLUS_DENYDIR);
 strcat(fn, "/");
 strcat(fn, name);

 if(stat(fn, &st) >= 0)
  {
   if(st.st_size && ((f = open(fn, O_RDONLY, 0)) >= 0))
    {
     retval = tac_malloc(st.st_size+1);
     read(f, retval, st.st_size);
     retval[st.st_size] = '\0';
     close(f);
    }
   else
    {
      retval =
#ifdef MORE_SUBST
	substitute2(cfg_get_deny_default(realname, recurse), (void *)data);
#else
        tac_strdup(cfg_get_deny_default(realname, recurse));
#endif
    }
  }

 free(fn);
 return retval;
}

int
set_sysadm_status(name, data, recurse)
char *name;
struct authen_data *data;
int recurse;
{
  USER *user, *group;

  char *retval = NULL;

  if (debug & DEBUG_AUTHEN_FLAG)
	report(LOG_DEBUG, "set_sysadm_status: name=%s rec=%d",
	       name, recurse);

  /* find the user/group entry */

  user = get_hash_user(name);

  if (!user) {
  /* user from default passwd file possible */
	if (debug & DEBUG_AUTHEN_FLAG)
	    report(LOG_DEBUG, "set_sysadm_status: no user named %s", name);
        recurse = 0;		/* must be 0 now, but... */
  }

#ifndef DENY_THEN_ALLOW
  if(check_allow_file(name))
    return 0;
#endif

  if(
     (retval = check_deny_file(name, name, data, recurse)) == NULL &&
#ifdef DENY_THEN_ALLOW
     !check_allow_file(name) &&
#endif
     recurse
    )
   {
    /* no file. Check containing group */
      char *member;
      int reclim = 0;
      if (GETMEMBER(user, name))
	group = (USER *) hash_lookup(grouptable, member);
      else
	group = NULL;

      while (group) {
	if (++reclim > 32) { report (LOG_ERR, "group recurse limit exceeded for %s", name); break; }
	if (debug & DEBUG_AUTHEN_FLAG)
	    report(LOG_DEBUG, "set_sysadm_status: recurse group = %s",
		   group->name);

#ifndef DENY_THEN_ALLOW
        if(check_allow_file(group->name))
          break;
#endif

        if((retval = check_deny_file(group->name, name, data, recurse)) != NULL)
          break;

#ifdef DENY_THEN_ALLOW
        if(check_allow_file(group->name))
          break;
#endif

	if (GETMEMBER(group, name))
	    group = (USER *) hash_lookup(grouptable, member);
	else
	    group = NULL;
      }
   }

  if(retval != NULL)
   {
    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
    fr_fi(data->server_msg, retval);
    if (debug & DEBUG_AUTHEN_FLAG)
      report(LOG_INFO, "user %s login not allowed by deny/allow rules", name);
    return 1;
   }

  return 0;
}

#endif


#ifdef CRON

static char *cronfn;

#ifdef CRON_REGEXP

static int regport(p, x)
char *p, *x;
{
 regex_t *r=tac_malloc(sizeof(regex_t));
 int i;

 if(regcomp(r, p, REG_EXTENDED))
  {
   report(LOG_WARNING, "bad regular expression %s in cron file %s ignored", p, cronfn);
   return 0;
  }

 i = !regexec(r, x, 0, NULL, 0);

 free(r);
 return i;
}

#else
 #define regport(p, x) (!fnmatch(p, x, 0))
#endif /* CRON_REGEXP */

static int port_match(data, list)
struct authen_data *data;
char *list;
{
  char *p;
  char tmp[256];


  if (debug & DEBUG_AUTHEN_FLAG)
     report(LOG_DEBUG, "port_match: user=%s nas=%s port=%s list=%s",
	 data->NAS_id->username,
	 data->NAS_id->NAS_name,
	 data->NAS_id->NAS_port,
	 list);

  for (p = strtok(list, " \t\n"); p; p = strtok(NULL, " \t\n"))
   {

#if defined(ALIASES) && defined(TTYS)
    if((data->NAS_id->NAS_flags & (A_FLAG_CHECK | A_FLAG_PCHECK)) ==
                                                (A_FLAG_CHECK | A_FLAG_PCHECK))
     {
      snprintf(tmp, sizeof(tmp), "%s@%s", data->NAS_id->NAS_palias,
                                data->NAS_id->NAS_alias);
      if (regport(p, tmp))
        return 1;
     }
#endif

#ifdef TTYS
    if(data->NAS_id->NAS_flags & A_FLAG_PCHECK)
     {
      snprintf(tmp, sizeof(tmp), "%s@%s", data->NAS_id->NAS_palias,
                                data->NAS_id->NAS_name);
      if (regport(p, tmp))
        return 1;
     }
#endif

#ifdef ALIASES
    if(data->NAS_id->NAS_flags & A_FLAG_CHECK)
     {
      snprintf(tmp, sizeof(tmp), "%s@%s", data->NAS_id->NAS_port,
                                data->NAS_id->NAS_alias);
      if (regport(p, tmp))
        return 1;
     }
#endif

    snprintf(tmp, sizeof(tmp), "%s@%s", data->NAS_id->NAS_port,
                                data->NAS_id->NAS_name);
    if (regport(p, tmp))
      return 1;
   }

  return 0;
}

int
set_time_status(name, data, pw, recurse)
char *name;
struct authen_data *data;
struct passwd *pw;
int recurse;
{
  time_t t = time(NULL);
  struct tm *tm = localtime(&t);
  char *envp[] = { NULL };
  entry *e;
  int minute, hour, dom, month, dow;
  FILE *f;
  static char *system_cron_dir = TAC_PLUS_CRONDIR;
  USER *user, *group;

  int retval = 0;

  if (debug & DEBUG_AUTHEN_FLAG)
	report(LOG_DEBUG, "set_time_status: name=%s rec=%d",
	       name, recurse);

  /* find the user/group entry */
  user = get_hash_user(name);

  if (!user) {
  /* user from default passwd file possible */
	if (debug & DEBUG_AUTHEN_FLAG)
	    report(LOG_DEBUG, "set_time_status: no user named %s", name);
        recurse = 0;		/* must be 0 now, but... */
  }

  cronfn = tac_malloc(strlen(system_cron_dir) + strlen(name) + 2);
  strcpy(cronfn, system_cron_dir);
  strcat(cronfn, "/");
  strcat(cronfn, name);

  if (!(f = fopen(cronfn, "r")) && recurse)
   {
    /* no file. Check containing group */
      char *member;
      int reclim = 0;
      if (GETMEMBER(user, name))
	group = (USER *) hash_lookup(grouptable, member);
      else
	group = NULL;

      while (group) {
	if (++reclim > 32) { report (LOG_ERR, "group recurse limit exceeded for %s", name); break; }
	if (debug & DEBUG_AUTHEN_FLAG)
	    report(LOG_DEBUG, "set_time_status: recurse group = %s",
		   group->name);

        free(cronfn);
        cronfn = tac_malloc(strlen(system_cron_dir) + strlen(group->name) + 2);
        strcpy(cronfn, system_cron_dir);
        strcat(cronfn, "/");
        strcat(cronfn, group->name);

        if ((f = fopen(cronfn, "r")) != NULL) /* he is john! */
          break;

	if (GETMEMBER(group, name))
	    group = (USER *) hash_lookup(grouptable, member);
	else
	    group = NULL;
      }
   }

  if(f)
   {
    minute = tm->tm_min -FIRST_MINUTE;
    hour = tm->tm_hour -FIRST_HOUR;
    dom = tm->tm_mday -FIRST_DOM;
    month = tm->tm_mon +1 /* 0..11 -> 1..12 */ -FIRST_MONTH;
    dow = tm->tm_wday -FIRST_DOW;

    retval = 1;

    while ((e = load_entry(f, NULL, pw, envp)))
     {
      if(bit_test(e->minute, minute)
	&& bit_test(e->hour, hour)
	&& bit_test(e->month, month)
	&& ( ((e->flags & DOM_STAR) || (e->flags & DOW_STAR))
	     ? (bit_test(e->dow,dow) && bit_test(e->dom,dom))
	     : (bit_test(e->dow,dow) || bit_test(e->dom,dom))
	     )
	&& port_match(data, e->cmd)
	)
       {
        free_entry(e);
        retval = 0;
        break;
       }

      free_entry(e);
     }

    free(cronfn);
    fclose(f);

    if(retval)
     {
      data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
#ifdef MORE_SUBST
      fr_fi(data->server_msg,
	substitute2(cfg_get_cron_default(name, recurse), (void *)data));
#else
      fr_fi(data->server_msg,
	tac_strdup(cfg_get_cron_default(name, recurse)));
#endif
     }
   }

  if((debug & DEBUG_AUTHEN_FLAG) && retval)
    report(LOG_INFO, "user %s login not allowed by cron rules", name);

  return retval;
}

#endif

#ifdef DBF_1
#include "../acctd/acctd-config.h"
#ifdef HAVE_DB_185_H
#include <db_185.h>
#else
#include <db.h>
#endif

int find_ug_tl_cond(TQ_st *tq) { return (tq->tl == TIME_LIMIT_NO || tq->tl == TIME_LIMIT_OFF || tq->tll > 0); }
/* FIXME was (tq->tl >= 0) */
int find_ug_dl_cond(TQ_st *tq) { return (tq->dl == TIME_LIMIT_NO || tq->dl == TIME_LIMIT_OFF || tq->dll > 0); }
/* FIXME was (tq->dl >= 0) */
#ifdef PARROT
#include <float.h>
static double parrot_off=PARROT_OFF; /* avoid rounding errors */
int find_ug_pl_cond(TQ_st *tq) { return (tq->parrot == parrot_off || (tq->parrot + tq->parrot_credit) > DBL_EPSILON); }

/* find_parrot_file: find parrot file for user */
FILE *find_parrot_file(char *name) {
  USER *user, *group;
  char *cronfn;
  char *member;
  FILE *f;
  int reclim = 0;

  cronfn = tac_malloc(strlen(PARROTDIR "/") + strlen(name) + 1);
  strcpy(cronfn, PARROTDIR "/");
  strcat(cronfn, name);
  if (f = fopen(cronfn, "r")) {
    report(LOG_DEBUG, "find_parrot_file: using file %s for %s", cronfn, name);
    free(cronfn);
    return f;
  }
  free(cronfn);

  user = get_hash_user(name);
  if (!user) {
    if (debug & DEBUG_CONFIG_FLAG)
      report(LOG_INFO, "find_parrot_file: no user named %s", name);
    return NULL;
  }

  /* no value. Check containing group */
  if (GETMEMBER(user, name))
    group = (USER *) hash_lookup(grouptable, member);
  else
    group = NULL;

  while (group) {
    if (++reclim > 32) { report (LOG_ERR, "group recurse limit exceeded for %s", name); break; }
    if (debug & DEBUG_CONFIG_FLAG)
      report(LOG_DEBUG, "find_parrot_file: recurse group = %s",
	     group->name);
    cronfn = tac_malloc(strlen(PARROTDIR "/") + strlen(group->name) + 1);
    strcpy(cronfn, PARROTDIR "/");
    strcat(cronfn, group->name);
    if (f = fopen(cronfn, "r")) {
      report(LOG_DEBUG, "find_parrot_file: using file %s for %s", cronfn, name);
      free(cronfn);
      return f;
    }
    free(cronfn);
    /* still nothing. Check containing group and so on */
    if (GETMEMBER(group, name))
      group = (USER *) hash_lookup(grouptable, member);
    else
      group = NULL;
  }

  return NULL;
}

#endif

/* find_ug: find user or group record in DB matching condition.
 * return values:
 * -1: error reading db
 * 0: found valid user record
 * 1: found valid group record
 * 2: not found in get_hash_user (should not happen?), no user record
 * 3: not found in get_hash_user (should not happen?), user record present
 * 4: no groups (at all or with a resource) found, no user record
 * 5: no groups (at all or with a resource) found, user record present
 * 6: not found in groups, no user record
 * 7: not found in groups, user record present
 */
int find_ug(DB *db, TQ_st *tq, char *name, int (*cond)(TQ_st *), DBT *key, char **gname) {
  USER *user, *group;
  DBT val[1];
  TQ_st u_tq;
  int u=0;
  int g=0;
  char *member;
  int reclim = 0;

  *gname = NULL;
  key->data = name;
  key->size = strlen(name) + 1;
  switch (u = (*db->get)(db, key, val, NULL)) {
  case -1:
    report(LOG_ERR, "find_ug: reading user db: %s", strerror(errno));
    return -1;
  case 0:
    if (!val->data) { u=0; break; } /* FIXME could this happen? */
    memcpy(tq, val->data, sizeof(TQ_st));
    memcpy(&u_tq, val->data, sizeof(TQ_st));
    if (cond(tq)) return 0;
  }
  u = 1-u;

  user = get_hash_user(name);
  if (!user) {
    if (debug & DEBUG_CONFIG_FLAG)
      report(LOG_DEBUG, "fing_ug: no user named %s", name);
    return 2+u;
  }

  /* no value. Check containing group */
  if (GETMEMBER(user, name))
    group = (USER *) hash_lookup(grouptable, member);
  else
    group = NULL;

  if (group == NULL) return 4+u;

  while (group) {
    if (++reclim > 32) { report (LOG_ERR, "group recurse limit exceeded for %s", name); break; }
    if (debug & DEBUG_CONFIG_FLAG)
      report(LOG_DEBUG, "find_ug: recurse group = %s",
	     group->name);

    key->data = group->name;
    key->size = strlen(group->name) + 1;
    if (!((*db->get)(db, key, val, NULL))) {
      g=1;
      memcpy(tq, val->data, sizeof(TQ_st));
      if (cond(tq)) { *gname = group->name; return 1; }
    }
    /* still nothing. Check containing group and so on */
    if (GETMEMBER(group, name))
      group = (USER *) hash_lookup(grouptable, member);
    else
      group = NULL;
  }

  /* no value for this user or her containing groups */
  if (u) {
    key->data = name;
    key->size = strlen(name) + 1;
    memcpy(tq, &u_tq, sizeof(TQ_st));
  }
  if (!g) return 4+u; else return 6+u;
}
#endif /* DBF_1 */

#ifdef ACL
/*
 * Check if we are a member of a given group.
 */
int     acl_checkgrp(rp, name)
	regex_t *rp;
	char            *name;
{
	USER            *user, *group = (USER *) 0;
	char            *member;
	int reclim = 0;

	user = (USER *) hash_lookup(usertable, name);
	if (!user) {
		if (debug & DEBUG_CONFIG_FLAG)
			report(LOG_DEBUG, "cfg_get_value: no user named %s", name);
#ifdef UDB
		goto checkgrp;
#else
		return (ACL_DENY);
#endif
	}
	if (GETMEMBER(user, name))
		group = (USER *) hash_lookup(grouptable, member);
	else
		group = NULL;
#ifdef UDB
checkgrp:
#endif
	while (group) {
	  if (++reclim > 32) { report (LOG_ERR, "group recurse limit exceeded for %s", name); break; }
	  if (!regexec(rp, group->name, 0, NULL, 0))
	    return (ACL_PERMIT);
	  if (GETMEMBER(group, name))
	    group = (USER *) hash_lookup(grouptable, member);
	  else
	    group = NULL;
	}
	return (ACL_DENY);
}

/*
 * Parse ACL statements
 */
#define ACL_PARSE_RVAL(p) \
	sym_get();                              \
	parse(S_separator);                     \
        p=(regex_t *)tac_malloc(sizeof(regex_t)); \
	if (regcomp(p, sym_buf, REG_EXTENDED)) { \
		parse_error("syntax error in regexp when defining access-group clause, line %d\n", \
			sym_line);              \
		return (1);                     \
	}                                       \
	sym_get();                              \
	continue;

int     acl_parse(sym, ap)
	int     sym;
	struct  agroup  *ap;
{
	int     op;
	struct  aexp    *ae;

	op = (sym == S_permit) ? ACL_PERMIT : ACL_DENY;
	parse(S_openbra);
	ae = (struct aexp *) tac_malloc(sizeof(*ae));
	ae->op = op;
	ae->user   = ae->group = ae->from = ae->tty = (regex_t *) 0;
	ae->agroup = (char *) 0;
	ae->nasg   = (struct nasgrp *) 0;
	ae->tg     = (struct tgrade *) 0;
	ae->next   = (struct aexp *) 0;

	while (sym_code != S_closebra) {
		switch (sym_code) {

		    case S_group:
			ACL_PARSE_RVAL(ae->group)

		    case S_user:
			ACL_PARSE_RVAL(ae->user)

		    case S_tty:
			ACL_PARSE_RVAL(ae->tty)

		    case S_from:
			ACL_PARSE_RVAL(ae->from)
#ifdef DXS1
		    case S_dxs1:
			ACL_PARSE_RVAL(ae->dxs1)
#endif
		    case S_time:
			sym_get();
			parse(S_separator);
			if (!(ae->tg = acl_parse_tgrade(sym_buf))) {
				parse_error("cannot parse timegrade, line %d\n",
					sym_line);
				return (1);
			}
			sym_get();
			continue;

		    case S_nasgrp:
			sym_get();
			parse(S_separator);
			if (!(ae->nasg = acl_getnas(sym_buf))) {
				parse_error("undefined NAS group %s, line %d\n",
					sym_buf, sym_line);
				return (1);
			}
			sym_get();
			continue;

		   case S_agroup:
			sym_get();
			parse(S_separator);
			if (!hash_lookup(acltable, sym_buf)) {
				parse_error("referencing undefined ACL group %s, line %d\n",
					sym_buf, sym_line);
				return(1);
			}
			ae->agroup = tac_strdup(sym_buf);
			sym_get();
			continue;

		   default:
			parse_error("Unknown access-group clause %s, line %d\n",
				sym_buf, sym_line);
			return(1);
		}
	}
	parse(S_closebra);
	if (!ap->atail) {
		ap->aexps = ap->atail = ae;
		return (0);
	}
	ap->atail->next = ae;
	ap->atail = ae;
	return (0);
}

#endif
