/*
 * advanced tacacs+ server based on Cisco Systems tac_plus code.
 * For Cisco copyright notice, see LICENSE.CISCO
 *
 * This is a copyrighted commercial software; see LICENSE for details.
 */

#include "tac_plus.h"
#include "dynipd.h"

#define LOCALHOST       "127.0.0.1"
#define UNKN            "SERVER-FAIL"

static  u_long  saddr  = 0;
static  int     serial = 0;
static  u_long  lport  = DPORT;

/*
 * Resolve and store dynamic IP host name. Also optional port
 * number can be supplied after hostname using colon, e.g.
 * mv.demos.su:20005
 */
void    dyna_sethost(hname)
	char    *hname;
{
	struct  hostent         *hp;
	char                    *p;

	serial = getpid();

	if ((p = strchr(hname, ':'))) {
		*p = '\0';
		if (sscanf(++p, "%ld", &lport) != 1) {
			report(LOG_ERR, "bad dynipd port: %s", p);
			tac_exit(1);
		}
	}
	if (!(saddr = inet_addr(hname))) {
		if (!(hp = gethostbyname(hname))) {
			report(LOG_ERR, "dyna_sethost: can't resolve dynamic ip server address (%s)", hname);
			tac_exit(1);
		}
		saddr = *((u_long *) hp->h_addr);
	}
	report(LOG_DEBUG, "dynip server addr set to %s, lport now %u",
		inet_ntoa(*((struct in_addr *) &saddr)), lport);
}

#define RFAIL   if (sd > 0) close(sd); return tac_strdup(UNKN);

/*
 * Process server request
 * (also serve duplicate requests - cache last successful one)
 */
char   *dyna_req(char *name, char *addr, char *nasinfo, int mode) {
	struct sockaddr_in      sca;
	struct  dynp            dd;
	int                     sd = -1, one = 1;
	char                    *pool = NULL;
	char *tmp;

	pool  = cfg_get_pvalue(name, TAC_IS_USER,
	    S_apool, TAC_PLUS_RECURSE);
#ifdef UDB
	if (pool && *pool == '%') {
		DB      *db = (DB *) cfg_get_udb(name);
		if (!db) {
			report(LOG_ERR, "dyna_req: No database assigned for user %s", name);
			RFAIL
		}
		pool = db_replace(name, db, pool);
		if (*pool == '%') {
			report(LOG_ERR, "dyna_req: can't subst apool %s from db", pool);
			RFAIL
		}
	}
#endif /* UDB */
	if (!saddr)
		saddr = inet_addr(LOCALHOST);

	if ((sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
		report(LOG_ERR, "dyna_req: can't get socket: %s", strerror(errno));
		RFAIL
	}
	if (setsockopt(sd, SOL_SOCKET, SO_KEEPALIVE,
	    (char *) &one, sizeof (one)) < 0) {
		report(LOG_ERR, "dyna_req: setsockopt KEEPALIVE: %s", strerror(errno));
		RFAIL
	}
#ifdef HAS_SNDLOWAT
	one = sizeof (dd);
	if (setsockopt(sd, SOL_SOCKET, SO_SNDLOWAT,
	    (char *) &one, sizeof (one)) < 0) {
		report(LOG_ERR, "dyna_req: setsockopt SNDLOWAT: %s", strerror(errno));
		RFAIL
	}
#endif
	bzero((char *) &sca, sizeof (sca));
	sca.sin_family      = AF_INET;
	sca.sin_addr.s_addr = saddr;
	sca.sin_port        = htons(lport);

	if (connect(sd, (struct sockaddr *) &sca, sizeof(sca)) < 0) {
		report(LOG_ERR, "dyna_req: can't connect to dynip server, addr %s",
			inet_ntoa(*((struct in_addr *) &saddr)));
		RFAIL
	}
	dd.serial = serial;
	dd.magic  = DYNP_MAGIC;

	strncpy(dd.argb, name, MAXARGBUF/2 - 1);
	switch (mode) {
	case 0: /* address request */
	  tmp = &dd.argb[MAXARGBUF/2];
	  if (pool) {
	    dd.op     = DYNP_IARQ2;
	    strncpy(tmp, pool, &dd.argb[MAXARGBUF] - tmp);
	    tmp += strlen(pool) + 1;
	  } else
	    dd.op     = DYNP_IARQ;
	  strncpy(tmp, nasinfo, &dd.argb[MAXARGBUF] - tmp);
	  break;
	case 1: /* logout notification */
	  dd.op = DYNP_LONOT;
	  sprintf(&dd.argb[MAXARGBUF/2], "%s", addr);
	  break;
	case 2: /* login notification */
	  tmp = &dd.argb[MAXARGBUF/2];
	  dd.op = DYNP_LINOT;
	  tmp += sprintf(&dd.argb[MAXARGBUF/2], "%s", addr) + 1;
	  strncpy(tmp, nasinfo, &dd.argb[MAXARGBUF] - tmp);
	  break;
	}

	if (write(sd, (char *) &dd, 1) != 1 ||
	    write(sd, (char *) &dd, sizeof(dd)) != sizeof(dd)) {
		report(LOG_ERR, "dyna_req: can't write to dynip socket");
		RFAIL
	}
	dd.serial = dd.op = dd.magic = 0;
	read(sd, (char *) &dd, 1);
	if (read(sd, (char *) &dd, sizeof(dd)) != sizeof(dd)) {
		report(LOG_ERR, "dyna_req: reading dynip socket");
		RFAIL
	}
	if (dd.serial != serial || dd.magic != DYNP_MAGIC ||
	    (dd.op != DYNP_IAREPL && dd.op != DYNP_IERROR && dd.op != DYNP_OKAY)) {
		report(LOG_ERR, "dyna_req: protocol error, bad serial/magic/op");
		RFAIL
	}
	if (dd.op == DYNP_IERROR) {
		report(LOG_ERR, "dyna_req: dynip server error: %s", dd.argb);
		RFAIL
	}
	if (dd.op == DYNP_IAREPL && !inet_addr(dd.argb)) {
		report(LOG_ERR, "dyna_req: dynip server returned bad ipn (%s)", dd.argb);
		RFAIL
	}
	close(sd);

	if (mode == 0)
	  return tac_strdup(dd.argb);
	else
	  return NULL;
}
