/*
   Copyright (c) 1995-2000 by Cisco systems, Inc.

   Permission to use, copy, modify, and distribute modified and
   unmodified copies of this software for any purpose and without fee is
   hereby granted, provided that (a) this copyright and permission notice
   appear on all copies of the software and supporting documentation, (b)
   the name of Cisco Systems, Inc. not be used in advertising or
   publicity pertaining to distribution of the program without specific
   prior permission, and (c) notice be given in supporting documentation
   that use, modification, copying and distribution is by permission of
   Cisco Systems, Inc.

   Cisco Systems, Inc. makes no representations about the suitability
   of this software for any purpose.  THIS SOFTWARE IS PROVIDED ``AS
   IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
   WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS FOR A PARTICULAR PURPOSE.
*/

/* Keywords of the configuration language */

#include "tac_plus.h"

static void *wordtable[HASH_TAB_SIZE];	/* Table of keyword declarations */

struct keyword {
    char *word;
    void *hash;
    u_char value;
};

typedef struct keyword KEYWORD;

static void
declare(name, value)
    char *name;
    int value;
{
    KEYWORD *n;
    KEYWORD *k = (KEYWORD *)tac_malloc(sizeof(KEYWORD));

    k->word = tac_strdup(name);
    k->value = value;

    n = hash_add_entry(wordtable, (void *) k);

    if (n) {
	report(LOG_ERR, "Attempt to multiply define keyword %s",
	       name);
	tac_exit(1);
    }
}

/* Declare keywords of the "configuration language". */

void 
parser_init()
{
    bzero(wordtable, sizeof(wordtable));

    declare("access", S_access);
    declare("accounting", S_accounting);
    declare("after", S_after);
    declare("allow-from", S_allow_from);
    declare("arap", S_arap);
    declare("attribute", S_attr);
    declare("authentication", S_authentication);
    declare("authorization", S_authorization);
    declare("before", S_before);
    declare("chap", S_chap);
    declare("acct_start", S_acct_start);
    declare("acct_stop", S_acct_stop);
    declare("acct_update", S_acct_update);
#ifdef MSCHAP
    declare("ms-chap", S_mschap);
#endif /* MSCHAP */
    declare("cleartext", S_cleartext);
    declare("cmd", S_cmd);
    declare("default", S_default);
    declare("deny", S_deny);
    declare("des", S_des);
#ifdef EMESSAGE
    declare("emessage", S_emessage);
#endif
#ifdef PERUSER_ENABLE
    declare("enable", S_enable);
#endif
    declare("exec", S_exec);
    declare("expires", S_expires);
    declare("file", S_file);
    declare("group", S_group);
    declare("global", S_global);
    declare("host", S_host);
    declare("ip", S_ip);
    declare("ipx", S_ipx);
    declare("key", S_key);
    declare("nas", S_nas);
    declare("nasaddr", S_nasaddr);
    declare("rsh-user", S_rshuser);
#ifdef ENABLE_SNMP
    declare("snmp-community", S_snmpcomm);
#endif
    declare("lcp", S_lcp);
    declare("listen-on", S_listen_on);
#ifdef MAXSESS
    declare("maxsess", S_maxsess);
#endif
    declare("member", S_member);
    declare("message", S_message);
    declare("name", S_name);
#if defined(NONE_LOGIN) || defined(PERUSER_ENABLE)
    declare("none", S_none);
    declare("nopassword", S_nopasswd);
#endif
    declare("optional", S_optional);
    declare("login", S_login);
    declare("permit", S_permit);
    declare("pap", S_pap);
    declare("opap", S_opap);
    declare("ppp", S_ppp);
#ifdef UNIX_LOGIN
    declare("prompt", S_prompt);
#endif
    declare("protocol", S_protocol);
#ifdef UNIX_LOGIN
    declare("reject", S_reject);
#endif
#ifdef TAC_SMB_AUTH
    declare("smb", S_smb);
    declare("primary", S_smbprimary);
    declare("backup", S_smbbackup);
    declare("domain", S_smbdomain);
#endif
#ifdef TAC_LDAP_AUTH
    declare("ldap", S_ldap);
    declare("server", S_ldapserver);
    declare("basedn", S_ldapbasedn);
    declare("binddn", S_ldapbinddn);
    declare("bindpw", S_ldapbindpw);
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_SASL)
    declare("saslmech", S_ldapsaslmech);
    declare("saslcred", S_ldapsaslcred);
#endif
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_NS_SSL)
    declare("sslpath", S_ldapsslpath);
#endif
#endif
#ifdef TAC_ROAMING
    declare("tacacs+", S_tacacs);
    declare("tacaddr", S_traddr);
#endif
#ifdef RAD_ROAMING
    declare("radius", S_radius);
    declare("authhost1", S_authhost1);
    declare("authhost2", S_authhost2);
    declare("accthost1", S_accthost1);
    declare("accthost2", S_accthost2);
    declare("secret", S_secret);
#endif
#ifdef TAC_DCE_AUTH
    declare("dce", S_dce);
#endif
    declare("skey", S_skey);
    declare("slip", S_slip);
    declare("service", S_svc);
    declare("user", S_user);
#if defined(AUTO_MEMBERS)
    declare("users", S_users);
#if defined(AM_REGEXP)
    declare("new-user", S_newuser);
#endif
#endif
#ifdef CRON
    declare("crondeny", S_crondeny);
#endif
#ifdef DENY
    declare("admdeny", S_admdeny);
#endif
#ifdef UDB
    declare("db", S_db);
#if defined(AUTO_MEMBERS) && defined(AM_REGEXP) && defined(MORE_SUBST)
    declare("db-user", S_db_user);
#endif
#endif
#ifdef DYNA
    declare("dynip-server", S_dynip);
    declare("address-pool", S_apool);
#endif
#ifdef ACL
    declare("tty", S_tty);
    declare("nas-group", S_nasgrp);
    declare("from", S_from);
    declare("time", S_time);
    declare("access-group", S_agroup);
#ifdef DXS1
    declare("dxs1", S_dxs1);
#endif
#endif
#ifdef EXTERN_AUTHEN
    declare("extern", S_extern);
#endif
}

/* Return a keyword code if a keyword is recognized. 0 otherwise */
int
keycode(keyword)
char *keyword;
{
    KEYWORD *k = hash_lookup(wordtable, keyword);

    if (k)
	return (k->value);
    return (S_unknown);
}

char *
codestring(type)
int type;
{
    switch (type) {
    default:
	return ("<unknown symbol>");
    case S_eof:
	return ("end-of-file");
    case S_unknown:
	return ("unknown");
    case S_separator:
	return ("=");
    case S_string:
	return ("<string>");
    case S_openbra:
	return ("{");
    case S_closebra:
	return ("}");
    case S_key:
	return ("key");
    case S_nas:
	return ("nas");
    case S_nasaddr:
	return ("nasaddr");
    case S_rshuser:
	return ("rsh-user");
#ifdef ENABLE_SNMP
    case S_snmpcomm:
	return ("snmp-community");
#endif
    case S_user:
	return ("user");
    case S_group:
	return ("group");
    case S_host:
	return ("host");
    case S_file:
	return ("file");
    case S_skey:
	return ("skey");
    case S_name:
	return ("name");
#if defined(NONE_LOGIN) || defined(PERUSER_ENABLE)
    case S_none:
	return ("none");
    case S_nopasswd:
	return("nopassword");
#endif
    case S_login:
	return ("login");
#ifdef EMESSAGE
    case S_emessage:
        return("emessage");
#endif
#ifdef PERUSER_ENABLE
    case S_enable:
        return("enable");
#endif
    case S_member:
	return ("member");
#if defined(AUTO_MEMBERS)
    case S_users:
	return ("users");
#if defined(AM_REGEXP)
    case S_newuser:
	return ("new-user");
#endif
#endif
#ifdef MAXSESS
    case S_maxsess:
	return ("maxsess");
#endif
    case S_expires:
	return ("expires");
    case S_after:
	return ("after");
    case S_before:
	return ("before");
    case S_acct_start:
	return ("acct_start");
    case S_acct_stop:
	return ("acct_stop");
    case S_acct_update:
	return ("acct_update");
    case S_message:
	return ("message");
    case S_arap:
	return ("arap");
    case S_global:
	return ("global");
    case S_chap:
	return ("chap");
#ifdef MSCHAP
    case S_mschap:
	return ("ms-chap");
#endif /* MSCHAP */
    case S_pap:
	return ("pap");
    case S_opap:
	return ("opap");
    case S_cleartext:
	return ("cleartext");
    case S_des:
	return("des");
    case S_svc:
	return ("service");
    case S_default:
	return ("default");
    case S_access:
	return ("access");
    case S_deny:
	return ("deny");
    case S_permit:
	return ("permit");
    case S_exec:
	return ("exec");
#ifdef UNIX_LOGIN
    case S_prompt:
	return ("prompt");
#endif
    case S_protocol:
	return ("protocol");
#ifdef UNIX_LOGIN
    case S_reject:
	return ("reject");
#endif
    case S_optional:
	return ("optional");
    case S_ip:
	return ("ip");
    case S_ipx:
	return ("ipx");
    case S_slip:
	return ("slip");
    case S_ppp:
	return ("ppp");
    case S_authentication:
	return ("authentication");
    case S_authorization:
	return ("authorization");
    case S_cmd:
	return ("cmd");
    case S_attr:
	return ("attribute");
    case S_svc_dflt:
	return ("svc_dflt");
    case S_accounting:
	return ("accounting");
    case S_allow_from:
	return ("allow-from");
    case S_listen_on:
	return ("listen-on");
    case S_lcp:
	return("lcp");
#ifdef TAC_SMB_AUTH
    case S_smb:
	return("smb");
#endif
#ifdef TAC_LDAP_AUTH
    case S_ldap:
	return("ldap");
    case S_ldapserver:
	return("server");
    case S_ldapbasedn:
	return("basedn");
    case S_ldapbinddn:
	return("binddn");
    case S_ldapbindpw:
	return("bindpw");
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_SASL)
    case S_ldapsaslmech:
	return("saslmech");
    case S_ldapsaslcred:
	return("saslcred");
#endif
#if defined(TAC_LDAP_VERSION3) && defined(TAC_LDAP_NS_SSL)
    case S_ldapsslpath:
	return("sslpath");
#endif
#endif
#ifdef TAC_ROAMING
    case S_tacacs:
	return("tacacs+");
    case S_traddr:
	return("tacaddr");
#endif
#ifdef RAD_ROAMING
    case S_radius:
	return("radius");
    case S_authhost1:
	return("authhost1");
    case S_authhost2:
	return("authhost2");
    case S_accthost1:
	return("accthost1");
    case S_accthost2:
	return("accthost2");
    case S_secret:
	return("secret");
#endif
#ifdef TAC_DCE_AUTH
    case S_dce:
	return("dce");
#endif
#ifdef DENY
    case S_admdeny:
	return ("admdeny");
#endif
#ifdef CRON
    case S_crondeny:
	return ("crondeny");
#endif
#ifdef UDB
    case S_db:
	return ("db");
#if defined(AUTO_MEMBERS) && defined(AM_REGEXP) && defined(MORE_SUBST)
    case S_db_user:
	return ("db-user");
#endif
#endif
#ifdef DYNA
    case S_dynip:
        return ("dynip-server");
    case S_apool:
        return ("address-pool");
#endif
#ifdef ACL
    case S_tty:
	return ("tty");
    case S_nasgrp:
	return ("nas-group");
    case S_from:
	return ("from");
    case S_time:
	return ("time");
    case S_agroup:
	return ("access-group");
#ifdef DXS1
    case S_dxs1:
	return ("dxs1");
#endif
#endif
#ifdef EXTERN_AUTHEN
    case S_extern:
	return ("extern");
#endif
    }
}
