/*
   Copyright (c) 1995-2000 by Cisco systems, Inc.

   Permission to use, copy, modify, and distribute modified and
   unmodified copies of this software for any purpose and without fee is
   hereby granted, provided that (a) this copyright and permission notice
   appear on all copies of the software and supporting documentation, (b)
   the name of Cisco Systems, Inc. not be used in advertising or
   publicity pertaining to distribution of the program without specific
   prior permission, and (c) notice be given in supporting documentation
   that use, modification, copying and distribution is by permission of
   Cisco Systems, Inc.

   Cisco Systems, Inc. makes no representations about the suitability
   of this software for any purpose.  THIS SOFTWARE IS PROVIDED ``AS
   IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
   WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS FOR A PARTICULAR PURPOSE.
*/

/* Routines to fork children and communicate with them via pipes */

#include "tac_plus.h"
#include <ctype.h>
#include <signal.h>

#if defined(AUTO_MEMBERS) && defined(AM_REGEXP)
static char *lookup_regex(char *sym, struct identity *id) {
  int num = atoi(sym);
  if ((num < NMATCH) && (id->pmatch[num].rm_so >= 0)) {
    char *s, c = *(id->username+id->pmatch[num].rm_eo);
    *(id->username+id->pmatch[num].rm_eo) = 0;
    s = tac_strdup(id->username+id->pmatch[num].rm_so);
    *(id->username+id->pmatch[num].rm_eo) = c;
    return s;
  }
  return NULL;
}
#if defined(MORE_SUBST) && defined(UDB)
static char *lookup3(char *sym, struct identity *id) {
  if (isdigit((int)(*sym))) {
    char *tmp = lookup_regex(sym, id);
    if (tmp) return tmp;
  }
  return (tac_strdup("unknown"));
}
#endif
#endif

/* Support for dollar variables.  Look in the authorization data and
return strings representing values found there.  If not found, return
"unknown". Recognized strings and their interpolated value types are:

user    -- user name
name    -- NAS name
dyna    -- dynamic address (only when DYNA defined)
port    -- NAS port
address -- NAC address (remote user location)
priv    -- privilege level (0 to 15)
method  -- (1 to 4)
type    -- (1 to 4)
service -- (1 to 7)
status  -- (pass, fail, error, unknown) */

static char *
lookup(sym, data)
char *sym;
struct author_data *data;
{
    static char buf[32];

    if (STREQ(sym, "user")) {
	return (tac_strdup(data->id->username));
    }
    if (STREQ(sym, "name")) {
	return (tac_strdup(data->id->NAS_name));
    }
#ifdef DYNA
    if (STREQ(sym, "dyna")) {
	char nasinfo[32];
	strcpy(nasinfo, inet_ntoa(session.nasaddr));
	strcat(nasinfo, "/");
	strncat(nasinfo, data->id->NAS_port, 16);
	return ((char *) dyna_req(data->id->username, NULL, nasinfo, 0));
    }
#endif
#ifdef ALIASES
    if (STREQ(sym, "alias")) {
	return (tac_strdup(data->id->NAS_alias));
    }
#endif
#ifdef TTYS
    if (STREQ(sym, "palias")) {
	return (tac_strdup(data->id->NAS_palias));
    }
#endif
    if (STREQ(sym, "port")) {
	return (tac_strdup(data->id->NAS_port));
    }
    if (STREQ(sym, "address")) {
	return (tac_strdup(data->id->NAC_address));
    }
    if (STREQ(sym, "priv")) {
	sprintf(buf, "%d", data->id->priv_lvl);
	return (tac_strdup(buf));
    }
    if (STREQ(sym, "method")) {
	sprintf(buf, "%d", data->authen_method);
	return (tac_strdup(buf));
    }
    if (STREQ(sym, "type")) {
	sprintf(buf, "%d", data->authen_type);
	return (tac_strdup(buf));
    }
    if (STREQ(sym, "service")) {
	sprintf(buf, "%d", data->service);
	return (tac_strdup(buf));
    }
    if (STREQ(sym, "status")) {
	switch (data->status) {
	default:
	    return (tac_strdup("unknown"));
	case AUTHOR_STATUS_PASS_ADD:
	case AUTHOR_STATUS_PASS_REPL:
	    return (tac_strdup("pass"));
	case AUTHOR_STATUS_FAIL:
	    return (tac_strdup("fail"));
	case AUTHOR_STATUS_ERROR:
	    return (tac_strdup("error"));
	}
    }
#if defined(AUTO_MEMBERS) && defined(AM_REGEXP)
    if (isdigit((int)(*sym))) {
      char *tmp = lookup_regex(sym, data->id);
      if (tmp) return tmp;
    }
#endif
    return (tac_strdup("unknown"));
}

#ifdef MORE_SUBST

/* Support for dollar variables.  Look in the authentication data and
return strings representing values found there.  If not found, return
"unknown". Recognized strings and their interpolated value types are:

user    -- user name
name    -- NAS name
port    -- NAS port
address -- NAC address (remote user location)
priv    -- privilege level (0 to 15)
method  -- action (1 to 4)
type    -- (1 to 4)
service -- (1 to 8)
status  -- (pass, fail, get, restart, error, follow, unknown) */

static char *lookup2(char *sym, struct authen_data *data) {
    char buf[32];

    if (STREQ(sym, "user"))
	return (tac_strdup(data->NAS_id->username));
    if (STREQ(sym, "name"))
	return (tac_strdup(data->NAS_id->NAS_name));
#ifdef ALIASES
    if (STREQ(sym, "alias"))
	return (tac_strdup(data->NAS_id->NAS_alias));
#endif
    if (STREQ(sym, "port"))
	return (tac_strdup(data->NAS_id->NAS_port));
#ifdef TTYS
    if (STREQ(sym, "palias"))
	return (tac_strdup(data->NAS_id->NAS_palias));
#endif
    if (STREQ(sym, "address"))
	return (tac_strdup(data->NAS_id->NAC_address));
    if (STREQ(sym, "priv")) {
	sprintf(buf, "%d", data->NAS_id->priv_lvl);
	return (tac_strdup(buf));
    }
    if (STREQ(sym, "method")) {
	sprintf(buf, "%d", data->action);
	return (tac_strdup(buf));
    }
    if (STREQ(sym, "type")) {
	sprintf(buf, "%d", data->type);
	return (tac_strdup(buf));
    }
    if (STREQ(sym, "service")) {
	sprintf(buf, "%d", data->service);
	return (tac_strdup(buf));
    }
    if (STREQ(sym, "status"))
	switch (data->status) {
	default:
	    return (tac_strdup("unknown"));
	case TAC_PLUS_AUTHEN_STATUS_PASS:
	    return (tac_strdup("pass"));
	case TAC_PLUS_AUTHEN_STATUS_FAIL:
	    return (tac_strdup("fail"));
	case TAC_PLUS_AUTHEN_STATUS_GETDATA:
	case TAC_PLUS_AUTHEN_STATUS_GETUSER:
	case TAC_PLUS_AUTHEN_STATUS_GETPASS:
	    return (tac_strdup("get"));
	case TAC_PLUS_AUTHEN_STATUS_RESTART:
	    return (tac_strdup("restart"));
	case TAC_PLUS_AUTHEN_STATUS_ERROR:
	    return (tac_strdup("error"));
	case TAC_PLUS_AUTHEN_STATUS_FOLLOW:
	    return (tac_strdup("follow"));
	}
#if defined(AUTO_MEMBERS) && defined(AM_REGEXP)
    if (isdigit((int)(*sym))) {
      char *tmp = lookup_regex(sym, data->NAS_id);
      if (tmp) return tmp;
    }
#endif
    return (tac_strdup("unknown"));
}

#endif

/* Interpolate values of dollar variables into a string.  Determine
   values for the various $ variables by looking in the authorization
   data */

#ifdef MORE_SUBST
char *
substitute3(string, data, mode)
char *string;
struct author_data *data;
int mode;
#else
static char *
substitute(string, data)
char *string;
struct author_data *data;
#endif
{
    char *cp;
    char out[MAX_INPUT_LINE_LEN], *outp;
    char sym[MAX_INPUT_LINE_LEN], *symp;
    char *value, *valuep;

    if (debug & DEBUG_AUTHOR_FLAG)
	report(LOG_DEBUG, "substitute: %s", string);

    cp = string;
    outp = out;

    while (*cp) {
	if (*cp != DOLLARSIGN) {
	    *outp++ = *cp++;
	    continue;
	}
	cp++;			/* skip dollar sign */
	symp = sym;

	/* does it have curly braces e.g. ${foo} ? */
	if (*cp == '{') {
	    cp++;		/* skip { */
	    while (*cp && *cp != '}')
		*symp++ = *cp++;
	    cp++;		/* skip } */

	} else {
	    /* copy symbol into sym */
#if defined(AUTO_MEMBERS) && defined(AM_REGEXP)
	  if (*cp && isdigit((int)*cp))
	    *symp++ = *cp++;
	  else
#endif
	    while (*cp && isalpha(*cp))
		*symp++ = *cp++;
	}

	*symp = '\0';
	/* lookup value */

	if (debug & DEBUG_SUBST_FLAG)
	    report(LOG_DEBUG, "Lookup %s", sym);

#ifdef MORE_SUBST
	switch (mode) {
	case 0:
	  value = lookup2(sym, (struct authen_data *)data);
	  break;
	case 1:
	  value = lookup(sym, (struct author_data *)data);
	  break;
#if defined(AUTO_MEMBERS) && defined(AM_REGEXP) && defined(UDB)
	case 2:
	  value = lookup3(sym, (struct identity *)data);
	  break;
#endif
	}
	valuep = value;
#else
        valuep = value = lookup(sym, data);
#endif
	if (debug & DEBUG_SUBST_FLAG)
	    report(LOG_DEBUG, "Expands to: %s", value);

	/* copy value into output */
	while (valuep && *valuep)
	    *outp++ = *valuep++;
	free(value);
    }
    *outp++ = '\0';

    if (debug & DEBUG_AUTHOR_FLAG)
	report(LOG_DEBUG, "Dollar substitution: %s", out);

    return (tac_strdup(out));
}

/* Wait for a (child) pid to terminate. Return its status. Probably
   horribly implementation dependent. */

static int
waitfor(pid)
int pid;
{
    int ret;
    int status;

    ret = waitpid(pid, &status, 0);

    if (ret < 0) {
	report(LOG_ERR, "%s: pid %d no child exists", session.peer, pid);
	return (-1);
    }
    if (!WIFEXITED(status)) {
	report(LOG_ERR, "%s: pid %d child in illegal state", session.peer, pid);
	return (-1);
    }
    if (debug & DEBUG_AUTHOR_FLAG)
	report(LOG_DEBUG, "pid %d child exited status %d",
	       pid, WEXITSTATUS(status));

    return (WEXITSTATUS(status));
}

/* Write an argv array of strings to fd, adding a newline to each one */
static int
write_args(fd, args, arg_cnt)
int fd, arg_cnt;
char **args;
{
    int i, m;

    for (i = 0; i < arg_cnt; i++) {
	int n = strlen(args[i]);

	m = write(fd, args[i], n);
	m += write(fd, "\n", 1);

	if (m != (n + 1)) {
	    report(LOG_ERR, "%s: Process write failure", session.peer);
	    return (-1);
	}
    }
    return (0);
}

/* Close the three given file-descruptors */
static void
close_fds(fd1, fd2, fd3)
 int fd1, fd2, fd3;
{
    if (fd1 >= 0) {
	close(fd1);
    }
    if (fd2 >= 0) {
	close(fd2);
    }
    if (fd3 >= 0) {
	close(fd3);
    }
}

/* Fork a command. Return read and write file descriptors in readfdp
   and writefdp. Return the pid or -1 if unsuccessful */

static int
my_popen(cmd, readfdp, writefdp, errorfdp)
char *cmd;
int *readfdp, *writefdp, *errorfdp;
{
    int fd1[2], fd2[2], fd3[2];
    int pid;

    fd1[0] = fd1[1] = fd2[0] = fd2[1] = fd3[0] = fd3[1] = -1;
    *readfdp = *writefdp = *errorfdp = -1;
    
    if (pipe(fd1) < 0 || pipe(fd2) < 0 || pipe(fd3) < 0) {
	report(LOG_ERR, "%s: Cannot create pipes", session.peer);
	close_fds(fd1[0], fd2[0], fd3[0]);
	close_fds(fd1[1], fd2[1], fd3[1]);
	return (-1);
    }

    /* The parent who forked us is set to reap all children
       automatically. We disable this so we can explicitly reap our
       children to read their status */

    signal(SIGCHLD, SIG_DFL);

    pid = fork();

    if (pid < 0) {
	report(LOG_ERR, "%s: fork failure", session.peer);
	close_fds(fd1[0], fd2[0], fd3[0]);
	close_fds(fd1[1], fd2[1], fd3[1]);
	return (-1);
    }
    if (pid > 0) {
	/* parent */
	close_fds(fd1[0], fd2[1], fd3[1]);

	*writefdp = fd1[1];
	*readfdp = fd2[0];
	*errorfdp = fd3[0];

	return (pid);
    }
    /* child */
    closelog();
    close(session.sock);
    close_fds(fd1[1], fd2[0], fd3[0]);

    if (fd1[0] != STDIN_FILENO) {
	if (dup2(fd1[0], STDIN_FILENO) < 0)
	    exit(-1);
	close(fd1[0]);
    }
    if (fd2[1] != STDOUT_FILENO) {
	if (dup2(fd2[1], STDOUT_FILENO) < 0)
	    exit(-1);
	close(fd2[1]);
    }
    if (fd3[1] != STDERR_FILENO) {
	if (dup2(fd3[1], STDERR_FILENO) < 0)
	    exit(-1);
	close(fd3[1]);
    }
    (void) execl("/bin/sh", "sh", "-c", cmd, (char *) NULL);
    _exit(-1);
    return(0); /* keep Codecenter quiet */
}

/* read the file descriptor and stuff the data into the given array for
 * the number of bytes given. Throw the rest away.
 */
static int
read_string (fd, string, len)
int fd, len;
char *string;
{
    uint i, ret;
    char c;
    
    i=0;
    do {
	ret = read(fd, &c, 1);
	if ( (ret > 0) && ((i+1)<len) ) {
	    string[i++] = c;
	    string[i] = '\0';
	}
    } while ((i<len) && (ret>0));
    return(ret);
}

/* Read lines from fd and place them into an argv style array. Highly
   recursive so we don't have to count lines in advance. Uses "n" as
   the count of lines seen so far. When eof is read, the array is
   allocated, and the recursion unravels */

static char **
read_args(n, fd)
int n, fd;
{
    char buf[255], *bufp, c, **out;

    bufp = buf;

    while (read(fd, &c, 1) > 0) {
	if (c != '\n') {
	    *bufp++ = c;
	    continue;
	}
	*bufp = '\0';
	out = read_args(n + 1, fd);
	out[n] = (char *) tac_malloc(strlen(buf) + 1);
	strcpy(out[n], buf);
	return (out);
    }
    /* eof */
    out = (char **) tac_malloc(sizeof(char *) * (n + 1));
    out[n] = NULL;

    return (out);
}


/* Do variable interpolation on a string, then invoke it as a shell
   command. Write an appropriate set of AV pairs to the command's
   standard input and read its standard output into outarray. Return
   the commands final status when it terminates */

int
call_pre_process(string, data, outargsp, outargs_cntp, error, err_len)
char *string, *error;
struct author_data *data;
char ***outargsp;
int *outargs_cntp, err_len;
{
    char **new_args;
    int readfd, writefd, errorfd;
    int status, i;
    char *cmd = substitute(string, data);
    int pid = my_popen(cmd, &readfd, &writefd, &errorfd);

    memset(error, '\0', err_len);
    
    free(cmd);

    if (pid < 0) {
	close_fds(readfd, writefd, errorfd);
	return (1);		/* deny */
    }

    for (i = 0; i < data->num_in_args; i++) {
	if (debug & DEBUG_AUTHOR_FLAG)
	    report(LOG_DEBUG, "input %s", data->input_args[i]);
    }

    if (write_args(writefd, data->input_args, data->num_in_args)) {
	close_fds(readfd, writefd, errorfd);
	return (1);		/* deny */
    }

    close(writefd);
    writefd = -1;

    new_args = read_args(0, readfd);
    *outargsp = new_args;

    if (debug & DEBUG_AUTHOR_FLAG) {
	for (i = 0; new_args[i]; i++) {
	    report(LOG_DEBUG, "output %s", new_args[i]);
	}
    }

    read_string(errorfd, error, err_len);
    if (error[0] != '\0') {
	report(LOG_ERR, "Error from program (%d): \"%s\" ",
	       strlen(error), error);
    }
	
    /* count the args */
    for (i = 0; new_args[i]; i++)
	 /* NULL stmt */ ;

    *outargs_cntp = i;

    status = waitfor(pid);
    close_fds(readfd, writefd, errorfd);
    return (status);
}

/* Do variable interpolation on a string, then invoke it as a shell
   command. Write an appropriate set of AV pairs to the command's
   standard input and read its standard output into outarray. Return
   the commands final status when it terminates */

int
call_post_process(string, data, outargsp, outargs_cntp)
char *string;
struct author_data *data;
char ***outargsp;
int *outargs_cntp;
{
    char **new_args;
    int status;
    int readfd, writefd, errorfd;
    int i;
    char *cmd = substitute(string, data);
    int pid = my_popen(cmd, &readfd, &writefd, &errorfd);

    free(cmd);

    if (pid < 0) {
	close_fds(readfd, writefd, errorfd);
	return (1);		/* deny */
    }

    /* If the status is AUTHOR_STATUS_PASS_ADD then the current output args
     * represent *additions* to the input args, not the full set */

    if (data->status == AUTHOR_STATUS_PASS_ADD) {

	for (i = 0; i < data->num_in_args; i++) {
	    if (debug & DEBUG_AUTHOR_FLAG)
		report(LOG_DEBUG, "input %s", data->input_args[i]);
	}

	if (write_args(writefd, data->input_args, data->num_in_args)) {
	    close_fds(readfd, writefd, errorfd);
	    return (1);		/* deny */
	}
    }
    for (i = 0; i < data->num_out_args; i++) {
	if (debug & DEBUG_AUTHOR_FLAG)
	    report(LOG_DEBUG, "input %s", data->output_args[i]);
    }

    if (write_args(writefd, data->output_args, data->num_out_args)) {
	close_fds(readfd, writefd, errorfd);
	return (1);		/* deny */
    }

    close(writefd);
    writefd = -1;

    new_args = read_args(0, readfd);
    *outargsp = new_args;

    if (debug & DEBUG_AUTHOR_FLAG) {
	for (i = 0; new_args[i]; i++) {
	    report(LOG_DEBUG, "output %s", new_args[i]);
	}
    }
    /* count the output args */
    for (i = 0; new_args[i]; i++)
	 /* NULL stmt */ ;

    *outargs_cntp = i;

    status = waitfor(pid);
    close_fds(readfd, writefd, errorfd);
    return (status);
}

#ifdef TAC_ASSEMBLE_MSG
/* Scan outarray generated by call_pre_process() or call_post_process()
   function, assemble message from appropriate set of AV pairs (msg=) and
   clean up these pairs from outarray. Return assembled message.
   Written by Vyacheslav Silakov <seal@nordlink.ru> */

char *assemble_msg(char ***outargsp, int *outargs_cntp) {
    char **new_args, **msg_args;
    char newargs_cnt = 0;
    char msgargs_cnt = 0;
    int  msg_size = 0;
    char *msg_mark = "msg=";
    char i;
    char *msg, *arg;

    if (*outargs_cntp <= 0)
	return (NULL);

    msg_args = (char **) tac_malloc (sizeof(char *) * (*outargs_cntp));
    for (i = 0; i < *outargs_cntp; i++) {
	for (arg = (*outargsp)[i]; *arg && isspace(*arg); arg++)
	    ;
	if (strncmp (arg, msg_mark, strlen (msg_mark)) == 0) {
	    msg_args[i] = arg + strlen (msg_mark);
	    msgargs_cnt++;
	    msg_size += strlen (msg_args[i]);
	} else
	    msg_args[i] = NULL;
    }

    /* We'll recover new line '\n' characters ignored by read_args()
       function. msgargs_cnt uses as a count of new line characters. */
    msg_size += msgargs_cnt;

    msg = (char *) tac_malloc (sizeof(char) * (msg_size + 1));
    new_args =
      (char **) tac_malloc (sizeof(char *) * (*outargs_cntp - msgargs_cnt));

    *msg = '\0';
    for (i = 0; i < *outargs_cntp; i++) {
	if (msg_args[i]) {
	    strcat (msg, msg_args[i]);
	    strcat (msg, "\n");
	} else {
	    new_args[newargs_cnt] = tac_strdup ((*outargsp)[i]);
	    if (debug & DEBUG_AUTHOR_FLAG)
		report (LOG_DEBUG, "assemble_msg(): rewrited output[%d] %s",
		    newargs_cnt, new_args[newargs_cnt]);
	    newargs_cnt++;
	}
	free ((*outargsp)[i]);
    }

    /* Delete last '\n'. It will be added by NAS itself -:) */
    msg[msg_size - 1] = '\0';

    free (*outargsp);
    free (msg_args);

    if (debug & DEBUG_AUTHOR_FLAG)
	report (LOG_DEBUG, "assemble_msg(): msg=%s", msg);

    *outargs_cntp = newargs_cnt;
    *outargsp = new_args;
    return (msg);
}
#endif

#ifdef EXTERN_AUTHEN
int call_extern_authen(char *string, struct authen_data *data, char *pwd) {
  int status;
  int readfd, writefd, errorfd;
  char *cmd = substitute2(string, (void*)data);
  int pid = my_popen(cmd, &readfd, &writefd, &errorfd);
  int n, m;

  free(cmd);
  if (pid < 0) {
    close_fds(readfd, writefd, errorfd);
    return (0);
  }
  if (pwd) {
    m=strlen(pwd);
    n=write(writefd,pwd,m);
    if (n != m) {
      report(LOG_ERR, "%s: Process write failure", session.peer);
      return 0;
    }
  }
  close_fds(readfd, writefd, errorfd);

  status = waitfor(pid);
  return status;
}
#endif

static int write_fd(int fd, char *s) {
  int n, m;
  if (s) {
    m = strlen(s);
    n = write(fd, s, m);
    if (n != m) {
      report(LOG_ERR, "%s: Process write failure", session.peer);
      return 1;
    }
    if (write(fd, "\n", 1) != 1) {
      report(LOG_ERR, "%s: Process write failure", session.peer);
      return 1;
    }
  }
  return 0;
}

int call_extern_acct(char *string, struct acct_rec *rec) {
  int status, i, err = 0;
  int readfd, writefd, errorfd;
  int pid = my_popen(string, &readfd, &writefd, &errorfd);

  if (pid < 0) {
    close_fds(readfd, writefd, errorfd);
    return (0);
  }
  err += write_fd(writefd, rec->identity->username);
  err += write_fd(writefd, rec->identity->NAS_name);
  err += write_fd(writefd, rec->identity->NAS_port);
  err += write_fd(writefd, rec->identity->NAC_address);
  for (i=0; i < rec->num_args; i++) {
    err += write_fd(writefd, rec->args[i]);
  }
  close_fds(readfd, writefd, errorfd);
  if (err) report(LOG_ERR, "%s: Process write failure", session.peer);

  status = waitfor(pid);
  return status;
}

#if defined(ACL) && defined(DXS1) || defined(REGNAS)
char *call_rsh(char *string) {
  int status;
  int readfd, writefd, errorfd;
  int pid = my_popen(string, &readfd, &writefd, &errorfd);
  char *buf = tac_malloc(10240);
  bzero(buf, 10240);

  if (pid < 0) {
    close_fds(readfd, writefd, errorfd);
    return NULL;
  }
  read(readfd, buf, 10240);
  status = waitfor(pid);
  close_fds(readfd, writefd, errorfd);
  if (status) return NULL; else return buf;
}
#endif
