#include "tac_plus.h"

#if defined(ALIASES) || defined(TTYS)
 #include "own_wtmp.h"
#endif

#ifdef DBF
#ifdef HAVE_DB_185_H
#include <db_185.h>
#else
#include <db.h>
#endif

#ifdef DBF_1
typedef struct {
    long tl;			/* time limit      */
    long tll;			/* time limit left */
    long dl;			/* day  limit      */
    long dll;			/* day  limit left */
    long tot;			/* total time      */
#ifdef PARROT
    double parrot;		/* parrots left    */
    double parrot_credit;
#endif
} TQ_st;
#endif
#endif

#define fill_pw()		\
  if(!pw)			\
  {				\
   pwdummy.pw_name = name;	\
   pwdummy.pw_uid = NOPUID;	\
   pwdummy.pw_gid = NOPGID;	\
   pwdummy.pw_dir = NOPHOME;	\
   pwdummy.pw_shell = NOPSHELL;	\
   pw = &pwdummy;		\
  }


int sys_adm_verify(name, pw, data, recurse)
char *name;
struct passwd *pw;
struct authen_data *data;
int recurse;
{
 struct passwd pwdummy;

 if (data->service == TAC_PLUS_AUTHEN_SVC_LOGIN)
  {
   char buff[4096];

   fill_pw();

#ifdef CRON
   if (set_time_status(name, data, pw, recurse))
     return 0;
#endif

#ifdef DENY
   if (set_sysadm_status(name, data, recurse))
     return 0;
#endif

#ifdef DBF
   if (set_usradm_status(name, data))
     return 0;
#endif

   snprintf(buff, sizeof(buff), "%s:%d:%d:%s:%s",
   pw->pw_name, pw->pw_uid, pw->pw_gid, pw->pw_dir, pw->pw_shell);
   data->server_data = tac_strdup(buff);
   data->server_dlen = strlen(buff)+1;
  }

#if (defined(CRON)  && (defined(CRON_FOR_ENABLE) || defined(CRON_FOR_PPP))) || \
    (defined(DENY)  && (defined(DENY_FOR_ENABLE) || defined(DENY_FOR_PPP))) || \
    (defined(DBF)   && (defined(DBF_FOR_ENABLE)  || defined(DBF_FOR_PPP)))

 else if(data->service == TAC_PLUS_AUTHEN_SVC_ENABLE
#ifdef PERUSER_ENABLE
     || data->service == TAC_PLUS_AUTHEN_SVC_ENUSER
#endif
#if defined(CRON_FOR_PPP) || defined(DENY_FOR_PPP) || defined(DBF_FOR_PPP)
     || data->service == TAC_PLUS_AUTHEN_SVC_PPP
#endif
        )
  {

#if defined(CRON) && (defined(CRON_FOR_ENABLE) || defined(CRON_FOR_PPP))
   fill_pw();
   if (set_time_status(name, data, pw, recurse))
     return 0;
#endif

#if defined(DENY) && (defined(DENY_FOR_ENABLE) || defined(DENY_FOR_PPP))
   if (set_sysadm_status(name, data, recurse))
     return 0;
#endif

#if defined(DBF) && (defined(DBF_FOR_ENABLE) || defined(DBF_FOR_PPP))
   if (set_usradm_status(name, data))
     return 0;
#endif

  }
#endif /* ... || ... */

  return (data->status == TAC_PLUS_AUTHEN_STATUS_PASS);
}


#ifdef DBF

#ifdef DBF_1
extern int find_ug(DB *, TQ_st *, char *, int (*)(TQ_st *), DBT *, char **);
extern int find_ug_tl_cond(TQ_st *);
extern int find_ug_dl_cond(TQ_st *);
#ifdef PARROT
extern int find_ug_pl_cond(TQ_st *);
static double parrot_off=PARROT_OFF; /* avoid rounding errors */
#endif
#endif

#ifndef MAX_MSG_LEN
#define MAX_MSG_LEN 256
#endif

int set_usradm_status(u, data)
char *u;
struct authen_data *data;
{
  static DB      *db;
  static DBT     key[1], val[1];

#ifdef DBF_1
  static TQ_st tq;
  int res;
  static char s2[MAX_MSG_LEN];
  char *gname;
#endif

  int retval = 0;
  char *s = tac_malloc(MAX_MSG_LEN);
  bzero(s, MAX_MSG_LEN);
  s2[0] = 0;

  if (debug & DEBUG_AUTHEN_FLAG)
	report(LOG_DEBUG, "set_usradm_status: name=%s", u);

  db = dbopen(TAC_PLUS_DBFILE, O_RDONLY, 0600, DB_HASH, NULL);
  if (!db) {
   if (debug & DEBUG_AUTHEN_FLAG)
     report(LOG_DEBUG, "set_usradm_status: no database file");
   return 0;
  }
  if (!tac_lockfd(TAC_PLUS_DBFILE, (*db->fd)(db), 0)) {
    report(LOG_ERR, "unable to lock user database %s: %s", TAC_PLUS_DBFILE, strerror(errno));
    return 0;
  }

#ifdef DBF_1
  res = find_ug(db, &tq, u, find_ug_tl_cond, key, &gname);
  if (res < 2 && !find_ug_tl_cond(&tq) || res > 1 && res != 4) {
    snprintf(s+strlen(s), MAX_MSG_LEN-strlen(s), "\n*** User %s: total time limit EXCEEDED!", u);
    if (debug & DEBUG_AUTHEN_FLAG) report(LOG_INFO, "user %s on line %s nas %s total time limit reached", u, data->NAS_id->NAS_port, data->NAS_id->NAS_name);
    goto e1;
  } else if (res != 4 && (tq.tl != TIME_LIMIT_OFF)) {
    if (tq.tl != TIME_LIMIT_NO)
      snprintf(s2+strlen(s2), MAX_MSG_LEN-strlen(s2), "\n*** total time left: %ld minutes", tq.tll / 60);
    else
      snprintf(s2+strlen(s2), MAX_MSG_LEN-strlen(s2), "\n*** total time left: unlimited");
    if (gname) snprintf(s2+strlen(s2), MAX_MSG_LEN-strlen(s2), " (in group %s)", gname);
  }

  res = find_ug(db, &tq, u, find_ug_dl_cond, key, &gname);
  if (res < 2 && !find_ug_dl_cond(&tq) || res > 1 && res != 4) {
    snprintf(s+strlen(s), MAX_MSG_LEN-strlen(s), "\n*** User %s: daily time limit EXCEEDED!", u);
    if (debug & DEBUG_AUTHEN_FLAG) report(LOG_INFO, "user %s on line %s nas %s day time limit reached", u, data->NAS_id->NAS_port, data->NAS_id->NAS_name);
    goto e1;
  } else if (res != 4 && (tq.dl != TIME_LIMIT_OFF)) {
    if (tq.dl != TIME_LIMIT_NO)
      snprintf(s2+strlen(s2), MAX_MSG_LEN-strlen(s2), "\n*** left for today: %ld minutes", tq.dll / 60);
    else
      snprintf(s2+strlen(s2), MAX_MSG_LEN-strlen(s2), "\n*** left for today: unlimited");
    if (gname) snprintf(s2+strlen(s2), MAX_MSG_LEN-strlen(s2), " (in group %s)", gname);
  }

#ifdef PARROT
  res = find_ug(db, &tq, u, find_ug_pl_cond, key, &gname);
  if (res < 2 && !find_ug_pl_cond(&tq) || res > 1 && res != 4) {
    snprintf(s+strlen(s), MAX_MSG_LEN-strlen(s), "\n*** User %s: parrot limit EXCEEDED!", u);
    if (debug & DEBUG_AUTHEN_FLAG) report(LOG_INFO, "user %s on line %s nas %s parrot limit reached", u, data->NAS_id->NAS_port, data->NAS_id->NAS_name);
    goto e1;
  } else if (res != 4 && tq.parrot != parrot_off) {
    snprintf(s2+strlen(s2), MAX_MSG_LEN-strlen(s2), "\n*** parrots left: %8.4f", tq.parrot);
    if (gname) snprintf(s2+strlen(s2), MAX_MSG_LEN-strlen(s2), " (in group %s)", gname);
  }
#endif

  if (!s2[0]) goto e2;
  sprintf(s, "\n*** User %s:%s", u, s2);

#else /* DBF_1 */
  key->data = u;
  key->size = strlen(u) + 1;
  if ((*db->get) (db, key, val, NULL)) {
    /* report(LOG_INFO, "set_usradm_status: can't get key value for %s", u); */
#ifdef DBF_STRICT
    goto e1;
#endif
  }

  if (!val->data) {
#ifdef DBF_SILENT
    goto e2;
#else
    snprintf(s+strlen(s), MAX_MSG_LEN-strlen(s), "\n*** User %s: time is unlimited.", u);
#endif
  } else {
    if (!*(time_t *)val->data) {
      snprintf(s+strlen(s), MAX_MSG_LEN-strlen(s), "\n*** User %s: time limit EXCEEDED!", u);
      goto e1;
    }
    snprintf(s+strlen(s), MAX_MSG_LEN-strlen(s), "\n*** User %s: time left %d minutes.",
     	     u, *(time_t *)val->data / 60);
  }
#endif /* DBF_1 */

  goto e0;

e1:
  data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
  retval = 1;

e0:
  fr_fi(data->server_msg, s);

  /* #ifdef DBF_SILENT */
e2:
  /* #endif */
  (*db->close) (db);
  if((debug & DEBUG_AUTHEN_FLAG) && retval)
    report(LOG_INFO, "user %s login not allowed by db rules", u);
  return retval;
}

#endif


#ifdef ALIASES

static char *nextvalid(p)
char **p;
{
  char	*ppp;

  while(**p && ((**p == ' ') || (**p == '\t'))) (*p)++;
  if(!*(ppp = *p) || *ppp == '#')
    return NULL;

  while(**p && (**p != ' ') && (**p != '\t')) (*p)++;
  if(**p)
   {
    **p = '\0';
    (*p)++;
   }

  return ppp;
}

static char *
value(p)
	char *p;
{

	return ((p = strchr(p, '=')) ? ++p : NULL);
}

char *get_peer_alias(peer, flags)
char *peer;
int *flags;
{
FILE	*f;
char  tmp[1024],
      *p;
char  *pp,
      *al,
      *s;
int tflag;
size_t u;

if ((f = fopen(TAC_PLUS_ALIAS_FILE, "r")) != NULL)
 {
  while(fgets(tmp, sizeof(tmp), f))
    {
      u = strlen(tmp);
      if(u && tmp[u - 1] == '\n')
        tmp[u - 1] = '\0';

      p = tmp;
      if((al = nextvalid(&p)) == NULL)	/* alias	*/
        continue;

      tflag = B_FLAG_INIT;			/* init flag	*/

#define	vcmp(e)	(!strncmp(pp, e, sizeof(e) - 1) && pp[sizeof(e) - 1] == '=')
#define mcmp(e) (!strncmp(s, e, sizeof(e) - 1))

					/* check flags		*/
      while((pp = nextvalid(&p)) != NULL) {

                if(!(s = value(pp)))	/* no value - no keyword */
                  break;

		     if(vcmp(_TTYS_CHECK))
		 {
		       if(mcmp(_TTYS_YES))
		         tflag |= A_FLAG_CHECK;
		  else if(mcmp(_TTYS_NO))
		         tflag &= ~A_FLAG_CHECK;
		 }
		else if(vcmp(_TTYS_PCHECK))
		 {
		       if(mcmp(_TTYS_YES))
		         tflag |= A_FLAG_PCHECK;
		  else if(mcmp(_TTYS_NO))
		         tflag &= ~A_FLAG_PCHECK;
		 }
		else if(vcmp(_TTYS_TTYS))
		 {
		       if(mcmp(_TTYS_YES))
		         tflag |= A_FLAG_TTYS;
		  else if(mcmp(_TTYS_NO))
		         tflag &= ~A_FLAG_TTYS;
		 }
		else if(vcmp(_TTYS_PUTMP) || vcmp(_TTYS_PWTMP))
		 {
		       if(mcmp(_TTYS_ALIAS))
		         tflag |= A_FLAG_PUTMP_AL;
		  else if(mcmp(_TTYS_TRUE))
		         tflag &= ~A_FLAG_PUTMP_AL;
		 }
		else if(vcmp(_TTYS_PACCT))
		 {
		       if(mcmp(_TTYS_ALIAS))
		         tflag |= A_FLAG_PACCT_AL;
		  else if(mcmp(_TTYS_TRUE))
		         tflag &= ~A_FLAG_PACCT_AL;
		 }
		else if(vcmp(_TTYS_UTMP) || vcmp(_TTYS_WTMP))
		 {
		       if(mcmp(_TTYS_ALIAS))
		        {
		         tflag &= ~A_FLAG_UTMP_X;
		         tflag |= A_FLAG_UTMP_AL;
		        }
		  else if(mcmp(_TTYS_TRUE))
		        {
		         tflag &= ~A_FLAG_UTMP_X;
		         tflag |= A_FLAG_UTMP_TR;
		        }
		  else if(mcmp(_TTYS_ASIS))
		        {
		         tflag &= ~A_FLAG_UTMP_X;
		         tflag |= A_FLAG_UTMP_AS;
		        }
		  else if(mcmp(_TTYS_NO))
		         tflag &= ~A_FLAG_UTMP_X;
		 }
		else if(vcmp(_TTYS_ACCT))
		 {
		       if(mcmp(_TTYS_ALIAS))
		        {
		         tflag &= ~A_FLAG_ACCT_X;
		         tflag |= A_FLAG_ACCT_AL;
		        }
		  else if(mcmp(_TTYS_TRUE))
		        {
		         tflag &= ~A_FLAG_ACCT_X;
		         tflag |= A_FLAG_ACCT_TR;
		        }
		  else if(mcmp(_TTYS_ASIS))
		        {
		         tflag &= ~A_FLAG_ACCT_X;
		         tflag |= A_FLAG_ACCT_AS;
		        }
		  else if(mcmp(_TTYS_NO))
		         tflag &= ~A_FLAG_ACCT_X;
		 }
		else
			break;
      }

						/* find peer	*/
      for(; pp; pp = nextvalid(&p))
       {
        if(!strcmp(peer, pp))
         {
          fclose(f);
          *flags = tflag;
          return tac_strdup(al);
         }
       }
    }
  fclose(f);
 }

*flags = B_FLAG_INIT;
return tac_strdup(peer);
}

#endif

#ifdef TTYS

char *get_port_alias(port, host, alias, flags, slot)
char *port, *host, *alias;
int *flags;
int *slot;
{
 char x[UT_LINESIZE + 1];
 char *trueline = x;

 *slot = t_ttyslot(port, host, alias, &trueline, flags);

 x[sizeof(x) - 1] = '\0';

 return tac_strdup(x);
}

#endif
