/* (C) 1997 Mike Shoyher msh@corbina.net, msh@apache.lexa.ru */
/* (C) 1999, 2000 Vladimir Volovich <vvv@vsu.ru> */

#include "tac_plus.h"
#include "tac_client.h"

static int tac_sequence;
static int tac_session_id;
static int tac_maxtry=3;
static int tac_timeout=15;
static int tac_fd;
static char *tac_key;
static char ourhost[128];
static int ourhost_len;
static char *ourtty="roaming";
static int ourtty_len;

static void fill_tac_hdr(struct tac_plus_pak_hdr *hdr) {
  hdr->version=TAC_PLUS_VER_0;
  hdr->type=TAC_PLUS_AUTHEN;
  hdr->seq_no=tac_sequence;
  hdr->flags |= TAC_PLUS_UNENCRYPTED_FLAG;
  hdr->session_id=tac_session_id;
}

static int send_data(void *buf, int buf_len, int fd) {
  fd_set fds;
  int rr;
  int try;
  struct timeval tv;

  FD_ZERO(&fds);
  FD_SET(fd,&fds);
  tv.tv_sec = tac_timeout;
  tv.tv_usec = 0;
  for(try=0;try<tac_maxtry;try++) {
    rr=select(fd+1,NULL,&fds,NULL,&tv);
    if (FD_ISSET(fd,&fds)) {
      if (buf_len!=write(fd,buf,buf_len)) continue;
      return 0;
    }
    break;
  }
  report(LOG_ERR, "send_data: write error");
  return -1;
}

static int read_data(void *buf, int buf_len, int fd) {
  fd_set fds;
  int rr;
  int try;
  struct timeval tv;

  FD_ZERO(&fds);
  FD_SET(fd,&fds);
  tv.tv_sec = tac_timeout;
  tv.tv_usec = 0;
  for(try=0;try<tac_maxtry;try++) {
    rr=select(fd+1,&fds,NULL,NULL,&tv);
    if (FD_ISSET(fd,&fds)) {
      if (buf_len!=read(fd,buf,buf_len)) continue;
      return 0;
    }
    break;
  }
  report(LOG_ERR, "read_data: read error");
  return -1;
}

static int read_reply(void **data) {
  int data_len;
  struct tac_plus_pak_hdr hdr;
  if (read_data(&hdr,TAC_PLUS_HDR_SIZE,tac_fd)==-1) return -1;
  data_len=ntohl(hdr.datalength);
  tac_sequence=hdr.seq_no+1;
  *data=tac_malloc(data_len);
  if (read_data(*data,data_len,tac_fd)==-1) return -1;
  md5_xor(&hdr, *data, tac_key);
  return data_len;
}

static void send_auth_cont(char *msg, int msg_len) {
  struct tac_plus_pak_hdr hdr;
  struct authen_cont ac;
  int datalength;
  void *buf;
  int buf_len;

  buf_len=TAC_PLUS_HDR_SIZE+TAC_AUTHEN_CONT_FIXED_FIELDS_SIZE+msg_len;
  buf=tac_malloc(buf_len);
  fill_tac_hdr(&hdr);
  ac.user_data_len=ac.flags=0;
  ac.user_msg_len=htons(msg_len);
  bcopy((const char*)msg,buf+TAC_PLUS_HDR_SIZE+TAC_AUTHEN_CONT_FIXED_FIELDS_SIZE,msg_len);
  datalength=msg_len+TAC_AUTHEN_CONT_FIXED_FIELDS_SIZE;
  hdr.datalength=htonl(datalength);
  bcopy((const char*)&hdr,buf,TAC_PLUS_HDR_SIZE);
  bcopy((const char*)&ac,buf+TAC_PLUS_HDR_SIZE,TAC_AUTHEN_CONT_FIXED_FIELDS_SIZE);
  md5_xor(buf, buf+TAC_PLUS_HDR_SIZE, tac_key);
  send_data(buf,buf_len,tac_fd);
  free(buf);
}

int init_tac_session(char *host_name, char *port_name, char *key, int timeout) {
  struct sockaddr_in tac_port;
  struct hostent *tac_h;
  struct servent *tac_serv;

  gethostname(ourhost,127);
  ourhost_len=strlen(ourhost);
  ourtty_len=strlen(ourtty);
  srand(time(NULL));
  if (timeout>0) tac_timeout=timeout;
  tac_key = strdup(key);
  tac_session_id=rand();
  tac_sequence=1;
  tac_port.sin_family = AF_INET;
  if (*host_name>='0' && *host_name<='9')
    tac_port.sin_addr.s_addr = inet_addr(host_name);
  else {
    tac_h = gethostbyname(host_name);
    if (!tac_h) {
      report(LOG_ERR, "init_tac_session: cannot resolve host name");
      return -1;
    }
    tac_port.sin_addr = *((struct in_addr *) tac_h->h_addr);
  }
  if (port_name==NULL) port_name="tacacs";
  if (*port_name>='0' && *port_name<='9')
    tac_port.sin_port=htons(atoi(port_name));
  else {
    tac_serv=getservbyname(port_name,"tcp");
    if (tac_serv)
      tac_port.sin_port=tac_serv->s_port;
    else {
      report(LOG_ERR, "init_tac_session: unknown port");
      return -1;
    }
  }
  if((tac_fd = socket(AF_INET, SOCK_STREAM, 0))<0) return -1;
  if (connect(tac_fd, (struct sockaddr *) &tac_port, sizeof tac_port) < 0) {
    report(LOG_ERR, "init_tac_session: connect to %s: %s", host_name, strerror(errno));
    return -1;
  }
  return tac_fd;
}

void deinit_tac_session(void) {
  shutdown(tac_fd,2);
  close(tac_fd);
  if (tac_key) free(tac_key);
}

/* Need to specify the lengths, because CHAP passwords etc may have NULs in them */
int make_authen(char *username, int user_len, char *password, int password_len, int authen_type) {
  struct tac_plus_pak_hdr hdr;
  struct authen_start as;
  int datalength;
  void *buf;
  void *data;
  int data_len;
  int buf_len;
  struct authen_reply *ar;

  fill_tac_hdr(&hdr);

  datalength=TAC_AUTHEN_START_FIXED_FIELDS_SIZE;

  as.action= TAC_PLUS_AUTHEN_LOGIN;
  as.priv_lvl= TAC_PLUS_PRIV_LVL_MIN;
  as.authen_type = authen_type;
  as.service= TAC_PLUS_AUTHEN_SVC_LOGIN;
  as.user_len=as.port_len=as.rem_addr_len=as.data_len=0;

  if (authen_type != TAC_PLUS_AUTHEN_TYPE_ASCII) {
    /* This will be a version 1 request with the username and password in the start */
    hdr.version=TAC_PLUS_VER_1;
    as.user_len = user_len;
    as.data_len = password_len;
  }
  buf=tac_malloc(buf_len=TAC_PLUS_HDR_SIZE+TAC_AUTHEN_START_FIXED_FIELDS_SIZE
	     +ourtty_len+ourhost_len+as.user_len+as.data_len);

  /* Append user name if required */
  bcopy(username,buf+datalength+TAC_PLUS_HDR_SIZE,as.user_len);
  datalength+=(as.user_len);

  bcopy(ourtty,buf+datalength+TAC_PLUS_HDR_SIZE,ourtty_len);
  datalength+=(ourtty_len);
  as.port_len=ourtty_len;

  bcopy(ourhost,buf+datalength+TAC_PLUS_HDR_SIZE,ourhost_len);
  datalength+=(ourhost_len);
  as.rem_addr_len=ourhost_len;

  /* Append password if required */
  bcopy(password,buf+datalength+TAC_PLUS_HDR_SIZE,as.data_len);
  datalength+=(as.data_len);

  hdr.datalength= htonl(datalength);
  bcopy((const char*)&hdr,buf,TAC_PLUS_HDR_SIZE);
  bcopy((const char*)&as,buf+TAC_PLUS_HDR_SIZE,TAC_AUTHEN_START_FIXED_FIELDS_SIZE);
  md5_xor(buf, buf+TAC_PLUS_HDR_SIZE, tac_key);
  send_data(buf,buf_len,tac_fd);
  free(buf);
  while ((data_len=read_reply(&data))!=-1) {
    ar=data;
    ar->msg_len = ntohs(ar->msg_len);
    ar->data_len = ntohs(ar->data_len);

#if 0 /* debug info */
    printf("make_authen: received status=");
    switch (ar->status) {
    case TAC_PLUS_AUTHEN_STATUS_PASS: printf("PASS, "); break;
    case TAC_PLUS_AUTHEN_STATUS_FAIL: printf("FAIL, "); break;
    case TAC_PLUS_AUTHEN_STATUS_GETDATA: printf("GETDATA, "); break;
    case TAC_PLUS_AUTHEN_STATUS_GETUSER: printf("GETUSER, "); break;
    case TAC_PLUS_AUTHEN_STATUS_GETPASS: printf("GETPASS, "); break;
    case TAC_PLUS_AUTHEN_STATUS_RESTART: printf("RESTART, "); break;
    case TAC_PLUS_AUTHEN_STATUS_ERROR: printf("ERROR, "); break;
    case TAC_PLUS_AUTHEN_STATUS_FOLLOW: printf("FOLLOW, "); break;
    default: printf("UNKNOWN (%i), ",ar->status); break;
    }
    printf("flags=%i, msg_len=%i, data_len=%i, msg=\"",ar->flags,ar->msg_len,ar->data_len);
    fwrite((char*)ar+TAC_AUTHEN_REPLY_FIXED_FIELDS_SIZE,1,ar->msg_len,stdout);
    printf("\", data=\"");
    fwrite((char*)ar+TAC_AUTHEN_REPLY_FIXED_FIELDS_SIZE+ar->msg_len,1,ar->data_len,stdout);
    printf("\"\n");
#endif

    switch (ar->status) {
    case TAC_PLUS_AUTHEN_STATUS_GETUSER:
      free(data);
      send_auth_cont(username, user_len);
      break;
    case TAC_PLUS_AUTHEN_STATUS_GETPASS:
      free(data);
      send_auth_cont(password, password_len);
      break;
    case TAC_PLUS_AUTHEN_STATUS_PASS:
      return 1;
    case TAC_PLUS_AUTHEN_STATUS_FAIL:
      report(LOG_WARNING, "make_authen: authentication failed");
      return 0;
    default :
      report(LOG_ERR, "make_authen: protocol error");
      return 0;
    }
  }
  report(LOG_ERR, "make_authen: unknown error");
  return 0;
}

#if 0
int make_author(char *username, int user_len, char *password, int password_len, int authen_type) {
  struct tac_plus_pak_hdr hdr;
  struct author as;
  int datalength;
  void *buf;
  void *data;
  int buf_len;
  int data_len;
  struct author_reply *ar;

  hdr.version=TAC_PLUS_VER_0; /* TAC_PLUS_VER_1 with username and passwd */
  hdr.type=TAC_PLUS_AUTHOR;
  tac_sequence=1;
  hdr.seq_no=tac_sequence;
  hdr.flags |= TAC_PLUS_UNENCRYPTED_FLAG;
  hdr.session_id=tac_session_id;

  datalength=TAC_AUTHOR_REQ_FIXED_FIELDS_SIZE;

  as.authen_method=AUTHEN_METH_TACACSPLUS; /* AUTHEN_METH_NONE, AUTHEN_METH_ENABLE, AUTHEN_METH_TACACSPLUS */
  as.priv_lvl=1; /* was: TAC_PLUS_PRIV_LVL_MIN */
  as.authen_type = authen_type;
  as.service= TAC_PLUS_AUTHEN_SVC_LOGIN;
  as.user_len=user_len;
  as.port_len=ourtty_len;
  as.rem_addr_len=ourhost_len;
  as.arg_cnt=2;

  buf=tac_malloc(buf_len=TAC_PLUS_HDR_SIZE+TAC_AUTHOR_REQ_FIXED_FIELDS_SIZE
	     +ourtty_len+ourhost_len+as.user_len+as.arg_cnt+17);

  /* arg lengths */
  bcopy("\15\4",buf+datalength+TAC_PLUS_HDR_SIZE,as.arg_cnt);
  datalength+=as.arg_cnt;

  /* Append user name if required */
  bcopy(username,buf+datalength+TAC_PLUS_HDR_SIZE,as.user_len);
  datalength+=(as.user_len);

  bcopy(ourtty,buf+datalength+TAC_PLUS_HDR_SIZE,ourtty_len);
  datalength+=(ourtty_len);

  bcopy(ourhost,buf+datalength+TAC_PLUS_HDR_SIZE,ourhost_len);
  datalength+=(ourhost_len);

  /* args */
  bcopy("service=shellcmd*",buf+datalength+TAC_PLUS_HDR_SIZE,17);
  datalength+=17;

  hdr.datalength= htonl(datalength);
  bcopy((const char*)&hdr,buf,TAC_PLUS_HDR_SIZE);
  bcopy((const char*)&as,buf+TAC_PLUS_HDR_SIZE,TAC_AUTHOR_REQ_FIXED_FIELDS_SIZE);
/* dump_tac_auth(buf); */
  md5_xor(buf, buf+TAC_PLUS_HDR_SIZE, tac_key);
  send_data(buf,buf_len,tac_fd);
  free(buf);

  if ((data_len=read_reply(&data))!=-1) {
    int ret=1;
    ar=data;
    ar->msg_len = ntohs(ar->msg_len);
    ar->data_len = ntohs(ar->data_len);

#if 0 /* debug info */
    printf("make_author: received status=");
    switch (ar->status) {
    case AUTHOR_STATUS_PASS_ADD: printf("PASS_ADD, "); break;
    case AUTHOR_STATUS_PASS_REPL: printf("PASS_REPL, "); break;
    case AUTHOR_STATUS_FAIL: printf("FAIL, "); ret=0; break;
    case AUTHOR_STATUS_ERROR: printf("ERROR, "); ret=0; break;
    default: printf("UNKNOWN (%i), ",ar->status); ret=0; break;
    }
    printf("arg_cnt=%i, msg_len=%i, data_len=%i, msg=\"",ar->arg_cnt,ar->msg_len,ar->data_len);
    fwrite((char*)ar+TAC_AUTHOR_REPLY_FIXED_FIELDS_SIZE+ar->arg_cnt,1,ar->msg_len,stdout);
    printf("\", data=\"");
    fwrite((char*)ar+TAC_AUTHOR_REPLY_FIXED_FIELDS_SIZE+ar->arg_cnt+ar->msg_len,1,ar->data_len,stdout);
    printf("\"\n");
#endif

    return ret;
  } else {
    report(LOG_ERR, "make_author: unknown error");
    return 0;
  }
}

int make_acct(char *username, int user_len, int authen_type, int action, char *args, int args_len, char *argl, int nargs) {
  struct tac_plus_pak_hdr hdr;
  struct acct as;
  int datalength;
  void *buf;
  void *data;
  int buf_len;
  int data_len;
  struct acct_reply *ar;

  hdr.version=TAC_PLUS_VER_0; /* TAC_PLUS_VER_1 with username and passwd */
  hdr.type=TAC_PLUS_ACCT;
  tac_sequence=1;
  hdr.seq_no=tac_sequence;
  hdr.flags |= TAC_PLUS_UNENCRYPTED_FLAG;
  hdr.session_id=tac_session_id;

  datalength=TAC_ACCT_REQ_FIXED_FIELDS_SIZE;

  as.flags=action;
  as.authen_method=AUTHEN_METH_TACACSPLUS; /* AUTHEN_METH_NONE, AUTHEN_METH_ENABLE, AUTHEN_METH_TACACSPLUS */
  as.priv_lvl=1; /* was: TAC_PLUS_PRIV_LVL_MIN */
  as.authen_type = authen_type;
  as.authen_service= TAC_PLUS_AUTHEN_SVC_LOGIN;
  as.user_len=user_len;
  as.port_len=ourtty_len;
  as.rem_addr_len=ourhost_len;
  as.arg_cnt=nargs;

  buf=tac_malloc(buf_len=TAC_PLUS_HDR_SIZE+TAC_ACCT_REQ_FIXED_FIELDS_SIZE
	     +ourtty_len+ourhost_len+as.user_len+as.arg_cnt+args_len);

  /* arg lengths */
  bcopy(argl,buf+datalength+TAC_PLUS_HDR_SIZE,as.arg_cnt);
  datalength+=as.arg_cnt;

  /* Append user name if required */
  bcopy(username,buf+datalength+TAC_PLUS_HDR_SIZE,as.user_len);
  datalength+=(as.user_len);

  bcopy(ourtty,buf+datalength+TAC_PLUS_HDR_SIZE,ourtty_len);
  datalength+=(ourtty_len);

  bcopy(ourhost,buf+datalength+TAC_PLUS_HDR_SIZE,ourhost_len);
  datalength+=(ourhost_len);

  /* args */
  bcopy(args,buf+datalength+TAC_PLUS_HDR_SIZE,args_len);
  datalength+=args_len;

  hdr.datalength= htonl(datalength);
  bcopy((const char*)&hdr,buf,TAC_PLUS_HDR_SIZE);
  bcopy((const char*)&as,buf+TAC_PLUS_HDR_SIZE,TAC_ACCT_REQ_FIXED_FIELDS_SIZE);
  md5_xor(buf, buf+TAC_PLUS_HDR_SIZE, tac_key);
  send_data(buf,buf_len,tac_fd);
  free(buf);

  if ((data_len=read_reply(&data))!=-1) {
    ar=data;
    ar->msg_len = ntohs(ar->msg_len);
    ar->data_len = ntohs(ar->data_len);

#if 0 /* debug info */
    printf("make_acct: received msg_len=%i, data_len=%i, status=",ar->msg_len,ar->data_len);
    switch (ar->status) {
    case TAC_PLUS_ACCT_STATUS_SUCCESS: printf("SUCCESS\n"); break;
    case TAC_PLUS_ACCT_STATUS_ERROR: printf("ERROR\n"); break;
    case TAC_PLUS_ACCT_STATUS_FOLLOW: printf("FOLLOW\n"); break;
    default: printf("UNKNOWN (%i)\n",ar->status); break;
    }
#endif

    return 1;
  } else {
    report(LOG_ERR, "make_acct: unknown error");
    return 0;
  }
}
#endif
