/* 
   Copyright (c) 1995-2000 by Cisco systems, Inc.

   Permission to use, copy, modify, and distribute modified and
   unmodified copies of this software for any purpose and without fee is
   hereby granted, provided that (a) this copyright and permission notice
   appear on all copies of the software and supporting documentation, (b)
   the name of Cisco Systems, Inc. not be used in advertising or
   publicity pertaining to distribution of the program without specific
   prior permission, and (c) notice be given in supporting documentation
   that use, modification, copying and distribution is by permission of
   Cisco Systems, Inc.

   Cisco Systems, Inc. makes no representations about the suitability
   of this software for any purpose.  THIS SOFTWARE IS PROVIDED ``AS
   IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
   WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS FOR A PARTICULAR PURPOSE.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define TAC_VERSION "F4.0.4.alpha.ia.096"

#define MD5_LEN           16
#ifdef MSCHAP
#define MSCHAP_DIGEST_LEN 49
#endif /* MSCHAP */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif
#ifdef HAVE_TERMIO_H
#include <termio.h>
#else
#ifdef HAVE_TERMIOS_H
#include <termios.h>
#endif
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include <stdio.h>
#include <errno.h>
#include <pwd.h>
#include <netdb.h>

#ifdef STDLIB_MALLOC
#include <stdlib.h>
#else
#include <malloc.h>
#endif

#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#else
#include <sys/syslog.h>
#endif

#include "../lib/libutil/tac_utmp.h"

#include <unistd.h>
#include <fcntl.h>
#if !defined(O_SYNC) && defined(O_FSYNC)
#define O_SYNC O_FSYNC
#endif

#if STDC_HEADERS
# include <string.h>
#else
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# else
#  define index strchr
#  define rindex strrchr
# endif
char *strchr (), *strrchr ();
# ifndef HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifndef CONFPATH
#define CONFPATH "/etc/tac_plus"
#endif
#ifndef DATAPATH
#define DATAPATH "/var/adm/tac_plus"
#endif
#ifndef RUNPATH
#define RUNPATH "/etc/tac_plus"
#endif
#ifndef LOGPATH
#define LOGPATH DATAPATH
#endif

#ifndef TAC_PLUS_PIDFILE
#define TAC_PLUS_PIDFILE RUNPATH "/tac_plus.pid"
#endif

#ifndef TAC_PLUS_CFGFILE
#define TAC_PLUS_CFGFILE CONFPATH "/tac_plus.conf"
#endif

#ifndef TAC_PLUS_ACCTFILE
#define TAC_PLUS_ACCTFILE DATAPATH "/acctfile"
#endif

/* This is a shared file used to maintain a record of who's on */
#ifndef WHOLOG_DEFAULT
#define WHOLOG_DEFAULT DATAPATH "/tac_who.log"
#endif
extern char *wholog;

#ifndef LOG_FACIL
#define LOG_FACIL LOG_LOCAL6
#endif

#define B_FLAG_PCHECK A_FLAG_NO
#define B_FLAG_PACCT  A_FLAG_NO
#define B_FLAG_PUTMP  A_FLAG_NO
#ifdef TTYS
 #ifdef PALIAS_CHECK
   #undef B_FLAG_PCHECK
   #define B_FLAG_PCHECK A_FLAG_PCHECK
 #endif
 #ifdef PALIAS_ACCT
   #undef B_FLAG_PACCT
   #define B_FLAG_PACCT A_FLAG_PACCT_AL
 #endif
 #if defined(UTMP) && defined(PALIAS_UTMP)
   #undef B_FLAG_PUTMP
   #define B_FLAG_PUTMP A_FLAG_PUTMP_AL
 #endif
#endif

#ifdef DENY
 #ifndef TAC_PLUS_DENYDIR
 #define TAC_PLUS_DENYDIR DATAPATH "/deny"
 #endif
 #ifndef TAC_PLUS_ALLOWDIR
 #define TAC_PLUS_ALLOWDIR DATAPATH "/allow"
 #endif
 #ifndef ADMDENY
 #define ADMDENY "\nSorry, this account has been blocked by administrator"
 #endif
#endif

#ifdef CRON
 #ifndef TAC_PLUS_CRONDIR
 #define TAC_PLUS_CRONDIR DATAPATH "/tabs"
 #endif
 #ifndef CRONDENY
 #define CRONDENY "\nSorry, you are not allowed to log in at this time on this line"
 #endif
#endif

#ifdef PARROT
 #ifndef PARROTDIR
 #define PARROTDIR DATAPATH "/parrot"
 #endif
 #ifndef PARROT_INTERVAL
 #define PARROT_INTERVAL 60
 #endif
#endif

#ifdef DBF
 #ifndef TAC_PLUS_DBFILE
 #define TAC_PLUS_DBFILE DATAPATH "/user.db"
 #endif
#endif

#ifdef UTMP
 #if defined(LOG_EMPTY) && !defined(EMPTY_NAME)
   #define EMPTY_NAME "unknown"
 #endif
#endif

#define B_FLAG_TTYS A_FLAG_NONE
#define B_FLAG_CHECK A_FLAG_NONE
#ifdef ALIASES
 #ifndef TAC_PLUS_ALIAS_FILE
  #define TAC_PLUS_ALIAS_FILE DATAPATH "/aliases"
 #endif
 #ifdef ALIAS_TTYS
   #undef B_FLAG_TTYS
   #define B_FLAG_TTYS A_FLAG_TTYS
 #endif
 #ifdef ALIAS_CHECK
   #undef B_FLAG_CHECK
   #define B_FLAG_CHECK A_FLAG_CHECK
 #endif
 #ifndef ACCT_LOG_MODE
  #define ACCT_LOG_MODE A_FLAG_ACCT_TR
 #endif
 #ifndef UTMP_LOG_MODE
  #define UTMP_LOG_MODE A_FLAG_UTMP_AS
 #endif
#else
 #define UTMP_LOG_MODE A_FLAG_UTMP_TR
 #define ACCT_LOG_MODE A_FLAG_ACCT_TR
#endif

#define B_FLAG_INIT (B_FLAG_PCHECK | B_FLAG_PACCT | B_FLAG_PUTMP | \
                     B_FLAG_TTYS | B_FLAG_CHECK | ACCT_LOG_MODE | UTMP_LOG_MODE)

#ifdef UNIX_LOGIN
 #ifndef UNIX_PROMPT
  #define UNIX_PROMPT "\nUser Access Verification\n\nlogin: "
 #endif
 #ifndef UNIX_REJECT
  #define UNIX_REJECT "Login incorrect"
 #endif
#endif

#if defined(NAMEISREGEXP) && defined(NAM_REGEXP)
 #ifndef ENAB_PREFIX
  #define ENAB_PREFIX "__"
 #endif
#else
 #define ENAB_PREFIX "$"
#endif

#ifndef NOPUID
 #define NOPUID 32767
#endif

#ifndef NOPGID
 #define NOPGID 32766
#endif

#ifndef NOPHOME
 #define NOPHOME ""
#endif

#ifndef NOPSHELL
 #define NOPSHELL ""
#endif

#ifdef TAC_PLUS_USERNAME
#if defined(TAC_PLUS_USERID) || defined(TAC_PLUS_GROUPID)
#error Define either TAC_PLUS_USERNAME or TAC_PLUS_*ID, not both
#endif
extern int tac_plus_userid;
extern int tac_plus_groupid;
#define TAC_PLUS_USERID tac_plus_userid
#define TAC_PLUS_GROUPID tac_plus_groupid
#endif

/* 
 * You probably shouldn't be changing much below this line unless you really
 * know what you are doing.
 */

#define DOLLARSIGN '$'

/*
 * XTACACSP protocol defintions
 */

/*
 * This structure describes an authentication method.
 *   authen_name     contains the name of the authentication method.
 *   authen_func     is a pointer to the authentication function.
 *   authen_method   numeric value of authentication method
 */

#define AUTHEN_NAME_SIZE 128

struct authen_type {
    char authen_name[AUTHEN_NAME_SIZE];
    int (*authen_func)();
    int authen_type;
};

#if defined(AUTO_MEMBERS) && defined(AM_REGEXP) || defined(ACL) || defined(REGNAS)
#include REGEX_H
#endif

/*
 * This structure describes a principal that is to be authenticated.
 *   username        is the principals name (ASCII, null terminated)
 *   NAS_name        is the name of the NAS where the user is
 *   NAS_port        is the port on the NAS where the user is
 *   NAC_address     is the remote user location.  This may be
 *                   a remote IP address or a caller-ID or ...
 *   priv_lvl        user's requested privilege level.
 */

struct identity {
    char *username;
    char *NAS_name;
#ifdef ALIASES
    char *NAS_alias;
#endif
    char *NAS_port;
#ifdef TTYS
    char *NAS_palias;
    int NAS_slot;
#endif
#if defined(ALIASES) || defined(TTYS)
    int NAS_flags;
#endif
    char *NAC_address;
    int priv_lvl;
#if defined(AUTO_MEMBERS) && defined(AM_REGEXP)
#define NMATCH 10  /* maximal number of sub-expressions is regexps */
    regmatch_t pmatch[NMATCH];
    char *newuser;
#endif
};

/*
 * The authen_data structure is the data structure for passing
 * information to and from the authentication function
 * (authen_type.authen_func).
 */

struct authen_data {
    struct identity *NAS_id;	/* user identity */
    char *server_msg;		/* null-terminated output msg */

    int server_dlen;		/* output data length */
    char *server_data;		/* output data */

    char *client_msg;		/* null-terminated input msg a user typed */

    int client_dlen;		/* input data length */
    char *client_data;		/* input data */

    void *method_data;		/* opaque private method data */
    int action;			/* what's to be done */
    int service;		/* calling service */
    int status;			/* Authen status */
    int type;			/* Authen type */
    u_char flags;               /* input & output flags fields */
};


/* return values for  choose_authen(); */

#define CHOOSE_FAILED -1     /* failed to choose an authentication function */
#define CHOOSE_OK      0     /* successfully chose an authentication function */
#define CHOOSE_GETUSER 1     /* need a username before choosing */
#define CHOOSE_BADTYPE 2     /* Invalid preferred authen function specified */


/*
 * This structure is the data structure for passing information to
 * and from the authorization function (do_author()).
 */
struct author_data {
    struct identity *id;	/* user id */
    int authen_method;		/* authentication method */

#define AUTHEN_METH_NONE             0x01
#define AUTHEN_METH_KRB5             0x02
#define AUTHEN_METH_LINE             0x03
#define AUTHEN_METH_ENABLE           0x04
#define AUTHEN_METH_LOCAL            0x05
#define AUTHEN_METH_TACACSPLUS       0x06
#define AUTHEN_METH_RCMD             0x20

    int authen_type;		/* authentication type see authen_type */
    int service;		/* calling service */
    char *msg;		        /* optional NULL-terminated return message */
    char *admin_msg;	        /* optional NULL-terminated admin message */
    int status;			/* return status */

#define AUTHOR_STATUS_PASS_ADD       0x01
#define AUTHOR_STATUS_PASS_REPL      0x02
#define AUTHOR_STATUS_FAIL           0x10
#define AUTHOR_STATUS_ERROR          0x11

    int num_in_args;		/* input arg count */
    char **input_args;		/* input arguments */
    int num_out_args;		/* output arg cnt */
    char **output_args;		/* output arguments */

};

/* An API accounting record structure */
struct acct_rec {
    int acct_type;		/* start, stop, update */

#define ACCT_TYPE_START      1
#define ACCT_TYPE_STOP       2
#define ACCT_TYPE_UPDATE     3

    struct identity *identity;
    int authen_method;
    int authen_type;
    int authen_service;
    char *msg;       /* output field */
    char *admin_msg; /* output field */
    int num_args;
    char **args;
};

#ifndef TAC_PLUS_PORT
#define	TAC_PLUS_PORT			49
#endif

#define TAC_PLUS_READ_TIMEOUT		180	/* seconds */
#define TAC_PLUS_WRITE_TIMEOUT		180	/* seconds */

#define NAS_PORT_MAX_LEN                255

typedef struct _naslist {        /* nas address-key list */
  struct in_addr addr;  /* nas address */
  char *key;            /* key for this nas */
  char *rshuser;        /* username for rsh commands to this nas */
#ifdef ENABLE_SNMP
  char *snmpcomm;       /* snmp community */
#endif
  struct _naslist *next;
} naslist;
extern naslist *nas_list;
extern naslist *nas_new;

struct session {
    int session_id;                /* host specific unique session id */
    int aborted;                   /* have we received an abort flag? */
    int seq_no;                    /* seq. no. of last packet exchanged */
    time_t last_exch;              /* time of last packet exchange */
    int sock;                      /* socket for this connection */
    char *key;                     /* the key */
    int keyline;                   /* line number key was found on */
    char *peer;                    /* name of connected peer */
    struct in_addr nasaddr;        /* nas ip address */
    naslist *nas;                  /* pointer to nas info struct */ /* FIXME duplication of nas->addr */
#ifdef ALIASES
    char *peeralias;               /* alias of connected peer */
    int flags;			   /* flags			*/
#endif
    char *cfgfile;                 /* config file name */
    char *acctfile;                /* name of accounting file */
    char port[NAS_PORT_MAX_LEN+1]; /* For error reporting */
    u_char version;                /* version of last packet read */
};
extern char *defaultkey;

extern struct session session;     /* the session */

/* Global variables */

extern int debug;                  /* debugging flag */
extern int logging;                /* syslog logging flag */
extern int single;                 /* do not fork (for debugging) */
extern int console;                /* log to console */
extern FILE *ostream;              /* for logging to console */
extern int parse_only;             /* exit after parsing verbosely */
extern int sendauth_only;          /* don't do sendauth */

/* All tacacs+ packets have the same header format */

struct tac_plus_pak_hdr {
    u_char version;

#define TAC_PLUS_MAJOR_VER_MASK 0xf0
#define TAC_PLUS_MAJOR_VER      0xc0

#define TAC_PLUS_MINOR_VER_0    0x0
#define TAC_PLUS_VER_0  (TAC_PLUS_MAJOR_VER | TAC_PLUS_MINOR_VER_0)

#define TAC_PLUS_MINOR_VER_1    0x01
#define TAC_PLUS_VER_1  (TAC_PLUS_MAJOR_VER | TAC_PLUS_MINOR_VER_1)

    u_char type;

#define TAC_PLUS_AUTHEN			1
#define TAC_PLUS_AUTHOR			2
#define TAC_PLUS_ACCT			3

    u_char seq_no;		/* packet sequence number */
    u_char flags;

#define TAC_PLUS_UNENCRYPTED_FLAG    0x01 /* packet is encrypted or cleartext */
#define TAC_PLUS_SINGLE_CONNECT_FLAG 0x04 /* the single-connection flag */

    int session_id;		/* session identifier FIXME: Is this needed? */
    int datalength;		/* length of encrypted data following this
				 * header */
    /* datalength bytes of encrypted data */
};

#define HASH_TAB_SIZE 157        /* user and group hash table sizes */

#define TAC_PLUS_HDR_SIZE 12

typedef struct tac_plus_pak_hdr HDR;

/* Authentication packet NAS sends to us */ 

struct authen_start {
    u_char action;

#define TAC_PLUS_AUTHEN_LOGIN    0x1
#define TAC_PLUS_AUTHEN_CHPASS   0x2
#define TAC_PLUS_AUTHEN_SENDPASS 0x3 /* deprecated */
#define TAC_PLUS_AUTHEN_SENDAUTH 0x4

    u_char priv_lvl;

#define TAC_PLUS_PRIV_LVL_MIN 0x0
#define TAC_PLUS_PRIV_LVL_MAX 0xf

    u_char authen_type;

#define TAC_PLUS_AUTHEN_TYPE_ASCII  1
#define TAC_PLUS_AUTHEN_TYPE_PAP    2
#define TAC_PLUS_AUTHEN_TYPE_CHAP   3
#define TAC_PLUS_AUTHEN_TYPE_ARAP   4
#ifdef MSCHAP
#define TAC_PLUS_AUTHEN_TYPE_MSCHAP 5
#endif /* MSCHAP */

    u_char service;

#define TAC_PLUS_AUTHEN_SVC_LOGIN  1
#define TAC_PLUS_AUTHEN_SVC_ENABLE 2
#define TAC_PLUS_AUTHEN_SVC_PPP    3
#define TAC_PLUS_AUTHEN_SVC_ARAP   4
#define TAC_PLUS_AUTHEN_SVC_PT     5
#define TAC_PLUS_AUTHEN_SVC_RCMD   6
#define TAC_PLUS_AUTHEN_SVC_X25    7
#define TAC_PLUS_AUTHEN_SVC_NASI   8

#ifdef PERUSER_ENABLE
 #define TAC_PLUS_AUTHEN_SVC_ENUSER 9
#endif

    u_char user_len;
    u_char port_len;
    u_char rem_addr_len;
    u_char data_len;
    /* <user_len bytes of char data> */
    /* <port_len bytes of char data> */
    /* <rem_addr_len bytes of u_char data> */
    /* <data_len bytes of u_char data> */
};

#define TAC_AUTHEN_START_FIXED_FIELDS_SIZE 8

/* Authentication continue packet NAS sends to us */ 
struct authen_cont {
    u_short user_msg_len;
    u_short user_data_len;
    u_char flags;

#define TAC_PLUS_CONTINUE_FLAG_ABORT 0x1

    /* <user_msg_len bytes of u_char data> */
    /* <user_data_len bytes of u_char data> */
};

#define TAC_AUTHEN_CONT_FIXED_FIELDS_SIZE 5

/* Authentication reply packet we send to NAS */ 
struct authen_reply {
    u_char status;

#define TAC_PLUS_AUTHEN_STATUS_PASS     1
#define TAC_PLUS_AUTHEN_STATUS_FAIL     2
#define TAC_PLUS_AUTHEN_STATUS_GETDATA  3
#define TAC_PLUS_AUTHEN_STATUS_GETUSER  4
#define TAC_PLUS_AUTHEN_STATUS_GETPASS  5
#define TAC_PLUS_AUTHEN_STATUS_RESTART  6
#define TAC_PLUS_AUTHEN_STATUS_ERROR    7 
#define TAC_PLUS_AUTHEN_STATUS_FOLLOW   0x21

    u_char flags;

#define TAC_PLUS_AUTHEN_FLAG_NOECHO     0x1

    u_short msg_len;
    u_short data_len;

    /* <msg_len bytes of char data> */
    /* <data_len bytes of u_char data> */
};

#define TAC_AUTHEN_REPLY_FIXED_FIELDS_SIZE 6

/* An authorization request packet */
struct author {
    u_char authen_method;
    u_char priv_lvl;
    u_char authen_type;
    u_char service;

    u_char user_len;
    u_char port_len;
    u_char rem_addr_len;
    u_char arg_cnt;		/* the number of args */

    /* <arg_cnt u_chars containing the lengths of args 1 to arg n> */
    /* <user_len bytes of char data> */
    /* <port_len bytes of char data> */
    /* <rem_addr_len bytes of u_char data> */
    /* <char data for each arg> */
};

#define TAC_AUTHOR_REQ_FIXED_FIELDS_SIZE 8

/* An authorization reply packet */
struct author_reply {
    u_char status;
    u_char arg_cnt;
    u_short msg_len;
    u_short data_len;

    /* <arg_cnt u_chars containing the lengths of arg 1 to arg n> */
    /* <msg_len bytes of char data> */
    /* <data_len bytes of char data> */
    /* <char data for each arg> */
};

#define TAC_AUTHOR_REPLY_FIXED_FIELDS_SIZE 6

struct acct {
    u_char flags;

#define TAC_PLUS_ACCT_FLAG_MORE     0x1
#define TAC_PLUS_ACCT_FLAG_START    0x2
#define TAC_PLUS_ACCT_FLAG_STOP     0x4
#define TAC_PLUS_ACCT_FLAG_WATCHDOG 0x8
	    
    u_char authen_method;
    u_char priv_lvl;
    u_char authen_type;
    u_char authen_service;
    u_char user_len;
    u_char port_len;
    u_char rem_addr_len;
    u_char arg_cnt; /* the number of cmd args */
    /* one u_char containing size for each arg */
    /* <user_len bytes of char data> */
    /* <port_len bytes of char data> */
    /* <rem_addr_len bytes of u_char data> */
    /* char data for args 1 ... n */
};

#define TAC_ACCT_REQ_FIXED_FIELDS_SIZE 9

struct acct_reply {
    u_short msg_len;
    u_short data_len;
    u_char status;

#define TAC_PLUS_ACCT_STATUS_SUCCESS 0x1
#define TAC_PLUS_ACCT_STATUS_ERROR   0x2
#define TAC_PLUS_ACCT_STATUS_FOLLOW  0x21

};

#define TAC_ACCT_REPLY_FIXED_FIELDS_SIZE 5

/* Odds and ends */
#define TAC_PLUS_MAX_ITERATIONS 50
#undef MIN
#define MIN(a,b) ((a)<(b)?(a):(b))
#define STREQ(a,b) (strcmp(a,b)==0)
#define MAX_INPUT_LINE_LEN 255

/* Debugging flags */

#define DEBUG_PARSE_FLAG     2
#define DEBUG_FORK_FLAG      4
#define DEBUG_AUTHOR_FLAG    8
#define DEBUG_AUTHEN_FLAG    16
#define DEBUG_PASSWD_FLAG    32
#define DEBUG_ACCT_FLAG      64
#define DEBUG_CONFIG_FLAG    128
#define DEBUG_PACKET_FLAG    256
#define DEBUG_HEX_FLAG       512
#define DEBUG_MD5_HASH_FLAG  1024
#define DEBUG_XOR_FLAG       2048
#define DEBUG_CLEAN_FLAG     4096
#define DEBUG_SUBST_FLAG     8192
#define DEBUG_PROXY_FLAG     16384
#define DEBUG_MAXSESS_FLAG   32768
#define DEBUG_LOCK_FLAG      65536

extern char *codestring();
extern int keycode();

#define TAC_IS_USER           1
#define TAC_PLUS_RECURSE      1
#define TAC_PLUS_NORECURSE    0

#ifndef DEFAULT_USERNAME
 #define DEFAULT_USERNAME "DEFAULT"
#endif

#include "parse.h"

/* Node types */

#define N_arg           50
#define N_optarg        51
#define N_svc_exec      52
#define N_svc_slip      53
#define N_svc_ppp       54
#define N_svc_arap      55
#define N_svc_cmd       56
#define N_permit        57
#define N_deny          58
#define N_svc           59

/* A parse tree node */
struct node {
    int type;     /* node type (arg, svc, proto) */
    void *next;   /* pointer to next node in chain */
    void *value;  /* node value */
    void *value1; /* node value */
    int dflt;     /* default value for node */
    int line;     /* line number declared on */
};

typedef struct node NODE;

union v {
    int intval;
    void *pval;
};

typedef union v VALUE;

/* acct.c */
extern void accounting();

/* report.c */
extern void report_string();
extern void report_hex();
#ifdef __STDC__
extern void report(int priority, char *fmt,...);
#else
extern void report();
#endif

/* packet.c */
extern u_char *get_authen_continue();
extern int send_authen_reply();
/* extern void send_authen_reply(int, char *, u_short, char *, u_short, u_char); */

/* utils.c */
extern char *tac_malloc();
extern char *tac_strdup();
extern char *tac_make_string();
extern char *tac_find_substring();
extern char *tac_realloc();

/* dump.c */
extern char *summarise_outgoing_packet_type();
extern char *summarise_incoming_packet_type();

/* author.c */
extern void author();

/* hash.c */
extern void *hash_add_entry();
extern void **hash_get_entries();
extern void *hash_lookup();

/* config.c */
extern int cfg_get_intvalue();
extern char * cfg_get_pvalue();
extern char *cfg_get_authen_default();
extern char **cfg_get_svc_attrs();
extern NODE *cfg_get_cmd_node();
extern NODE *cfg_get_svc_node();
extern char *cfg_get_expires();
extern char *cfg_get_login_secret();
extern int cfg_get_user_nopasswd();
extern char *cfg_get_arap_secret();
extern char *cfg_get_chap_secret();
#ifdef MSCHAP
extern char *cfg_get_mschap_secret();
#endif /* MSCHAP */
extern char *cfg_get_pap_secret();
extern char *cfg_get_opap_secret();
extern char *cfg_get_global_secret();
extern void cfg_clean_config();
extern char *cfg_nodestring();

#ifdef DENY
 extern int set_sysadm_status(char *, struct authen_data *, int);
 extern char *cfg_get_deny_default(char *, int);
#endif
#ifdef CRON
 extern int set_time_status(char *, struct authen_data *, struct passwd *, int);
 extern char *cfg_get_cron_default(char *, int);
#endif
#ifdef UNIX_LOGIN
 extern char *cfg_get_prompt_default();
 extern char *cfg_get_reject_default(char *, int);
#endif

/* pw.c */
extern struct passwd *tac_passwd_lookup();

/* parse.c */
extern void parser_init();

/* pwlib.c */
extern void set_expiration_status();

/* set_adm_status.c */
extern int sys_adm_verify(char *, struct passwd *, struct authen_data *, int);
#ifdef DBF
extern int set_usradm_status(char *, struct authen_data *);
#endif
#ifdef ALIASES
extern char *get_peer_alias(char *, int *);
#endif
#ifdef TTYS
extern char *get_port_alias(char *, char*, char *, int *, int *);
#endif

/* programs.c */
#ifdef MORE_SUBST
extern char *substitute3(char *, struct author_data *, int);
#define substitute(s, d) substitute3(s, d, 1)
#define substitute2(s, d) substitute3(s, d, 0)
#endif

/* miscellaneous */
#ifndef HAS_SYSERRLIST
extern const char *const sys_errlist[];
extern const int sys_nerr;
#endif
/* extern char *sys_errlist[]; */
/* extern int sys_nerr; */
#ifndef HAVE_STRERROR
#define strerror(n) \
    (((n) >= 0 && (n) < sys_nerr) ? sys_errlist[n] : "unknown error")
#endif
extern int errno;
extern int sendauth_fn();
extern int sendpass_fn();
extern int enable_fn();
extern int default_fn();
extern int default_v0_fn();
extern int skey_fn();
#ifdef MSCHAP
extern void mschap_lmchallengeresponse();
extern void mschap_ntchallengeresponse();
#endif /* MSCHAP */

#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif

#ifdef MAXSESS

extern void maxsess_loginit();
extern int maxsess_check_count();

/*
 * This is state kept per user/session
 */
struct peruser {
    char username[64];		/* User name */
    char NAS_name[32];		/* NAS user logged into */
    char NAS_port[32];		/*  ...port on that NAS */
    char NAC_address[32];	/*  ...IP address of NAS */
};

#endif /* MAXSESS */

#ifndef LOGFILE_DEFAULT
#define LOGFILE_DEFAULT LOGPATH "/tac_plus.log"
#endif
extern struct timeval started_at;
extern char *logfile;
extern char *wtmpfile;
extern int wtmpfd;

#define fr_fi(x, y) {if(x) free(x); x = y;}

#ifdef ACL
#include "acl.h"
#endif

#ifdef UDB
#ifdef HAVE_DB_185_H
#include <db_185.h>
#else
#include <db.h>
#endif
extern DB *db_register();
#endif

/* some internal prototypes */
#ifdef TACACS_PLUS
#ifdef PERUSER_ENABLE
extern char *cfg_get_enable(char *, int);
#endif
#ifdef REGNAS
extern void reg_nas_users(void);
#endif
#ifdef TAC_ROAMING
extern void cfg_get_tac_roaming_data(char *, int, char **, char **);
#endif
#ifdef UDB
char *db_replace(char *, DB *, char *);
#endif
#ifdef DYNA
extern char *dyna_req(char *, char *, char *, int);
extern void dyna_sethost(char *);
#endif
extern char *crypt(const char *, const char *);
extern int call_post_process(char *, struct author_data *, char ***, int *);
extern int call_pre_process(char *, struct author_data *, char ***, int *, char *, int);
extern int call_extern_authen(char *, struct authen_data *, char *);
extern int cfg_no_user_permitted(void);
extern int cfg_ppp_is_configured(char *, int);
extern int cfg_read_config(char *);
extern int cfg_user_exists(char *);
extern int cfg_user_svc_default_is_permit(char *);
extern int choose_authen(struct authen_data *, struct authen_type *);
extern int des_verify(char *, char *);
extern int do_acct(struct acct_rec *);
extern int do_author(struct author_data *);
extern int md5_xor(HDR *, u_char *, char *);
extern int tac_lockfd(char *, int, int);
extern int tac_unlockfd(char *, int, int);
extern int verify(char *, char *, struct authen_data *, int);
extern u_char *read_packet(void);
extern int write_packet(u_char *);
/* extern void dump_nas_pak(u_char *); */
/* extern void dump_tacacs_pak(u_char *); */
/* extern void loguser(struct acct_rec *); */
/* extern void send_acct_reply(u_char, char *, char *); */
/* extern void send_authen_error(char *); */
/* extern void send_author_reply(u_char, char *, char *, int, char **); */
/* extern void send_error_reply(int, char *); */
extern int tac_exit(int);
extern int do_wtmp(struct acct_rec *);
extern void retrieve_secret(char *, char **);
#if defined(ACL) && defined(DXS1) || defined(REGNAS)
extern char *call_rsh(char *);
#endif
extern int call_extern_acct(char *, struct acct_rec *);
#endif
