/*
 * advanced tacacs+ server based on Cisco Systems tac_plus code.
 * For Cisco copyright notice, see LICENSE.CISCO
 *
 * This is a copyrighted commercial software; see LICENSE for details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#ifdef HAVE_DB_185_H
#include <db_185.h>
#else
#include <db.h>
#endif

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <errno.h>

extern char *crypt(const char *, const char *);

static  char    dbname[FILENAME_MAX];
static  DB      *db;
int     verbose = 0;
int	tacspec = 0;

void    edit(char *), dbappend(), opendb(int),
	change_pass(int, char *, char *), error();

int main(ac, av)
	int     ac;
	char   *av[];
{
	int     c;
	DBT     k[1], d[1];
	char    *p;
	int	i;
	struct in_addr iptmp;

	if (ac < 3)
		goto usage;
	av++;   ac--;

	if ((p = (char *) strrchr(*av, '.')) && !strcmp(p, ".db"))
		strcpy(dbname, *av);
	else
		sprintf(dbname, "%s.db", *av);

	while ((c = getopt(ac, av, "ve:p:P:LCad:g:i")) != -1) {
		switch (c) {
		    case 'v':
			verbose++;
			break;
		    case 'i':
			tacspec++;
			break;
		    case 'e':                   /* edit entry */
			opendb(O_RDWR);
			if (optind != ac)
				goto usage;
			edit(optarg);
			exit(0);
		    case 'P':
		    case 'p':                   /* change password */
			opendb(O_RDWR);
			if (optind == ac)
				change_pass(c == 'P', optarg, NULL);
			else if (optind == (ac - 1))
				change_pass(c == 'P', optarg, av[optind]);
			else
				goto usage;
			exit(0);
		    case 'L':                   /* database dump */
			opendb(O_RDONLY);
			while(!(*db->seq) (db, k, d, R_NEXT)) {
				if (tacspec) {
					fwrite(k->data + sizeof(u_long), 1, k->size - sizeof(u_long), stdout);
					fputc(' ',stdout);
					for (i=k->size - sizeof(u_long); i<10; i++) fputc(' ',stdout);
					iptmp.s_addr = ntohl(* (u_long *) k->data);
					fprintf(stdout, "time=%li\taddr=%s\n", * (time_t *) d->data, inet_ntoa(iptmp));
				} else {
					fwrite(k->data, 1, k->size, stdout);
					putc('\t', stdout);
					fwrite(d->data, 1, d->size, stdout);
					putc('\n', stdout);
				}
			}
			(*db->close) (db);
			exit(0);
		    case 'C':                   /* create db    */
			if (optind != ac)
				goto usage;
			if (!access(dbname, F_OK) && unlink(dbname) < 0)
				error(1, "can't unlink old db %s\n", dbname);
			if (verbose)
				error(0, "creating db %s\n", dbname);
			opendb(O_RDWR|O_CREAT);
			dbappend();
			exit(0);
		    case 'a':                   /* append to db */
			if (optind != ac)
				goto usage;
			opendb(O_RDWR);
			dbappend();
			exit(0);
		    case 'd':                   /* delete key  */
			optind--;
			opendb(O_RDWR);
			do {
			  if (tacspec) {
			    int found = 0;
			    while(!(*db->seq) (db, k, d, R_NEXT)) {
			      char *tmp = (char *) malloc(k->size + 1 - sizeof(u_long));
			      strncpy(tmp, k->data + sizeof(u_long), k->size - sizeof(u_long));
			      tmp[k->size - sizeof(u_long)]='\0';
			      if (strcmp(av[optind], tmp)) {
				free(tmp);
				continue;
			      }
			      found++;
			      if ((*db->del) (db, k, NULL))
				error(0, "unable to delete key %s\n", tmp);
			      free(tmp);
			    }
			    if (!found)
			      error(0, "can't get key %s\n", av[optind]);
			  } else {
			    k->size = strlen(av[optind]);
			    k->data = av[optind];
			    if (verbose)
			      error(0, "deleting key %s\n", k->data);
			    if ((*db->del) (db, k, NULL))
			      error(0, "unable to delete key %s\n", k->data);
			  }
			  optind++;
			} while (optind < ac);
			(*db->close) (db);
			exit(0);
		    case 'g':                   /* get key     */
			optind--;
			opendb(O_RDONLY);
			do {
				if (tacspec) {
				  int found = 0;
				  while(!(*db->seq) (db, k, d, R_NEXT)) {
				    char *tmp = (char *) malloc(k->size + 1 - sizeof(u_long));
				    strncpy(tmp, k->data + sizeof(u_long), k->size - sizeof(u_long));
				    tmp[k->size - sizeof(u_long)]='\0';
				    if (strcmp(av[optind], tmp)) {
				      free(tmp);
				      continue;
				    }
				    free(tmp);
				    found++;
				    fwrite(k->data + sizeof(u_long), 1, k->size - sizeof(u_long), stdout);
				    fputc(' ',stdout);
				    for (i=k->size - sizeof(u_long); i<10; i++) fputc(' ',stdout);
				    iptmp.s_addr = ntohl(* (u_long *) k->data);
				    fprintf(stdout, "time=%li\taddr=%s\n", *(time_t *)d->data, inet_ntoa(iptmp));
				  }
				  if (!found)
				    error(0, "can't get key %s\n", av[optind]);
				} else {
					k->size = strlen(av[optind]);
					k->data = av[optind];
					if ((*db->get) (db, k, d, NULL))
						error(0, "can't get key %s\n", k->data);
					else {
						fwrite(k->data, 1, k->size, stdout);
						putc('\t', stdout);
						fwrite(d->data, 1, d->size, stdout);
						putc('\n', stdout);
					}
				}
				optind++;
			} while (optind < ac);
			(*db->close) (db);
			exit(0);
		     default:
usage:			error(1, "Usage: tdb dbname [OPTION]\n"
			"  -a append to db\n"
			"  -C create db\n"
			"  -d key ... delete key\n"
			"  -e key edit entry\n"
			"  -g key ... get key\n"
			"  -i specific for dynipd db\n"
			"  -L database dump\n"
			"  -p|-P user [passw] change password\n"
			"  -v verbose\n");
		}
	}
	goto usage;
}

/*
 * Open database
 */
void    opendb(mode)
	int     mode;
{
	struct flock flock;

	db = dbopen(dbname, mode, 0600, DB_HASH, NULL);
	if (!db)
		error(1, "can't %s database %s\n", (mode & O_CREAT)?
			"create":"open", dbname);
	flock.l_type   = (mode != O_RDONLY) ? F_WRLCK : F_RDLCK;
	flock.l_whence = SEEK_SET;
	flock.l_start  = 0L;
	flock.l_len    = 0L;
	if (fcntl((*db->fd)(db), F_SETLKW, &flock) < 0)
	  error(1, "unable to lock database %s: %s", dbname, strerror(errno));
}

/*
 * Append to database
 */
void    dbappend()
{
	unsigned char   iline[BUFSIZ], *p;
	char *p1;
	DBT             k[1], d[1];
	int             line = 0;
	time_t time;
	u_long addr;
	char tkey[128];

	if (verbose)
		error(0, "append to db %s\n", dbname);

	while(fgets(iline, BUFSIZ, stdin)) {
		line++;
		if (*iline == '#')
			continue;       /* comment */
		for (p = iline; *p && *p != '\n'; p++)
			;
		*p = '\0';
		for (p = iline; *p != '\t' && *p != ' ' && *p; p++)
			;
		if (!*p) {
badline:
			error(0, "bad input line %d; ignored\n", line);
			continue;
		}
		if (tacspec) {
			*p = '\0';
			for(p++; *p && (*p == ' ' || *p == '\t'); p++);
			if (!*p) goto badline;
			if (strncmp(p,"time=",5)) goto badline; p+=5;
			errno=0;
			time=strtoul(p,&p1,10);
			if (errno != 0) goto badline;
			p=p1;
			for(p++; *p && (*p == ' ' || *p == '\t'); p++);
			if (!*p) goto badline;
			if (strncmp(p,"addr=",5)) goto badline; p+=5;
			addr=htonl(inet_addr(p));
			if (addr == -1) goto badline;
			if (verbose)
				error(0, "adding user=%s, addr=%s, time=%i\n", iline, p, time);

			k->size = strlen(iline) + sizeof(u_long);
			k->data = tkey;
			* (u_long *) tkey = addr;
			strcpy(k->data + sizeof(u_long), iline);
			d->data = &time;
			d->size = sizeof(time);
		} else {
			*p++ = '\0';
			while(*p == ' ' || *p == '\t')
				p++;
			if (!*p)
				goto badline;
			if (verbose)
				error(0, "adding key=%s data=%s\n", iline, p);
			k->data = iline;
			k->size = strlen(iline);
			d->data = p;
			d->size = strlen(p);
		}
		if ((*db->put) (db, k, d, NULL))
			error(0, "can't put key %s to db %s (input line %d)\n",
				iline, dbname, line);
	}
	(*db->close) (db);
}

/*
 * Edit entry
 */
void    edit(user)
	char    *user;
{
	DBT     k[1], d[1];
	char    *temp, *editor, buf[BUFSIZ];
	char    proto[] = "/tmp/tdbXXXXXX";
	FILE    *fp;
	int	i;
	struct in_addr iptmp;

	temp = (char *) mktemp(proto);
	if (!(fp = fopen(temp, "w")))
err:
		error(1, "can't create/reopen temp file %s\n", temp);
	if (tacspec) {
	  int found = 0;
	  while(!(*db->seq) (db, k, d, R_NEXT)) {
	    char *tmp = (char *) malloc(k->size + 1 - sizeof(u_long));
	    strncpy(tmp, k->data + sizeof(u_long), k->size - sizeof(u_long));
	    tmp[k->size - sizeof(u_long)]='\0';
	    if (strcmp(user, tmp)) {
	      free(tmp);
	      continue;
	    }
	    free(tmp);
	    found++;
	    fwrite(k->data + sizeof(u_long), 1, k->size - sizeof(u_long), fp);
	    fputc(' ',fp);
	    for (i=k->size - sizeof(u_long); i<10; i++) fputc(' ',fp);
	    iptmp.s_addr = ntohl(* (u_long *) k->data);
	    fprintf(fp, "time=%li\taddr=%s\n", *(time_t *)d->data, inet_ntoa(iptmp));
	  }
	  if (!found)
	    error(1, "can't get db entry for user %s\n", user);
	} else {
	  k->size = strlen(user);
	  k->data = user;
	  if ((*db->get) (db, k, d, NULL))
	    error(1, "can't get db entry for user %s\n", user);
	  fwrite(k->data, 1, k->size, fp);
	  putc('\t', fp);
	  fwrite(d->data, 1, d->size, fp);
	  putc('\n', fp);
	}
	fclose(fp);
	editor = (char *) getenv("EDITOR");
	if (!editor)
		editor = "vi";
	sprintf(buf, "%s %s", editor, temp);
	if(system(buf))
		error(1, "error while executing editor\n");
	close(0);
	if (!fopen(temp, "r"))
		goto err;
	unlink(temp);
	/* gluk ... */
	sprintf(buf, "%s.bak", temp);
	unlink(buf);
	dbappend();
}

/*
 * Change password for user
 */
void    change_pass(docrypt, user, pass)
	int     docrypt;
	char    *user, *pass;
{
	DBT     k[1], d[1];
	char    buf[BUFSIZ], *p, salt[2], *np1, np2[BUFSIZ];
	int     ntries, offs;

	k->size = strlen(user);
	k->data = user;

	if ((*db->get) (db, k, d, NULL))
		error(1, "can't get db entry for user %s\n", user);

	for (p = d->data, offs = 0; offs < d->size; offs++, p++)
		if (*p == '|')
			break;
	salt[0] = 'A' + (getpid() % 17);
	salt[1] = 'a' + (time(NULL) % 17);
	ntries = 0;
	if (!pass) {
		for (;;) {
			np1 = getpass("Enter new password: ");
			strcpy(np2, np1);
			np1 = getpass("Repeat new password: ");
			if (!strcmp(np1, np2))
				break;
			error(0, "Passwords mismatch; please try again.\n");
			if (ntries++ > 2)
				error(1, "\nAborted.\n");
		}
		np1 = crypt(np2, salt);
	} else
		np1 = docrypt ? crypt(pass, salt) : pass;
	strcpy(buf, np1);
	if (*p == '|')
		strncat(buf, p, d->size - offs);
	d->data = buf;
	d->size = strlen(buf);

	if ((*db->put) (db, k, d, NULL))
		error(1, "can't update key %s\n", user);

	(*db->close)(db);
}

void error(lvl, s, a1, a2, a3, a4)
	char    *s;
	int     lvl, a1, a2, a3, a4;
{
	fprintf(stderr, s, a1, a2, a3, a4);
	if (lvl)
		exit(lvl);
}
